# ezcrypt [![CI status](https://ci.codeberg.org/api/badges/14265/status.svg)](https://ci.codeberg.org/repos/14265)

ezcrypt is an easy to use tool for strong file encryption.

* [About](#about)
* [Installation](#installation)
* [Example usage](#example-usage)
* [Principles](#principles)
* [Resources](#resources)

## About

### Goals

ezcrypt is a (serious) hobby project with a few design goals:

* **Security**: Strong encryption and robust architecture.
* **Simplicity**: Easy to use.
* **Portability**: Works on most systems (even low end systems).
* **Public domain**: Open source and completely free for everyone.

### Features

#### Easy to use

- Plain and simple encryption/decryption of any file with a passphrase.
- No cryptographic keys required (although pepper files are supported).
- Familiar CLI interface, similar to [gzip](https://www.gzip.org/).

#### Resistant against cryptanalytic attacks

- Strong encryption, making [brute-force](https://en.wikipedia.org/wiki/Brute-force_attack) key attacks impractical.
  - Four levels of encryption ([cascade encryption](https://en.wikipedia.org/wiki/Multiple_encryption)), each with a 256-bit key.
  - The total effective key space is <span title="179,769,313,486,231,590,772,930,519,078,902,473,361,797,697,894,230,657,273,430,081,157,732,675,805,500,963,132,708,477,322,407,536,021,120,113,879,871,393,357,658,789,768,814,416,622,492,847,430,639,474,124,377,767,893,424,865,485,276,302,219,601,246,094,119,453,082,952,085,005,768,838,150,682,342,462,881,473,913,110,540,827,237,163,350,510,684,586,298,239,947,245,938,479,716,304,835,356,329,624,224,137,216">2<sup>1024</sup></span> (for reference, the age of the universe is less than 2<sup>79</sup> microseconds).
- High cost [key derivation function](https://en.wikipedia.org/wiki/Key_derivation_function), making brute-force passphrase attacks impractical.
  - Configurable cost, up to several minutes per passphrase-to-key derivation on a 5 GHz CPU core.
  - Cache hard algorithm, making GPU implementations inefficient.
  - Strong [salt](https://en.wikipedia.org/wiki/Salt_(cryptography)), making precomputed [rainbow table](https://en.wikipedia.org/wiki/Rainbow_table) attacks impractical.
  - Optional strong [secret pepper](https://en.wikipedia.org/wiki/Pepper_(cryptography)) for additional security.
- [Authenticated encryption](https://en.wikipedia.org/wiki/Authenticated_encryption) using [Poly1305](https://en.wikipedia.org/wiki/Poly1305).
  - Both privacy and authenticity of the data is assured.
  - Tampering with the encrypted data is infeasible.

#### Portable

- Written in portable [C11](https://en.wikipedia.org/wiki/C11_(C_standard_revision)).
- Works on most operating systems (including Linux, macOS, Windows, FreeBSD).
- Works on most CPU architectures (including 64- and 32-bit x86, ARM, RISC-V, etc).
- Fully self contained without any dependencies on 3rd party cryptography libraries.

#### Free, open source and public domain

All code is free and unencumbered software released into the public domain, including the cryptographic algorithms.

For more information, see [unlicense.org](https://unlicense.org/).

### FAQ

See [Frequently asked questions](doc/faq.md).

## Installation

Prerequisites: A C compiler and [CMake](https://cmake.org/). For Linux targets, GTK 3 is also recommended to enable GUI dialogs (e.g. `apt install libgtk-3-dev` on Ubuntu).

To build:

```bash
mkdir out && cd out
cmake -DCMAKE_BUILD_TYPE=Release ../src
cmake --build .
```

The resulting executable file is `out/ezcrypt`.

To install (from the `out` folder):

```bash
sudo cmake --install .
```

### Testing

To run a full build-and-test suite in a Docker environment (from the repo root):

```bash
docker-compose build
docker-compose run --rm ezcrypt-test
```

## Example usage

The canonical help for ezcrypt can be obtained with:

```bash
$ ezcrypt --help
```

### Encrypt a file

Encrypt the file `myfile`, with the passphrase provided via a terminal prompt (or a GUI prompt where available). The output file is called `myfile.z` (the original file is kept):

```bash
$ ezcrypt myfile
Enter passphrase:
Please repeat the passphrase:
```

### Decrypt a file

Decrypt the file `myfile.z`, with the passphrase provided via a terminal prompt. The output file is called `myfile` (the original file is kept):

```bash
$ ezcrypt -d myfile.z
Enter passphrase:
```

### Decrypt and print a file

Decrypt the file `myfile.z` to stdout, with the passphrase provided via the environment variable `$SECRET`:

```bash
$ ezcrypt --show -E SECRET myfile.z
```

### Encrypt & decrypt via pipes

```bash
$ echo "Hello world!" | ezcrypt -E SECRET | ezcrypt -d -E SECRET
Hello world!
```

### Edit an encrypted text file

Edit the plaintext contents of the encrypted file `myfile.z`, using the default text editor (e.g. `$EDITOR` or `notepad.exe`):

```bash
$ ezcrypt --edit myfile.z
```

Note: If the plaintext is not modified by the editor, `myfile.z` remains unmodified.

## Principles

### File format

![ezcrypt file format](doc/fileformat.png)

Encryption is done in four layers. At each level a different [cipher](https://en.wikipedia.org/wiki/Cipher) is used, and each level has its own [encyrption key](https://en.wikipedia.org/wiki/Key_(cryptography)) and its own [initialization vector](https://en.wikipedia.org/wiki/Initialization_vector) (IV). The different ciphers are:

1. [AES](https://en.wikipedia.org/wiki/Advanced_Encryption_Standard), CBC, 256-bit key (outermost level)
2. [Twofish](https://en.wikipedia.org/wiki/Twofish), CBC, 256-bit key
3. [Serpent](https://en.wikipedia.org/wiki/Serpent_(cipher)), CBC, 256-bit key
4. [ChaCha](https://en.wikipedia.org/wiki/Salsa20#ChaCha_variant), 20 rounds, 256-bit key, XChaCha20 construction (innermost level)

The [salt](https://en.wikipedia.org/wiki/Salt_(cryptography)) and the IV for each encryption level is generated from system level [entropy](https://en.wikipedia.org/wiki/Entropy_(computing)) (i.e. highly random data), and is different for each run of ezcrypt. Thus encrypting the same file twice will result in two different [ciphertexts](https://en.wikipedia.org/wiki/Ciphertext) (even if the same passphrase is used).

Note that the encrypted file does not contain any header or other identification metadata. This is by design.

### Key derivation

![ezcrypt key derivation](doc/key-derivation.png)

The key at each level is generated from a combination of the user supplied [passphrase](https://en.wikipedia.org/wiki/Passphrase), an optional user supplied [pepper](https://en.wikipedia.org/wiki/Pepper_(cryptography)) file (hashed to 256 bits) and a per-level 256-bit [salt](https://en.wikipedia.org/wiki/Salt_(cryptography)). This is done using a compute intensive key derivation function called [ZKDF](doc/zkdf.md).

## Resources

* [Codeberg](https://codeberg.org/ezcrypt/ezcrypt) (main host)
* [repo.or.cz](https://repo.or.cz/ezcrypt.git) (mirror)
* <a rel="me" href="https://techhub.social/@ezcrypt">Mastodon</a>
