// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// ezcrypt - Encryption & decryption CLI tool.
//------------------------------------------------------------------------------

#if !defined(_WIN32)
// Define this macro to get struct sigaction from signal.h.
// NOLINTNEXTLINE(bugprone-reserved-identifier,cert-dcl37-c,cert-dcl51-cpp)
#define _XOPEN_SOURCE 700
#endif

#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32)
#include <fcntl.h>
#include <io.h>
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

#if !defined(_WIN32)
#include <sys/resource.h>
#include <sys/time.h>
#define HAS_SETRLIMIT
#endif

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#include <sys/param.h>
#if defined(BSD) && (BSD >= 201408)
#include <sys/procctl.h>
#define HAS_PROCCTL
#endif
#endif

#if defined(__linux__) || defined(linux) || defined(__linux) || \
    defined(__gnu_linux__)
#include <sys/prctl.h>
#define HAS_PRCTL
#endif

#include "libezcrypt/arguments.h"
#include "libezcrypt/crypt_stream.h"
#include "libezcrypt/error.h"
#include "libezcrypt/fileutils.h"
#include "libezcrypt/passphrase.h"
#include "libezcrypt/pepper.h"
#include "libezcrypt/stringutils.h"

//------------------------------------------------------------------------------
// Constants and macros.
//------------------------------------------------------------------------------

// File extension for encrypted files.
#define FILE_EXT ".z"
#define FILE_EXTLEN 2

// Maximum length of a passphrase.
#define MAX_PASSPHRASE_SIZE 1024

// Maximum supported full file name length (for statically allocated strings).
#define MAX_FILE_NAME_SIZE 1024

//------------------------------------------------------------------------------
// Helper functions.
//------------------------------------------------------------------------------

static char* make_outfile_name(const char* infile, bool encrypt) {
  if (infile == NULL) {
    return NULL;
  }

  size_t len = strlen(infile);
  char* outfile = NULL;
  if (encrypt) {
    // Append FILE_EXT to the infile name.
    outfile = (char*)malloc(len + FILE_EXTLEN + 1);
    if (outfile != NULL) {
      memcpy(outfile, infile, len);
      memcpy(outfile + len, FILE_EXT, FILE_EXTLEN);
      outfile[len + FILE_EXTLEN] = 0;
    }
  } else {
    // Check that infile ends with FILE_EXT.
    const char* p = strrchr(infile, '.');
    if (p != NULL && p != infile && strcmp(p, FILE_EXT) == 0) {
      // Remove FILE_EXT from the infile name.
      outfile = (char*)malloc(len - FILE_EXTLEN + 1);
      if (outfile != NULL) {
        memcpy(outfile, infile, len - FILE_EXTLEN);
        outfile[len - FILE_EXTLEN] = 0;
      }
    }
  }

  return outfile;
}

//------------------------------------------------------------------------------
// Helpers for handling the editor.
//------------------------------------------------------------------------------

static const char* get_default_editor(void) {
  // First try the EDITOR environment variable.
  const char* editor = getenv("EDITOR");
  if (editor != NULL) {
    return editor;
  }

#if !defined(_WIN32)
  // Second, try some known editors in order of (some) preference.
  static const int NUM_EDITOR_ALTERNATIVES = 2;
  static const char* EDITOR_ALTERNATIVES[2] = {"nano", "vim"};
  for (int i = 0; i < NUM_EDITOR_ALTERNATIVES; ++i) {
    const char* alternative = EDITOR_ALTERNATIVES[i];
    if (is_exe_in_path(alternative)) {
      return alternative;
    }
  }
#endif

#if defined(_WIN32)
  // Notepad should always be installed on Windows.
  return "notepad.exe";
#else
  // According to the POSIX standard, vi is a mandatory utility.
  // https://pubs.opengroup.org/onlinepubs/9699919799/utilities/vi.html
  return "vi";
#endif
}

static void get_edit_file_ext(const char* ciphertext_filename,
                              char* ext,
                              const size_t ext_size) {
  PRECONDITION(ciphertext_filename != NULL);
  PRECONDITION(ext != NULL);
  PRECONDITION(ext_size > 0U);

  // Default extension if none could be derived.
  if (!string_copy(".txt", ext, ext_size)) {
    FATAL_ERROR("Too short file extension buffer");
  }

  // Get the plaintext filename (drop the cipher extension).
  char* plainttext_filename = make_outfile_name(ciphertext_filename, false);
  if (plainttext_filename != NULL) {
    const char* plaintext_ext = strrchr(plainttext_filename, '.');
    if (plaintext_ext != NULL) {
      (void)string_copy(plaintext_ext, ext, ext_size);
    }
    free(plainttext_filename);
  }
}

//------------------------------------------------------------------------------
// Resource cleanup logic.
//------------------------------------------------------------------------------

static FILE* s_inf = NULL;
static FILE* s_outf = NULL;
static bool s_inf_opened = false;
static bool s_outf_opened = false;

static char s_edit_file_name[MAX_FILE_NAME_SIZE];
static char s_unfinished_output_file_name[MAX_FILE_NAME_SIZE];

static void close_open_files(void) {
  if (s_inf_opened) {
    fclose(s_inf);
    s_inf_opened = false;
  }

  if (s_outf_opened) {
    fclose(s_outf);
    s_outf_opened = false;
  }
}

static void delete_edit_file(void) {
  if (s_edit_file_name[0] != 0) {
    if (!delete_file(s_edit_file_name)) {
      fprintf(stderr,
              "***Warning: Unable to delete temporary edit file: %s\n",
              s_edit_file_name);
    }
    s_edit_file_name[0] = 0;
  }
}

static void set_unfinished_output_file_name(const char* file_name) {
  if (!string_copy(file_name,
                   s_unfinished_output_file_name,
                   sizeof(s_unfinished_output_file_name))) {
    FATAL_ERROR("Too long output file path");
  }
}

static void clear_unfinished_output_file_name(void) {
  s_unfinished_output_file_name[0] = 0;
}

static void delete_unfinished_output_file(void) {
  if (s_unfinished_output_file_name[0] != 0) {
    if (!delete_file(s_unfinished_output_file_name)) {
      fprintf(stderr,
              "***Warning: Unable to delete unfinished output file: %s\n",
              s_unfinished_output_file_name);
    }
    clear_unfinished_output_file_name();
  }
}

// This function is called at exit or at program interruption (CTRL+C or kill),
// to ensure that the edited plaintext file is deleted for instance.
static void program_cleanup(void) {
  close_open_files();
  delete_edit_file();
  delete_unfinished_output_file();
  kill_passphrase_child_processes();
}

//------------------------------------------------------------------------------
// atexit handler.
//------------------------------------------------------------------------------

#if defined(_WIN32)
typedef int process_id_t;
#else
typedef pid_t process_id_t;
#endif
static process_id_t s_main_process_id;

static inline process_id_t get_process_id(void) {
#if defined(_WIN32)
  return _getpid();
#else
  return getpid();
#endif
}

static void atexit_handler(void) {
  if (get_process_id() == s_main_process_id) {
    // Only do cleanup work from the main process.
    program_cleanup();
  }
}

static void setup_atexit_handler(void) {
  // Register the main process ID. This is used by the atexit handler to only do
  // cleanup from the main process.
  s_main_process_id = get_process_id();

  // Register our cleanup code to be run when we exit the program.
  atexit(atexit_handler);
}

//------------------------------------------------------------------------------
// Signal handling.
//------------------------------------------------------------------------------

static void signal_catcher(int signal_id) {
#if defined(_WIN32)
  // Change the signal handler from SIG_DFL (e.g. kill process) to SIG_IGN
  // (ignore) to avoid being interrupted during this handler.
  (void)signal(signal_id, SIG_IGN);
#else
  (void)signal_id;  // Unused.
#endif

  // The purpose of this signal handler is to ensure that the same functions
  // are called as during exit. According to the POSIX standard you can (should)
  // only call async-signal-safe functions, and exit() is not such a function so
  // we can't rely on the atexit functionality.
  program_cleanup();

  // _Exit() is a more light-weight version of exit(). It does not call any
  // atexit routines, for instance.
  _Exit(1);
}

static void setup_signal_handlers(void) {
#if defined(_WIN32)
  // Windows does not have sigaction(), so use signal() instead.
  (void)signal(SIGINT, &signal_catcher);
  (void)signal(SIGTERM, &signal_catcher);
#else
  struct sigaction sa;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = signal_catcher;
  (void)sigaction(SIGINT, &sa, NULL);
  (void)sigaction(SIGTERM, &sa, NULL);
  (void)sigaction(SIGQUIT, &sa, NULL);
#endif
}

static void disable_core_dumps(void) {
#if defined(HAS_SETRLIMIT)
  {
    // Limit the size of core dumps to zero bytes.
    struct rlimit core_limit;
    core_limit.rlim_cur = 0;
    core_limit.rlim_max = 0;
    if (setrlimit(RLIMIT_CORE, &core_limit) < 0) {
      fprintf(stderr,
              "***Warning: Unable to disable crash core dumps (setrlimit).\n");
    }
  }
#endif

#if defined(HAS_PROCCTL)
  {
    // Disable traces and dumps (BSD).
    int trace_ctl = PROC_TRACE_CTL_DISABLE;
    if (procctl(P_PID, getpid(), PROC_TRACE_CTL, &trace_ctl) != 0) {
      fprintf(stderr,
              "***Warning: Unable to disable crash core dumps (procctl).\n");
    }
  }
#endif

#if defined(HAS_PRCTL)
  // Disable dumps (Linux).
  if (prctl(PR_SET_DUMPABLE, 0) != 0) {
    fprintf(stderr,
            "***Warning: Unable to disable crash core dumps (prctl).\n");
  }
#endif
}

//------------------------------------------------------------------------------
// Main program entry.
//------------------------------------------------------------------------------

int main(int argc, char** argv) {
  // Handle cleanup when the program exists.
  setup_atexit_handler();

  // We don't want to leak any information to core dumps in case we crash.
  disable_core_dumps();

  // Catch external signals that tries to kill our process (e.g. CTRL+C).
  setup_signal_handlers();

  // Parse arguments.
  args_t args;
  if (!parse_args(argc, argv, &args)) {
    exit(1);
  }

  // Make a pepper?
  if (args.command == ARG_CMD_MAKE_PEPPER) {
    // Open output file.
    s_outf = create_private_file(args.outfile);
    if (s_outf == NULL) {
      FATAL_ERRORF("Unable to open output file: %s", args.outfile);
    }
    s_outf_opened = true;

    // Generate the pepper file.
    const size_t PEPPER_FILE_LENGTH = 4096;
    if (!make_pepper_file(s_outf, PEPPER_FILE_LENGTH)) {
      // Delete the potentially bad (e.g. zero length) pepper file so that it is
      // not accidentally used.
      close_open_files();
      delete_file(args.outfile);

      // Bail.
      FATAL_ERRORF("Failed to write to the file: %s", args.outfile);
    }

    // Perform program cleanup.
    program_cleanup();
    exit(0);
  }

  // Check if the required files exist. We do this before asking for a
  // passphrase to improve UX, and we avoid acutally opening the files this
  // early as it could potentially leak information in case the passphrase query
  // is compromised.
  if (args.infile != NULL && !file_exists(args.infile)) {
    FATAL_ERRORF("File does not exist: %s", args.infile);
  }
  if (args.pepper_file != NULL && !file_exists(args.pepper_file)) {
    FATAL_ERRORF("File does not exist: %s", args.pepper_file);
  }

  // Get the passphrase.
  char passphrase[MAX_PASSPHRASE_SIZE];
  const bool verify = (args.command == ARG_CMD_ENCRYPT);
  if (!get_passphrase(&args, verify, passphrase, MAX_PASSPHRASE_SIZE)) {
    FATAL_ERROR("No passphrase provided");
  }
  if ((args.command == ARG_CMD_ENCRYPT) && (passphrase[0] == 0) &&
      !args.allow_empty_passphrase) {
    FATAL_ERROR("No passphrase provided");
  }

  // Get the 256-bit pepper (zero by default).
  uint8_t pepper[PEPPER_BYTES];
  memset(&pepper[0], 0, sizeof(pepper));
  if (args.pepper_file != NULL) {
    FILE* pepper_file = fopen(args.pepper_file, "rb");
    if (pepper_file == NULL) {
      FATAL_ERRORF("Unable to open pepper file: %s", args.pepper_file);
    }
    const bool pepper_success = derive_pepper_from_file(pepper_file, pepper);
    fclose(pepper_file);
    if (!pepper_success) {
      FATAL_ERRORF("Unable to read pepper file: %s", args.pepper_file);
    }
  }

  // If we're editing an encrypted file, we need a temporary file name for the
  // plaintext.
  if (args.command == ARG_CMD_EDIT) {
    // Select a file name for the temporarily decrypted file used for editing.
    char ext[100];
    get_edit_file_ext(args.infile, ext, sizeof(ext));
    make_safe_tmp_filename(s_edit_file_name, sizeof(s_edit_file_name), ext);

    // Also make sure that we have an editor.
    if (args.editor == NULL) {
      args.editor = get_default_editor();
    }
  }

  // Open the input file.
  if (args.infile != NULL) {
    s_inf = fopen(args.infile, "rb");
    if (s_inf == NULL) {
      FATAL_ERRORF("Unable to open input file: %s", args.infile);
    }
    s_inf_opened = true;
  } else {
    s_inf = stdin;
#if defined(_WIN32)
    _setmode(_fileno(s_inf), _O_BINARY);
#endif
  }

  // Open the output file.
  const char* outfile_name =
      (args.command == ARG_CMD_EDIT) ? s_edit_file_name : args.outfile;
  const char* path_to_delete = NULL;
  if (args.write_to_stdout) {
    // Write to standard output.
    s_outf = stdout;
#if defined(_WIN32)
    _setmode(_fileno(s_outf), _O_BINARY);
#endif
  } else if (args.command != ARG_CMD_TEST) {
    char* outfile_malloced = NULL;
    const char* outfile = NULL;
    if (outfile_name == NULL) {
      // Generate an output file name from the input file name.
      if (args.infile == NULL) {
        FATAL_ERROR("No input file name provided");
      }
      outfile_malloced =
          make_outfile_name(args.infile, args.command == ARG_CMD_ENCRYPT);
      if (outfile_malloced == NULL) {
        FATAL_ERROR("No reasonable output file name possible");
      }
      outfile = outfile_malloced;
      if (!args.keep) {
        path_to_delete = args.infile;
      }
    } else {
      // Use the user specified output file name.
      outfile = outfile_name;
    }

    // We will not overwrite an existing file without the "force" flag.
    if ((!args.force) && file_or_dir_exists(outfile)) {
      FATAL_ERRORF("File already exists: %s", outfile);
    }

    // Open the output file.
    s_outf = create_private_file(outfile);
    if (s_outf == NULL) {
      FATAL_ERRORF("Unable to open output file: %s", outfile);
    }
    s_outf_opened = true;
    if (outfile != NULL && outfile != s_edit_file_name) {
      set_unfinished_output_file_name(outfile);
    }
    if (outfile_malloced != NULL) {
      free(outfile_malloced);
    }
  }

  // Encrypt or decrypt.
  int cost = 1;
  bool success = false;
  switch (args.command) {
    case ARG_CMD_ENCRYPT:
      success = encrypt_stream(passphrase, pepper, s_inf, s_outf, args.cost);
      break;
    case ARG_CMD_DECRYPT:
    case ARG_CMD_EDIT:
    case ARG_CMD_TEST:
      // TODO: Delete output file (if applicable) on FATAL_ERROR and !success.
      success = decrypt_stream(passphrase, pepper, s_inf, s_outf, &cost);
      break;
    default:
      // Should never happen.
      FATAL_ERROR("Unknown command");
  }

  if (success) {
    // The output file is no longer considered unfinished.
    clear_unfinished_output_file_name();
  }

  // Close files.
  close_open_files();

  if (success && path_to_delete != NULL) {
    if (!delete_file(path_to_delete)) {
      fprintf(stderr, "***Warning: Unable to delete: %s\n", path_to_delete);
    }
  }

  // If this was a test, print the success.
  if (args.command == ARG_CMD_TEST && success) {
    const char* name = (args.infile == NULL) ? "stdin" : args.infile;
    fprintf(stderr, "%s: OK\n", name);
  }

  // Are we done?
  if (args.command != ARG_CMD_EDIT || !success) {
    exit(success ? 0 : 1);
  }

  //----------------------------------------------------------------------------
  // If we're editing an encrypted file we have some more work to do...
  //----------------------------------------------------------------------------

  {
    // Get the signature of the edit file before it is edited.
    uint8_t hash_before[SHA256_HASH_BYTES];
    compute_file_sha256(s_edit_file_name, hash_before);

    // Build the editor command string.
    char cmd[2048];
    const int cnt =
        snprintf(cmd, sizeof(cmd), "%s \"%s\"", args.editor, s_edit_file_name);
    if (cnt < 0 || ((size_t)cnt) >= sizeof(cmd)) {
      fprintf(stderr, "Editor command too long. Aborting.\n");
      exit(1);
    }

    // Run the editor.
    const int exit_code = run_command(cmd);
    if (exit_code != 0) {
      fprintf(stderr, "Editor terminated with an error. Aborting.\n");
      exit(1);
    }

    // Get the signature of the edit file after it is edited.
    uint8_t hash_after[SHA256_HASH_BYTES];
    compute_file_sha256(s_edit_file_name, hash_after);
    const int file_changed =
        (memcmp(hash_before, hash_after, sizeof(hash_before)) != 0);

    // Abort if the file was not changed.
    if (!file_changed) {
      fprintf(stderr,
              "There were no edits. Leaving the original file untouched.\n");
      exit(1);
    }
  }

  // Open the edited file for input.
  s_inf = fopen(s_edit_file_name, "rb");
  if (s_inf == NULL) {
    FATAL_ERRORF("Unable to open edited file: %s", s_edit_file_name);
  }
  s_inf_opened = true;

  // Open the output file.
  {
    // If the outfile is not specified, write back to the input file.
    outfile_name = args.outfile ? args.outfile : args.infile;

    // TODO: Derive all the keys BEFORE creating the output file, in order to
    // minimize the risk of ending up with an empty/corrupt file if the process
    // is killed prematurely.
    s_outf = create_private_file(outfile_name);
    if (s_outf == NULL) {
      FATAL_ERRORF("Unable to open output file: %s", outfile_name);
    }
    s_outf_opened = true;
  }

  // Encrypt the edited file.
  success = encrypt_stream(passphrase, pepper, s_inf, s_outf, cost);

  // Perform program cleanup.
  program_cleanup();

  exit(success ? 0 : 1);
}
