// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "arguments.h"

#include "error.h"
#include "version.h"
#include <stdio.h>
#include <string.h>

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

static const char HELP_SHORT_STR[] =
    "COMMANDS\n"
    "  -e, --encrypt       encrypt%s\n"
    "  -d, --decrypt       decrypt%s\n"
    "  -s, --show          decrypt file to stdout (same as -d -c)\n"
    "      --view          (same as --show)\n"
    "      --edit[=EDITOR] edit an encrypted file\n"
    "  -t, --test          test an encrypted file (no output)\n"
    "  -m, --make-pepper   make a pepper file (requires -o)\n"
    "  -V, --version       show the program version and exit\n"
    "  -h, --help          show program help and exit\n"
    "\n"
    "OPTIONS\n"
    "  -c, --stdout        write output to stdout\n"
    "  -o OUTFILE, --output OUTFILE\n"
    "                      write output to OUTFILE\n"
    "  -k, --keep          keep (don't delete) input file (default)\n"
    "  -n, --no-keep       don't keep (delete) input file\n"
    "  -f, --force         force overwrite of existing file\n"
    "  -E PASSENV, --pass-env PASSENV\n"
    "                      get passphrase from $PASSENV\n"
    "  -F PASSFILE, --pass-file PASSFILE\n"
    "                      get passphrase from the file PASSFILE\n"
    "  -P PEPPERFILE, --pepper-file PEPPERFILE\n"
    "                      derive and use a pepper from the file PEPPERFILE\n"
    "  -N, --no-gui        don't use a GUI passphrase dialog\n"
    "      --allow-empty   allow the passphrase to be empty\n"
    "  -1 .. -5            KDF computational cost (default: 2)\n";

static const char HELP_EXTRA_ENCRYPTION_STR[] =
    "\n"
    "ENCRYPTION & DECRYPTION\n"
    "  The default behaviour is that when a file is encrypted, the encrypted\n"
    "  file is called the same as the plaintext file but with the extension\n"
    "  \".z\". When a file is decrypted, the default behavior is that the\n"
    "  plaintext file is called the same as the encrypted file but without\n"
    "  the \".z\" extension.\n"
    "\n"
    "  If FILE is not specified, or if FILE is \"-\", the standard input is\n"
    "  encrypted or decrypted to the standard output. Note that in this case\n"
    "  the passphrase must be supplied via the -E/-F option.\n";

static const char HELP_EXTRA_PASSPHRASE_STR[] =
    "\n"
    "PASSPHRASE & PEPPER\n"
    "  The passphrase can be provided in one of the following ways:\n"
    "\n"
    "    Interactively via a terminal prompt or a GUI prompt (default)\n"
    "    Via a named environment variable (-E)\n"
    "    Via a named text file (-F)\n"
    "\n"
    "  Please note that when the passphrase is given in a file, only the\n"
    "  first line is used (excluding any terminating line ending), and null\n"
    "  characters are not supported.\n"
    "\n"
    "  For added security it is possible to provide a pepper in the form of\n"
    "  a file using the --pepper-file (-P) option. Using a pepper file is\n"
    "  roughly equivalent to extending the passphrase with 40 random\n"
    "  characters, but also improves security by dividing the secret into\n"
    "  two physically different parts (a passphrase and a pepper file).\n"
    "\n"
    "  The pepper is the SHA2-256 hash of the file, which means that the\n"
    "  pepper file can be of any kind and any length. Pepper files should be\n"
    "  kept secret and it is strongly recommended that they are stored\n"
    "  privately with restrictive permissions (e.g. chmod 400 mypepper.bin).\n"
    "\n"
    "  To securely generate a strong pepper file:\n"
    "\n"
    "    %s --make-pepper -o mypepper.bin\n"
    "\n"
    "  While the passphrase is mandatory, the pepper is optional.\n";

static const char HELP_EXTRA_EDITING_STR[] =
    "\n"
    "EDITING FILES\n"
    "  Use the --edit command to edit an encrypted file. The encrypted file\n"
    "  is decrypted to a temporary file, and an editor is launched. When the\n"
    "  editor program exits the edited plaintext file is re-encrypted and\n"
    "  the temporary file is deleted.\n"
    "\n"
    "  It is possible to use editors for non-text file types too by passing\n"
    "  the editor name together with the --edit command. E.g. to edit a PNG\n"
    "  image with GIMP, use:\n"
    "\n"
    "    %s --edit=gimp myimage.png.z\n"
    "\n"
    "  Note that it is possible to pass arguments to the editor:\n"
    "\n"
    "    %s --edit=\"gedit --standalone\" mysecrets.txt.z\n"
    "\n"
    "  Warning: Be aware that there may be additional vulnerabilities when\n"
    "  storing the plaintext in the file system and passing it to\n"
    "  potentially insecure and/or compromised editor programs.\n"
    "\n";

static const char HELP_EXTRA_CIPHERS_STR[] =
    "CIPHERS\n"
    "  The plaintext is encrypted in four levels using the following ciphers\n"
    "  (outermost level to innermost level):\n"
    "\n"
    "    Level 1: AES-256 (CBC)\n"
    "    Level 2: Twofish-256 (CBC)\n"
    "    Level 3: Serpent-256 (CBC)\n"
    "    Level 4: XChaCha20-256\n"
    "\n"
    "  The total key size is thus 256 x 4 = 1024 bits. For all intents and\n"
    "  purposes that key size is unbreakable for all eternity, so remember\n"
    "  that the entropy and management of your passphrases as well as the\n"
    "  handling of plaintexts are the deciding factors for keeping your data\n"
    "  secret.\n"
    "\n"
    "  The encrypted data is furthermore authenticated using the Poly1305\n"
    "  keyed hash function, which protects the encrypted data against\n"
    "  tampering and fraud.\n"
    "\n";

static const char HELP_EXTRA_KDF_STR[] =
    "KDF COMPUTATIONAL COST\n"
    "  The KDF computational cost controls how much time is required to\n"
    "  derive the cipher keys from the passphrase (a higher cost means that\n"
    "  it will take longer to crack the passphrase). The five cost levels\n"
    "  are as follows:\n"
    "\n"
    "    -1             Very quick         (1 ZKDF pass/cipher)\n"
    "    -2 (default)   20x slower than -1 (20 ZKDF passes/cipher)\n"
    "    -3             20x slower than -2 (400 ZKDF passes/cipher)\n"
    "    -4             20x slower than -3 (8,000 ZKDF passes/cipher)\n"
    "    -5             Unbelievably slow  (160,000 ZKDF passes/cipher)\n"
    "\n"
    "  With the default cost level, -2, the four cipher keys are derived in\n"
    "  less than a second on most computers from anno 2020 and later.\n"
    "\n"
    "  Increasing the KDF computational cost level by 1 (e.g. going from -2\n"
    "  to -3) is roughly equivalent to adding 4 bits of entropy to your\n"
    "  passphrase. Note that adding an extra random character to your\n"
    "  passphrase will add roughly 6 bits of entropy (i.e. it would take\n"
    "  about 60x longer time to brute-force guess the passphrase).\n";

enum help_type_t { HELP_TYPE_SHORT = 1, HELP_TYPE_LONG = 2 };

static void print_usage(const args_t* args, enum help_type_t help_type) {
  fprintf(stdout, "Usage: %s [COMMAND] [OPTION]... [FILE]\n\n", args->program);
  fprintf(stdout,
          HELP_SHORT_STR,
          args->is_ezdecrypt ? "" : " (default)",
          args->is_ezdecrypt ? " (default)" : "");
  if (help_type == HELP_TYPE_LONG) {
    fprintf(stdout, HELP_EXTRA_ENCRYPTION_STR);
    fprintf(stdout, HELP_EXTRA_PASSPHRASE_STR, args->program);
    fprintf(stdout, HELP_EXTRA_EDITING_STR, args->program, args->program);
    fprintf(stdout, HELP_EXTRA_CIPHERS_STR);
    fprintf(stdout, HELP_EXTRA_KDF_STR);
  } else {
    fprintf(stdout, "\nTry `%s --help' for more information.\n", args->program);
  }
}

static void print_version(void) {
  fprintf(stdout, PROGRAM_NAME " v" VERSION_STRING "\n");
  fprintf(stdout, PUBLIC_DOMAIN_STRING "\n");
  fprintf(stdout, URL_STRING "\n");
}

static bool invoked_as_ezdecrypt(const char* program_name) {
  if (program_name == NULL) {
    return false;
  }

  // Get the basename (drop path).
  const char* basename = strrchr(program_name, '/');
#if defined(_WIN32)
  if (basename == NULL) {
    basename = strrchr(program_name, '\\');
  }
#endif
  if (basename == NULL) {
    basename = program_name;
  }

  // Does the program name contain "ezdecrypt"?
  if (strstr(basename, "ezdecrypt") != NULL) {
    return true;
  }

  return false;
}

static bool validate_args(args_t* args) {
  if (args->command == ARG_CMD_MAKE_PEPPER) {
    // Ensure that outfile is specified.
    if (args->outfile == NULL) {
      fprintf(stderr,
              "%s: Missing output file (-o) (required when making a pepper).\n",
              args->program);
      print_usage(args, HELP_TYPE_SHORT);
      return false;
    }

    // There should be no input file.
    if (args->infile != NULL) {
      fprintf(stderr,
              "%s: Do not specify an input FILE when making a pepper.\n",
              args->program);
      print_usage(args, HELP_TYPE_SHORT);
      return false;
    }

    // That's all we need to check when making a pepper (the rest of the
    // argument checks are for other commands).
    return true;
  }

  // Special case: infile == "-" means stdin (i.e. set infile to NULL).
  if (args->infile != NULL && strcmp(args->infile, "-") == 0) {
    args->infile = NULL;
  }

  // If we're editing, the input file must be specified.
  if (args->command == ARG_CMD_EDIT && args->infile == NULL) {
    fprintf(
        stderr, "%s: Missing FILE (required when editing).\n", args->program);
    print_usage(args, HELP_TYPE_SHORT);
    return false;
  }

  // If we're using stdin for the input, we can't enter the passphrase via
  // stdin. Thus -E/-F must be used.
  if (args->infile == NULL && args->passphrase_env == NULL &&
      args->passphrase_file == NULL) {
    fprintf(stderr, "%s: Missing FILE or -E/-F option.\n", args->program);
    print_usage(args, HELP_TYPE_SHORT);
    return false;
  }

  // If we're using pipes (i.e. outfile is neither specified explicitly nor
  // implicitly from infile), we need to write to stdout.
  if (args->infile == NULL && args->outfile == NULL &&
      args->command != ARG_CMD_TEST) {
    args->write_to_stdout = true;
  }

  if (args->command == ARG_CMD_TEST) {
    // If we're testing, there should be no output file.
    if (args->outfile != NULL || args->write_to_stdout || args->force) {
      fprintf(
          stderr, "%s: Testing does not produce any output.\n", args->program);
      print_usage(args, HELP_TYPE_SHORT);
      return false;
    }

    // Don't allow --no-keep if we're testing.
    if (!args->keep) {
      fprintf(stderr,
              "%s: Source file can not be deleted as part of testing.\n",
              args->program);
      print_usage(args, HELP_TYPE_SHORT);
      return false;
    }
  }

  return true;
}

static bool starts_with(const char* str, const char* start_str) {
  return strncmp(start_str, str, strlen(start_str)) == 0;
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

bool parse_args(const int argc, char** argv, args_t* args) {
  PRECONDITION(argv != NULL);
  PRECONDITION(args != NULL);

  // Are we invoked for encryption or decryption?
  const bool is_ezdecrypt = invoked_as_ezdecrypt(argv[0]);

  // Set default arguments.
  memset(args, 0, sizeof(args_t));
  args->is_ezdecrypt = is_ezdecrypt;
  args->program = argv[0];
  args->command = is_ezdecrypt ? ARG_CMD_DECRYPT : ARG_CMD_ENCRYPT;
  args->cost = 2;
  args->keep = true;

  for (int k = 1; k < argc; ++k) {
    const char* arg = argv[k];
    if (strcmp(arg, "-h") == 0 || strcmp(arg, "--help") == 0) {
      args->command = ARG_CMD_HELP;
    } else if (strcmp(arg, "-e") == 0 || strcmp(arg, "--encrypt") == 0) {
      args->command = ARG_CMD_ENCRYPT;
    } else if (strcmp(arg, "-d") == 0 || strcmp(arg, "--decrypt") == 0) {
      args->command = ARG_CMD_DECRYPT;
    } else if (strcmp(arg, "-s") == 0 || strcmp(arg, "--show") == 0 ||
               strcmp(arg, "--view") == 0) {
      args->command = ARG_CMD_DECRYPT;
      args->write_to_stdout = true;
      args->outfile = NULL;
    } else if (strcmp(arg, "--edit") == 0 || starts_with(arg, "--edit=")) {
      args->command = ARG_CMD_EDIT;
      if (starts_with(arg, "--edit=")) {
        args->editor = strchr(arg, '=') + 1;
        if (args->editor[0] == 0) {
          fprintf(stderr, "%s: Missing EDITOR specification.\n", args->program);
          print_usage(args, HELP_TYPE_SHORT);
          return false;
        }
      } else {
        args->editor = NULL;
      }
      args->outfile = NULL;
    } else if (strcmp(arg, "-m") == 0 || strcmp(arg, "--make-pepper") == 0) {
      args->command = ARG_CMD_MAKE_PEPPER;
    } else if (strcmp(arg, "-t") == 0 || strcmp(arg, "--test") == 0) {
      args->command = ARG_CMD_TEST;
    } else if (strcmp(arg, "-V") == 0 || strcmp(arg, "--version") == 0) {
      args->command = ARG_CMD_VERSION;
    } else if (strcmp(arg, "-c") == 0 || strcmp(arg, "--stdout") == 0) {
      args->write_to_stdout = true;
      args->outfile = NULL;
    } else if (strcmp(arg, "-o") == 0 || strcmp(arg, "--output") == 0) {
      if (k >= (argc - 1)) {
        fprintf(stderr, "%s: Missing argument after %s.\n", args->program, arg);
        print_usage(args, HELP_TYPE_SHORT);
        return false;
      }
      args->outfile = argv[k + 1];
      args->write_to_stdout = false;
      ++k;
    } else if (strcmp(arg, "-k") == 0 || strcmp(arg, "--keep") == 0) {
      args->keep = true;
    } else if (strcmp(arg, "-n") == 0 || strcmp(arg, "--no-keep") == 0) {
      args->keep = false;
    } else if (strcmp(arg, "-f") == 0 || strcmp(arg, "--force") == 0) {
      args->force = true;
    } else if (strcmp(arg, "-E") == 0 || strcmp(arg, "--pass-env") == 0) {
      if (k >= (argc - 1)) {
        fprintf(stderr, "%s: Missing argument after %s.\n", args->program, arg);
        print_usage(args, HELP_TYPE_SHORT);
        return false;
      }
      args->passphrase_env = argv[k + 1];
      args->passphrase_file = NULL;
      ++k;
    } else if (strcmp(arg, "-F") == 0 || strcmp(arg, "--pass-file") == 0) {
      if (k >= (argc - 1)) {
        fprintf(stderr, "%s: Missing argument after %s.\n", args->program, arg);
        print_usage(args, HELP_TYPE_SHORT);
        return false;
      }
      args->passphrase_file = argv[k + 1];
      args->passphrase_env = NULL;
      ++k;
    } else if (strcmp(arg, "-P") == 0 || strcmp(arg, "--pepper-file") == 0) {
      if (k >= (argc - 1)) {
        fprintf(stderr, "%s: Missing argument after %s.\n", args->program, arg);
        print_usage(args, HELP_TYPE_SHORT);
        return false;
      }
      args->pepper_file = argv[k + 1];
      ++k;
    } else if (strcmp(arg, "-N") == 0 || strcmp(arg, "--no-gui") == 0) {
      args->disallow_gui = true;
    } else if (strcmp(arg, "--allow-empty") == 0) {
      args->allow_empty_passphrase = true;
    } else if (strcmp(arg, "-1") == 0) {
      args->cost = 1;
    } else if (strcmp(arg, "-2") == 0) {
      args->cost = 2;
    } else if (strcmp(arg, "-3") == 0) {
      args->cost = 3;
    } else if (strcmp(arg, "-4") == 0) {
      args->cost = 4;
    } else if (strcmp(arg, "-5") == 0) {
      args->cost = 5;
    } else if (arg[0] != '-' || strcmp(arg, "-") == 0) {
      if (args->infile != NULL) {
        print_usage(args, HELP_TYPE_SHORT);
        return false;
      }
      args->infile = arg;
    } else {
      fprintf(stderr, "%s: Unknown argument: %s\n", args->program, arg);
      print_usage(args, HELP_TYPE_SHORT);
      return false;
    }
  }

  if (args->command == ARG_CMD_HELP) {
    print_usage(args, HELP_TYPE_LONG);
    return false;
  }
  if (args->command == ARG_CMD_VERSION) {
    print_version();
    return false;
  }

  return validate_args(args);
}
