// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "crypt_stream.h"

#include "crypto/aes.h"
#include "crypto/chacha.h"
#include "crypto/entropy.h"
#include "crypto/poly1305.h"
#include "crypto/serpent.h"
#include "crypto/twofish.h"
#include "crypto/zkdf.h"
#include "error.h"
#include <assert.h>
#include <string.h>

//------------------------------------------------------------------------------
// Constants and macros.
//------------------------------------------------------------------------------

// The block length (in bytes) must be a multiple of the AES, Twofish & Serpent
// block lengths (16 bytes), and the ChaCha block length (64 bytes).
#define BLOCK_BYTES 64

// We encode some meta data (computational cost) in the un-encrypted AES salt
// area.
#define METADATA_COST 0
#define METADATA_LEN (4)

// Total salt bytes for the different ciphers.
// Note: We round up to a multiple of 64 in order for the ciphers to be able to
// handle the salts without padding. Trailing bytes are ignored.
#define ROUNDUP64(x) ((((x) + 63) / 64) * 64)
#define AES_HEADER_BYTES \
  ROUNDUP64(ZKDF_SALT_BYTES + AES_IV_BYTES + METADATA_LEN)
#define TWOFISH_HEADER_BYTES ROUNDUP64(ZKDF_SALT_BYTES + TWOFISH_IV_BYTES)
#define SERPENT_HEADER_BYTES ROUNDUP64(ZKDF_SALT_BYTES + SERPENT_IV_BYTES)
#define CHACHA_HEADER_BYTES ROUNDUP64(ZKDF_SALT_BYTES + XCHACHA_IV_BYTES)

// Computational cost index to number of ZKDF passes mapping.
static const int NUM_ZKDF_PASSES_FOR_COST[5] = {
    1,      // Very quick (mainly intended for testing).
    20,     // Fast:ish (for every day casual use).
    400,    // Slow (increased privacy).
    8000,   // Hands off!
    160000  // Paranoia extreme.
};

//------------------------------------------------------------------------------
// Helper functions.
//------------------------------------------------------------------------------

static int to_zkdf_passes(int computational_cost) {
  // Convert the computational cost index (1-5) to number of ZKDF passes.
  if (computational_cost < 1) {
    computational_cost = 1;
  } else if (computational_cost > 5) {
    computational_cost = 5;
  }
  return NUM_ZKDF_PASSES_FOR_COST[computational_cost - 1];
}

static inline uint32_t get_uint32_be(const uint8_t* bytes) {
  return (((uint64_t)bytes[0]) << 24) | (((uint64_t)bytes[1]) << 16) |
         (((uint64_t)bytes[2]) << 8) | (((uint64_t)bytes[3]) << 0);
}

static inline void set_uint32_be(const uint32_t x, uint8_t* bytes) {
  bytes[0] = (uint8_t)(x >> 24);
  bytes[1] = (uint8_t)(x >> 16);
  bytes[2] = (uint8_t)(x >> 8);
  bytes[3] = (uint8_t)(x >> 0);
}

static void encode32_mod(uint8_t* buffer,
                         const uint32_t value,
                         const uint32_t modulo) {
  PRECONDITION(value < modulo);

  // Read 4 bytes from the byte buffer as a 32-bit integer.
  const uint32_t x_original = get_uint32_be(buffer);

  // Ensure that X is divisible by MODULO, and that adding a value in
  // the range [0, MODULO) will not overflow.
  uint32_t x = x_original / modulo;
  x -= (x_original >> 31);
  x *= modulo;

  // Add the value to X.
  x += value;

  // Store the modified integer back into the buffer.
  set_uint32_be(x, buffer);
}

static uint32_t decode32_mod(const uint8_t* buffer, const uint32_t modulo) {
  const uint32_t x = get_uint32_be(buffer);
  return x % modulo;
}

static uint8_t* peppered(const uint8_t pepper[PEPPER_BYTES],
                         const uint8_t salt[ZKDF_SALT_BYTES]) {
  static_assert(PEPPER_BYTES == ZKDF_SALT_BYTES,
                "Pepper size must match salt size");

  static uint8_t peppered_salt[ZKDF_SALT_BYTES];

  // We pepper the salt by XOR:ing the salt and pepper together.
  for (unsigned i = 0; i < ZKDF_SALT_BYTES; ++i) {
    peppered_salt[i] = salt[i] ^ pepper[i];
  }
  return &peppered_salt[0];
}

void initialize_poly1305_context_from_chacha(poly1305_ctx_t* poly1305_ctx,
                                             chacha256_ctx_t* chacha_ctx) {
  static_assert(
      CHACHA_BLOCK_BYTES >= POLY1305_KEY_BYTES,
      "Poly1305 key size must be smaller than the ChaCha20 block size");

  // Encrypt a zero-filled buffer with ChaCha20. This assumes that the ChaCha20
  // context has just been initialized (i.e. its counter is zero). After this
  // operation, the ChaCha20 context will have the counter set to 1 (one).
  uint8_t buf[CHACHA_BLOCK_BYTES] = {0};
  chacha20_256_encrypt_buffer(chacha_ctx, buf, sizeof(buf));

  // Use the first 32 bytes of the encrypted buffer as the key for the Poly1305
  // hash context (the rest of the buffer is simply thrown away). This is the
  // same method that is used in the standardized ChaCha20-Poly1305
  // authenticated encryption algorithm.
  poly1305_init(poly1305_ctx, buf);
}

//------------------------------------------------------------------------------
// Encryption.
//------------------------------------------------------------------------------

bool encrypt_stream(const char* passphrase,
                    const uint8_t pepper[PEPPER_BYTES],
                    FILE* inf,
                    FILE* outf,
                    int computational_cost) {
  PRECONDITION(passphrase != NULL);
  PRECONDITION(inf != NULL);
  PRECONDITION(outf != NULL);

  bool success = false;

  // Generate a random AES salt.
  uint8_t salt_aes[AES_HEADER_BYTES];
  uint8_t* iv_aes = &salt_aes[ZKDF_SALT_BYTES];
  make_entropy(salt_aes, sizeof(salt_aes));

  // Encode the ZKDF computational cost into the AES salt area.
  uint8_t* metadata = &iv_aes[AES_IV_BYTES];
  encode32_mod(&metadata[METADATA_COST], computational_cost - 1, 5);

  // Get the number of ZKDF passes.
  const int zkdf_passes = to_zkdf_passes(computational_cost);

  // AES: Generate the key and initialize the context.
  uint8_t key_aes[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_aes[0]), zkdf_passes, key_aes);
  aes256_ctx_t aes_ctx;
  aes_256_init(&aes_ctx, key_aes, iv_aes);

  // Generate a random Twofish salt.
  uint8_t salt_twofish[TWOFISH_HEADER_BYTES];
  const uint8_t* iv_twofish = &salt_twofish[ZKDF_SALT_BYTES];
  make_entropy(salt_twofish, sizeof(salt_twofish));

  // Twofish: Generate the key and initialize the context.
  uint8_t key_twofish[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_twofish[0]), zkdf_passes, key_twofish);
  twofish256_ctx_t twofish_ctx;
  twofish_256_init(&twofish_ctx, key_twofish, iv_twofish);

  // Generate a random Serpent salt.
  uint8_t salt_serpent[SERPENT_HEADER_BYTES];
  const uint8_t* iv_serpent = &salt_serpent[ZKDF_SALT_BYTES];
  make_entropy(salt_serpent, sizeof(salt_serpent));

  // Serpent: Generate the key and initialize the context.
  uint8_t key_serpent[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_serpent[0]), zkdf_passes, key_serpent);
  serpent256_ctx_t serpent_ctx;
  serpent_256_init(&serpent_ctx, key_serpent, iv_serpent);

  // Generate a random ChaCha salt.
  uint8_t salt_chacha[CHACHA_HEADER_BYTES];
  const uint8_t* iv_chacha = &salt_chacha[ZKDF_SALT_BYTES];
  make_entropy(salt_chacha, sizeof(salt_chacha));

  // ChaCha: Generate the key and initialize the context.
  uint8_t key_chacha[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_chacha[0]), zkdf_passes, key_chacha);
  chacha256_ctx_t chacha_ctx;
  xchacha20_256_init(&chacha_ctx, key_chacha, iv_chacha);

  // Poly1305: Initialize the context. This must be done after initializing the
  // ChaCha20 context, but before using the ChaCha20 context for actual data
  // encryption.
  poly1305_ctx_t poly1305_ctx;
  initialize_poly1305_context_from_chacha(&poly1305_ctx, &chacha_ctx);

  // Write the AES salt to the destination file (in plaintext).
  if (fwrite(salt_aes, 1, sizeof(salt_aes), outf) != sizeof(salt_aes)) {
    FATAL_ERROR("Failed to write to output file");
  }

  // Encrypt and write the Twofish salt (encrypted by AES).
  aes_256_cbc_encrypt_buffer(&aes_ctx, salt_twofish, sizeof(salt_twofish));
  if (fwrite(salt_twofish, 1, sizeof(salt_twofish), outf) !=
      sizeof(salt_twofish)) {
    FATAL_ERROR("Failed to write to output file");
  }

  // Encrypt and write the Serpent salt (encrypted by AES+Twofish).
  twofish_256_cbc_encrypt_buffer(
      &twofish_ctx, salt_serpent, sizeof(salt_serpent));
  aes_256_cbc_encrypt_buffer(&aes_ctx, salt_serpent, sizeof(salt_serpent));
  if (fwrite(salt_serpent, 1, sizeof(salt_serpent), outf) !=
      sizeof(salt_serpent)) {
    FATAL_ERROR("Failed to write to output file");
  }

  // Encrypt and write the ChaCha salt (encrypted by AES+Twofish+Serpent).
  serpent_256_cbc_encrypt_buffer(
      &serpent_ctx, salt_chacha, sizeof(salt_chacha));
  twofish_256_cbc_encrypt_buffer(
      &twofish_ctx, salt_chacha, sizeof(salt_chacha));
  aes_256_cbc_encrypt_buffer(&aes_ctx, salt_chacha, sizeof(salt_chacha));
  if (fwrite(salt_chacha, 1, sizeof(salt_chacha), outf) !=
      sizeof(salt_chacha)) {
    FATAL_ERROR("Failed to write to output file");
  }

  // Include the file header (salts) in the Poly1305 hash (it's associated
  // data).
  poly1305_append(&poly1305_ctx, salt_aes, sizeof(salt_aes));
  poly1305_append(&poly1305_ctx, salt_twofish, sizeof(salt_twofish));
  poly1305_append(&poly1305_ctx, salt_serpent, sizeof(salt_serpent));
  poly1305_append(&poly1305_ctx, salt_chacha, sizeof(salt_chacha));

  // Read + encrypt + write.
  uint8_t data[BLOCK_BYTES];
  while (true) {
    // Read one block from the source file.
    size_t bytes_read = fread(data, 1, BLOCK_BYTES, inf);

    // Pad missing bytes with a random sequence and terminate with a byte that
    // encodes the number of pad bytes.
    if (bytes_read < BLOCK_BYTES) {
      const uint8_t num_pad_bytes = (uint8_t)BLOCK_BYTES - (uint8_t)bytes_read;
      make_entropy(&data[bytes_read], num_pad_bytes);
      static const uint8_t PAD_COUNT_MASK = (BLOCK_BYTES * 2) - 1;
      data[BLOCK_BYTES - 1] =
          (data[BLOCK_BYTES - 1] & ~PAD_COUNT_MASK) | num_pad_bytes;
    }

    // Encrypt the data buffer.
    chacha20_256_encrypt_buffer(&chacha_ctx, data, BLOCK_BYTES);
    serpent_256_cbc_encrypt_buffer(&serpent_ctx, data, BLOCK_BYTES);
    twofish_256_cbc_encrypt_buffer(&twofish_ctx, data, BLOCK_BYTES);
    aes_256_cbc_encrypt_buffer(&aes_ctx, data, BLOCK_BYTES);

    // Update the Poly1305 hash.
    poly1305_append(&poly1305_ctx, data, BLOCK_BYTES);

    // Write to the destination file.
    if (fwrite(data, 1, BLOCK_BYTES, outf) != BLOCK_BYTES) {
      FATAL_ERROR("Failed to write to output file");
    }

    // Done (EOF)?
    if (bytes_read < BLOCK_BYTES) {
      success = true;
      break;
    }
  }

  // Write a tail that is smaller than one cipher block size. This signals to
  // the decoder that this is the end of the file. Included in the tail are:
  //  * The Poly1305 authentication tag.
  //  * A random number of random bytes (to be ignored by the decoder).
  make_entropy(data, BLOCK_BYTES);
  const size_t tail_size =
      POLY1305_HASH_BYTES + (((size_t)data[BLOCK_BYTES - 1]) %
                             (BLOCK_BYTES - POLY1305_HASH_BYTES - 1));
  poly1305_finalize(&poly1305_ctx, &data[0]);
  if (fwrite(data, 1, tail_size, outf) != tail_size) {
    FATAL_ERROR("Failed to write to output file");
  }

  return success;
}

//------------------------------------------------------------------------------
// Decryption.
//------------------------------------------------------------------------------

bool decrypt_stream(const char* passphrase,
                    const uint8_t pepper[PEPPER_BYTES],
                    FILE* inf,
                    FILE* outf,
                    int* computational_cost) {
  PRECONDITION(passphrase != NULL);
  PRECONDITION(inf != NULL);
  PRECONDITION(computational_cost != NULL);

  bool success = false;

  // Read the raw (encrypted) header data.
  uint8_t raw_header[AES_HEADER_BYTES + TWOFISH_HEADER_BYTES +
                     SERPENT_HEADER_BYTES + CHACHA_HEADER_BYTES];
  if (fread(raw_header, 1, sizeof(raw_header), inf) != sizeof(raw_header)) {
    FATAL_ERROR("Premature end of input file");
  }

  // Get the AES salt (stored as plaintext).
  uint8_t salt_aes[AES_HEADER_BYTES];
  const uint8_t* iv_aes = &salt_aes[ZKDF_SALT_BYTES];
  memcpy(&salt_aes[0], &raw_header[0], sizeof(salt_aes));

  // Decode the computational cost (we also return the cost to the caller).
  const uint8_t* metadata = &iv_aes[AES_IV_BYTES];
  const int cost = ((int)decode32_mod(&metadata[METADATA_COST], 5)) + 1;
  *computational_cost = cost;

  // Get the number of ZKDF passes.
  const int zkdf_passes = to_zkdf_passes(cost);

  // AES: Generate the key and initialize the context.
  uint8_t key_aes[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_aes[0]), zkdf_passes, key_aes);
  aes256_ctx_t aes_ctx;
  aes_256_init(&aes_ctx, key_aes, iv_aes);

  // Get and decrypt the Twofish salt (encrypted by AES).
  uint8_t salt_twofish[TWOFISH_HEADER_BYTES];
  const uint8_t* iv_twofish = &salt_twofish[ZKDF_SALT_BYTES];
  memcpy(&salt_twofish[0], &raw_header[AES_HEADER_BYTES], sizeof(salt_twofish));
  aes_256_cbc_decrypt_buffer(&aes_ctx, salt_twofish, sizeof(salt_twofish));

  // Twofish: Generate the key and initialize the context.
  uint8_t key_twofish[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_twofish[0]), zkdf_passes, key_twofish);
  twofish256_ctx_t twofish_ctx;
  twofish_256_init(&twofish_ctx, key_twofish, iv_twofish);

  // Get and decrypt the Serpent salt (encrypted by AES+Twofish).
  uint8_t salt_serpent[SERPENT_HEADER_BYTES];
  const uint8_t* iv_serpent = &salt_serpent[ZKDF_SALT_BYTES];
  memcpy(&salt_serpent[0],
         &raw_header[AES_HEADER_BYTES + TWOFISH_HEADER_BYTES],
         sizeof(salt_serpent));
  aes_256_cbc_decrypt_buffer(&aes_ctx, salt_serpent, sizeof(salt_serpent));
  twofish_256_cbc_decrypt_buffer(
      &twofish_ctx, salt_serpent, sizeof(salt_serpent));

  // Serpent: Generate the key and initialize the context.
  uint8_t key_serpent[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_serpent[0]), zkdf_passes, key_serpent);
  serpent256_ctx_t serpent_ctx;
  serpent_256_init(&serpent_ctx, key_serpent, iv_serpent);

  // Get and decrypt the ChaCha salt (encrypted by AES+Twofish+Serpent).
  uint8_t salt_chacha[CHACHA_HEADER_BYTES];
  const uint8_t* iv_chacha = &salt_chacha[ZKDF_SALT_BYTES];
  memcpy(&salt_chacha[0],
         &raw_header[AES_HEADER_BYTES + TWOFISH_HEADER_BYTES +
                     SERPENT_HEADER_BYTES],
         sizeof(salt_chacha));
  aes_256_cbc_decrypt_buffer(&aes_ctx, salt_chacha, sizeof(salt_chacha));
  twofish_256_cbc_decrypt_buffer(
      &twofish_ctx, salt_chacha, sizeof(salt_chacha));
  serpent_256_cbc_decrypt_buffer(
      &serpent_ctx, salt_chacha, sizeof(salt_chacha));

  // ChaCha: Generate the key and initialize the context.
  uint8_t key_chacha[ZKDF_KEY_BYTES];
  zkdf_make_key(
      passphrase, peppered(pepper, &salt_chacha[0]), zkdf_passes, key_chacha);
  chacha256_ctx_t chacha_ctx;
  xchacha20_256_init(&chacha_ctx, key_chacha, iv_chacha);

  // Poly1305: Initialize the context. This must be done after initializing the
  // ChaCha20 context, but before using the ChaCha20 context for actual data
  // decryption.
  poly1305_ctx_t poly1305_ctx;
  initialize_poly1305_context_from_chacha(&poly1305_ctx, &chacha_ctx);

  // Include the raw file header (salts) in the Poly1305 hash.
  poly1305_append(&poly1305_ctx, raw_header, sizeof(raw_header));

  // Read + decrypt + write.
  uint8_t data_double_buffered[2][BLOCK_BYTES];
  uint8_t* data = data_double_buffered[0];
  uint8_t* data_next = data_double_buffered[1];
  size_t bytes_read = fread(data_next, 1, BLOCK_BYTES, inf);
  while (true) {
    // Done (EOF or trailing garbage)?
    if (bytes_read < BLOCK_BYTES) {
      // Check the Poly1305 authentication tag.
      uint8_t hash_poly1305[POLY1305_HASH_BYTES];
      poly1305_finalize(&poly1305_ctx, hash_poly1305);
      if (((bytes_read < POLY1305_HASH_BYTES) ||
           (memcmp(hash_poly1305, data_next, POLY1305_HASH_BYTES) != 0))) {
        FATAL_ERROR(
            "Authentication failed. Invalid credentials or corrupt stream?");
      }
      success = true;
      break;
    }

    // Swap data buffers (the previously loaded buffer becomes the current).
    {
      uint8_t* tmp = data_next;
      data_next = data;
      data = tmp;
    }

    // Update the Poly1305 hash.
    poly1305_append(&poly1305_ctx, data, BLOCK_BYTES);

    // Read the next block from the source file.
    // Note: We need to read one block ahead in order to detect EOF, i.e. if the
    // current block was the last block or not.
    size_t next_bytes_read = fread(data_next, 1, BLOCK_BYTES, inf);

    // We only have to decrypt the data if we're writing to an output file
    // (otherwise we're just checking the authentication tag).
    if (outf != NULL) {
      // Decrypt the data buffer.
      aes_256_cbc_decrypt_buffer(&aes_ctx, data, BLOCK_BYTES);
      twofish_256_cbc_decrypt_buffer(&twofish_ctx, data, BLOCK_BYTES);
      serpent_256_cbc_decrypt_buffer(&serpent_ctx, data, BLOCK_BYTES);
      chacha20_256_decrypt_buffer(&chacha_ctx, data, BLOCK_BYTES);

      // If this was the last block, handle padding.
      size_t actual_len = BLOCK_BYTES;
      if (next_bytes_read < BLOCK_BYTES) {
        // The padding length is encoded in the last padding byte.
        static const size_t PAD_COUNT_MASK = (BLOCK_BYTES * 2) - 1;
        size_t padded_bytes = ((size_t)data[BLOCK_BYTES - 1]) & PAD_COUNT_MASK;
        if (padded_bytes <= BLOCK_BYTES) {
          actual_len = BLOCK_BYTES - padded_bytes;
        }
      }

      // Write to the destination file.
      if (actual_len > 0U) {
        if (fwrite(data, 1, actual_len, outf) != actual_len) {
          FATAL_ERROR("Failed to write to output file");
        }
      }
    }

    bytes_read = next_bytes_read;
  }

  return success;
}
