// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "aes.h"

#include "../error.h"
#include <string.h>

// Can we use x86 intrinsics?
#if defined(__x86_64__) || defined(_M_AMD64) || defined(_M_X64) || \
    defined(__amd64__) || defined(__amd64)
#include "x86cpuid.h"
#include <wmmintrin.h>
#define USE_X86_AES
#endif  // __x86_64__

// Can we use ARMv8 intrinsics? NEON is a prerequisite, but crypto is requested
// explicitly and checked at runtime (for platforms that support that).
#if defined(_M_ARM64) ||     \
    (defined(__aarch64__) && \
     (defined(__ARM_NEON) || defined(__ARM_NEON__) || defined(__ARM_NEON_FP)))
#include <arm_neon.h>
#if (defined(__GNUC__) && (__GNUC__ >= 6)) || \
    (defined(__clang__) && (__clang_major__ >= 17))
#define ATTRIBUTE_AARCH64_CRYPTO __attribute__((target("+crypto")))
#else
#define ATTRIBUTE_AARCH64_CRYPTO
#endif
#if defined(__linux__) || defined(__linux) || defined(linux)
#include <sys/auxv.h>
#define USE_GETAUXVAL
#endif
#define USE_AARCH64_AES
#endif  // __aarch64__

// The number of columns comprising a state in AES. This is a constant in AES.
#define NB 4

// The number of 32 bit words in a key.
#define NK 8

// The number of rounds in AES cipher.
#define NUM_ROUNDS 14

// Acceleration method.
#define METHOD_GENERIC 0
#define METHOD_X86_AES 1
#define METHOD_AARCH64_AES 2

// state_t - An array holding the intermediate results during decryption.
typedef uint8_t state_t[4][4];

//------------------------------------------------------------------------------
// Look-up tables.
//------------------------------------------------------------------------------

static const uint8_t SBOX[256] = {
    0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b,
    0xfe, 0xd7, 0xab, 0x76, 0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0,
    0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0, 0xb7, 0xfd, 0x93, 0x26,
    0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
    0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2,
    0xeb, 0x27, 0xb2, 0x75, 0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0,
    0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84, 0x53, 0xd1, 0x00, 0xed,
    0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
    0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f,
    0x50, 0x3c, 0x9f, 0xa8, 0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5,
    0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2, 0xcd, 0x0c, 0x13, 0xec,
    0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
    0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14,
    0xde, 0x5e, 0x0b, 0xdb, 0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c,
    0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79, 0xe7, 0xc8, 0x37, 0x6d,
    0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
    0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f,
    0x4b, 0xbd, 0x8b, 0x8a, 0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e,
    0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e, 0xe1, 0xf8, 0x98, 0x11,
    0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
    0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f,
    0xb0, 0x54, 0xbb, 0x16};

static const uint8_t RSBOX[256] = {
    0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e,
    0x81, 0xf3, 0xd7, 0xfb, 0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87,
    0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb, 0x54, 0x7b, 0x94, 0x32,
    0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
    0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49,
    0x6d, 0x8b, 0xd1, 0x25, 0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16,
    0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92, 0x6c, 0x70, 0x48, 0x50,
    0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
    0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05,
    0xb8, 0xb3, 0x45, 0x06, 0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02,
    0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b, 0x3a, 0x91, 0x11, 0x41,
    0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
    0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8,
    0x1c, 0x75, 0xdf, 0x6e, 0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89,
    0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b, 0xfc, 0x56, 0x3e, 0x4b,
    0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
    0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59,
    0x27, 0x80, 0xec, 0x5f, 0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d,
    0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef, 0xa0, 0xe0, 0x3b, 0x4d,
    0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
    0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63,
    0x55, 0x21, 0x0c, 0x7d};

// The round constant word array, RCON[i], contains the values given by
// x to the power (i-1) being powers of x (x is denoted as {02}) in the field
// GF(2^8)
static const uint8_t RCON[11] =
    {0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36};

// Jordan Goulder points out in PR #12
// (https://github.com/kokke/tiny-AES-C/pull/12), that you can remove most of
// the elements in the RCON array, because they are unused.
//
// From Wikipedia's article on the Rijndael key schedule @
// https://en.wikipedia.org/wiki/Rijndael_key_schedule#RCON
//
// "Only the first some of these constants are actually used – up to RCON[10]
// for AES-128 (as 11 round keys are needed), up to RCON[8] for AES-192, up to
// RCON[7] for AES-256. RCON[0] is not used in AES algorithm."

//------------------------------------------------------------------------------
// Private functions.
//------------------------------------------------------------------------------

// This function produces NB(NUM_ROUNDS+1) round keys. The round keys are used
// in each round to decrypt the states.
static void key_expansion(uint8_t round_key[AES_ROUND_KEY_BYTES],
                          const uint8_t key[AES_KEY_BYTES]) {
  // The first round key is the key itself.
  for (unsigned i = 0; i < NK; ++i) {
    round_key[(i * 4) + 0] = key[(i * 4) + 0];
    round_key[(i * 4) + 1] = key[(i * 4) + 1];
    round_key[(i * 4) + 2] = key[(i * 4) + 2];
    round_key[(i * 4) + 3] = key[(i * 4) + 3];
  }

  // All other round keys are found from the previous round keys.
  for (unsigned i = NK; i < NB * (NUM_ROUNDS + 1); ++i) {
    uint8_t tempa[4];
    {
      unsigned k = (i - 1) * 4;
      tempa[0] = round_key[k + 0];
      tempa[1] = round_key[k + 1];
      tempa[2] = round_key[k + 2];
      tempa[3] = round_key[k + 3];
    }

    if (i % NK == 0) {
      // This function shifts the 4 bytes in a word to the left once.
      // [a0,a1,a2,a3] becomes [a1,a2,a3,a0]

      // RotWord
      {
        const uint8_t u8tmp = tempa[0];
        tempa[0] = tempa[1];
        tempa[1] = tempa[2];
        tempa[2] = tempa[3];
        tempa[3] = u8tmp;
      }

      // SubWord
      {
        tempa[0] = SBOX[tempa[0]];
        tempa[1] = SBOX[tempa[1]];
        tempa[2] = SBOX[tempa[2]];
        tempa[3] = SBOX[tempa[3]];
      }

      tempa[0] = tempa[0] ^ RCON[i / NK];
    }
    if (i % NK == 4) {
      // SubWord
      {
        tempa[0] = SBOX[tempa[0]];
        tempa[1] = SBOX[tempa[1]];
        tempa[2] = SBOX[tempa[2]];
        tempa[3] = SBOX[tempa[3]];
      }
    }
    unsigned j = i * 4;
    unsigned k = (i - NK) * 4;
    round_key[j + 0] = round_key[k + 0] ^ tempa[0];
    round_key[j + 1] = round_key[k + 1] ^ tempa[1];
    round_key[j + 2] = round_key[k + 2] ^ tempa[2];
    round_key[j + 3] = round_key[k + 3] ^ tempa[3];
  }
}

// This function adds the round key to state.
// The round key is added to the state by an XOR function.
static void add_round_key(int round, state_t* state, const uint8_t* round_key) {
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      (*state)[i][j] ^= round_key[(round * NB * 4) + (i * NB) + j];
    }
  }
}

// The sub_bytes function substitutes the values in the state matrix with values
// in an S-box.
static void sub_bytes(state_t* state) {
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      (*state)[j][i] = SBOX[(*state)[j][i]];
    }
  }
}

// The shift_rows() function shifts the rows in the state to the left.
// Each row is shifted with different offset.
// Offset = Row number. So the first row is not shifted.
static void shift_rows(state_t* state) {
  // Rotate first row 1 columns to left
  uint8_t temp = (*state)[0][1];
  (*state)[0][1] = (*state)[1][1];
  (*state)[1][1] = (*state)[2][1];
  (*state)[2][1] = (*state)[3][1];
  (*state)[3][1] = temp;

  // Rotate second row 2 columns to left
  temp = (*state)[0][2];
  (*state)[0][2] = (*state)[2][2];
  (*state)[2][2] = temp;

  temp = (*state)[1][2];
  (*state)[1][2] = (*state)[3][2];
  (*state)[3][2] = temp;

  // Rotate third row 3 columns to left
  temp = (*state)[0][3];
  (*state)[0][3] = (*state)[3][3];
  (*state)[3][3] = (*state)[2][3];
  (*state)[2][3] = (*state)[1][3];
  (*state)[1][3] = temp;
}

static uint8_t xtime(uint8_t x) {
  return ((x << 1) ^ (((x >> 7) & 1) * 0x1b));
}

// mix_columns function mixes the columns of the state matrix
static void mix_columns(state_t* state) {
  for (int i = 0; i < 4; ++i) {
    const uint8_t t = (*state)[i][0];
    const uint8_t Tmp =
        (*state)[i][0] ^ (*state)[i][1] ^ (*state)[i][2] ^ (*state)[i][3];
    uint8_t Tm = (*state)[i][0] ^ (*state)[i][1];
    Tm = xtime(Tm);
    (*state)[i][0] ^= Tm ^ Tmp;
    Tm = (*state)[i][1] ^ (*state)[i][2];
    Tm = xtime(Tm);
    (*state)[i][1] ^= Tm ^ Tmp;
    Tm = (*state)[i][2] ^ (*state)[i][3];
    Tm = xtime(Tm);
    (*state)[i][2] ^= Tm ^ Tmp;
    Tm = (*state)[i][3] ^ t;
    Tm = xtime(Tm);
    (*state)[i][3] ^= Tm ^ Tmp;
  }
}

// Used to multiply numbers in the field GF(2^8)
// Note: The last call to xtime() is unneeded, but often ends up generating a
// smaller binary.
// The compiler seems to be able to vectorize the operation better this way.
// See: https://github.com/kokke/tiny-AES-c/pull/34
static uint8_t mult(uint8_t x, uint8_t y) {
  return (((y & 1) * x) ^ ((y >> 1 & 1) * xtime(x)) ^
          ((y >> 2 & 1) * xtime(xtime(x))) ^
          ((y >> 3 & 1) * xtime(xtime(xtime(x)))) ^
          ((y >> 4 & 1) *
           xtime(xtime(
               xtime(xtime(x))))));  // This last call to xtime() can be omitted
}

static void mix_columns_inv(state_t* state) {
  for (int i = 0; i < 4; ++i) {
    const uint8_t a = (*state)[i][0];
    const uint8_t b = (*state)[i][1];
    const uint8_t c = (*state)[i][2];
    const uint8_t d = (*state)[i][3];

    (*state)[i][0] =
        mult(a, 0x0e) ^ mult(b, 0x0b) ^ mult(c, 0x0d) ^ mult(d, 0x09);
    (*state)[i][1] =
        mult(a, 0x09) ^ mult(b, 0x0e) ^ mult(c, 0x0b) ^ mult(d, 0x0d);
    (*state)[i][2] =
        mult(a, 0x0d) ^ mult(b, 0x09) ^ mult(c, 0x0e) ^ mult(d, 0x0b);
    (*state)[i][3] =
        mult(a, 0x0b) ^ mult(b, 0x0d) ^ mult(c, 0x09) ^ mult(d, 0x0e);
  }
}

static void sub_bytes_inv(state_t* state) {
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      (*state)[j][i] = RSBOX[(*state)[j][i]];
    }
  }
}

static void shift_rows_inv(state_t* state) {
  // Rotate first row 1 columns to right
  uint8_t temp = (*state)[3][1];
  (*state)[3][1] = (*state)[2][1];
  (*state)[2][1] = (*state)[1][1];
  (*state)[1][1] = (*state)[0][1];
  (*state)[0][1] = temp;

  // Rotate second row 2 columns to right
  temp = (*state)[0][2];
  (*state)[0][2] = (*state)[2][2];
  (*state)[2][2] = temp;

  temp = (*state)[1][2];
  (*state)[1][2] = (*state)[3][2];
  (*state)[3][2] = temp;

  // Rotate third row 3 columns to right
  temp = (*state)[0][3];
  (*state)[0][3] = (*state)[1][3];
  (*state)[1][3] = (*state)[2][3];
  (*state)[2][3] = (*state)[3][3];
  (*state)[3][3] = temp;
}

static void encrypt_one_block(uint8_t* buf, const uint8_t* round_key) {
  state_t* state = (state_t*)buf;

  // Add the first round key to the state before starting the rounds.
  add_round_key(0, state, round_key);

  // Perform NUM_ROUNDS rounds.
  //  - The first NUM_ROUNDS-1 rounds are identical.
  //  - Last one without mix_columns().
  for (int round = 1;; ++round) {
    sub_bytes(state);
    shift_rows(state);
    if (round == NUM_ROUNDS) {
      break;
    }
    mix_columns(state);
    add_round_key(round, state, round_key);
  }

  // Add round key to last round
  add_round_key(NUM_ROUNDS, state, round_key);
}

#if defined(USE_X86_AES)
static void encrypt_one_block_x86(uint8_t* buf, const uint8_t* round_key) {
  __m128i s = _mm_loadu_si128((__m128i*)buf);
  s = _mm_xor_si128(s, _mm_loadu_si128((const __m128i*)&round_key[0]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[16]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[32]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[48]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[64]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[80]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[96]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[112]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[128]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[144]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[160]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[176]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[192]));
  s = _mm_aesenc_si128(s, _mm_loadu_si128((const __m128i*)&round_key[208]));
  s = _mm_aesenclast_si128(s, _mm_loadu_si128((const __m128i*)&round_key[224]));
  _mm_storeu_si128((__m128i*)buf, s);
}
#endif  // USE_X86_AES

#if defined(USE_AARCH64_AES)
ATTRIBUTE_AARCH64_CRYPTO
static void encrypt_one_block_aarch64(uint8_t* buf, const uint8_t* round_key) {
  uint8x16_t s = vld1q_u8(buf);
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[0])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[16])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[32])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[48])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[64])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[80])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[96])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[112])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[128])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[144])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[160])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[176])));
  s = vaesmcq_u8(vaeseq_u8(s, vld1q_u8(&round_key[192])));
  s = vaeseq_u8(s, vld1q_u8(&round_key[208]));
  s = veorq_u8(s, vld1q_u8(&round_key[224]));
  vst1q_u8(buf, s);
}
#endif  // USE_AARCH64_AES

static void decrypt_one_block(uint8_t* buf, const uint8_t* round_key) {
  state_t* state = (state_t*)buf;

  // Add the first round key to the state before starting the rounds.
  add_round_key(NUM_ROUNDS, state, round_key);

  // The middle rounds are identical.
  for (int k = NUM_ROUNDS - 1; k >= 1; --k) {
    shift_rows_inv(state);
    sub_bytes_inv(state);
    add_round_key(k, state, round_key);
    mix_columns_inv(state);
  }

  // Final round is without mix_columns_inv().
  shift_rows_inv(state);
  sub_bytes_inv(state);
  add_round_key(0, state, round_key);
}

#if defined(USE_X86_AES)
static void decrypt_one_block_x86(uint8_t* buf, const uint8_t* round_key) {
  __m128i s = _mm_loadu_si128((__m128i*)buf);
  s = _mm_xor_si128(s, _mm_loadu_si128((const __m128i*)&round_key[224]));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[208])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[192])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[176])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[160])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[144])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[128])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[112])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[96])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[80])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[64])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[48])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[32])));
  s = _mm_aesdec_si128(
      s, _mm_aesimc_si128(_mm_loadu_si128((const __m128i*)&round_key[16])));
  s = _mm_aesdeclast_si128(s, _mm_loadu_si128((const __m128i*)&round_key[0]));
  _mm_storeu_si128((__m128i*)buf, s);
}
#endif  // USE_X86_AES

#if defined(USE_AARCH64_AES)
ATTRIBUTE_AARCH64_CRYPTO
static void decrypt_one_block_aarch64(uint8_t* buf, const uint8_t* round_key) {
  const uint8x16_t rk0 = vld1q_u8(round_key + 14 * 16);
  const uint8x16_t rk1 = vaesimcq_u8(vld1q_u8(round_key + 13 * 16));
  const uint8x16_t rk2 = vaesimcq_u8(vld1q_u8(round_key + 12 * 16));
  const uint8x16_t rk3 = vaesimcq_u8(vld1q_u8(round_key + 11 * 16));
  const uint8x16_t rk4 = vaesimcq_u8(vld1q_u8(round_key + 10 * 16));
  const uint8x16_t rk5 = vaesimcq_u8(vld1q_u8(round_key + 9 * 16));
  const uint8x16_t rk6 = vaesimcq_u8(vld1q_u8(round_key + 8 * 16));
  const uint8x16_t rk7 = vaesimcq_u8(vld1q_u8(round_key + 7 * 16));
  const uint8x16_t rk8 = vaesimcq_u8(vld1q_u8(round_key + 6 * 16));
  const uint8x16_t rk9 = vaesimcq_u8(vld1q_u8(round_key + 5 * 16));
  const uint8x16_t rk10 = vaesimcq_u8(vld1q_u8(round_key + 4 * 16));
  const uint8x16_t rk11 = vaesimcq_u8(vld1q_u8(round_key + 3 * 16));
  const uint8x16_t rk12 = vaesimcq_u8(vld1q_u8(round_key + 2 * 16));
  const uint8x16_t rk13 = vaesimcq_u8(vld1q_u8(round_key + 1 * 16));
  const uint8x16_t rk14 = vld1q_u8(round_key + 0 * 16);

  uint8x16_t s = vld1q_u8(buf);
  s = vaesimcq_u8(vaesdq_u8(s, rk0));
  s = vaesimcq_u8(vaesdq_u8(s, rk1));
  s = vaesimcq_u8(vaesdq_u8(s, rk2));
  s = vaesimcq_u8(vaesdq_u8(s, rk3));
  s = vaesimcq_u8(vaesdq_u8(s, rk4));
  s = vaesimcq_u8(vaesdq_u8(s, rk5));
  s = vaesimcq_u8(vaesdq_u8(s, rk6));
  s = vaesimcq_u8(vaesdq_u8(s, rk7));
  s = vaesimcq_u8(vaesdq_u8(s, rk8));
  s = vaesimcq_u8(vaesdq_u8(s, rk9));
  s = vaesimcq_u8(vaesdq_u8(s, rk10));
  s = vaesimcq_u8(vaesdq_u8(s, rk11));
  s = vaesimcq_u8(vaesdq_u8(s, rk12));
  s = veorq_u8(vaesdq_u8(s, rk13), rk14);
  vst1q_u8(buf, s);
}
#endif  // USE_AARCH64_AES

static void xor_with_iv(uint8_t buf[AES_BLOCK_BYTES],
                        const uint8_t iv[AES_BLOCK_BYTES]) {
#if defined(USE_X86_AES)
  // We only use instructions that are guaranteed to be supported by all x86_64
  // CPUs, so no need for any CPUID check.
  _mm_storeu_si128((__m128i*)buf,
                   _mm_xor_si128(_mm_loadu_si128((__m128i*)buf),
                                 _mm_loadu_si128((const __m128i*)iv)));
#elif defined(USE_AARCH64_AES)
  // We only use instructions that are guaranteed to be supported by all NEON
  // enabled AArch64 CPUs, so no need for any capability check.
  vst1q_u8(buf, veorq_u8(vld1q_u8(buf), vld1q_u8(iv)));
#else
  for (int i = 0; i < AES_BLOCK_BYTES; ++i) {
    buf[i] ^= iv[i];
  }
#endif
}

static void detect_cpu_capabilities(aes256_ctx_t* ctx) {
  // Default to the generic method.
  ctx->method = METHOD_GENERIC;

#if defined(USE_X86_AES)
  // Use CPUID to detect actual support for AES.
  cpuid_regs_t regs;
  x86cpuid(0x00000001U, 0U, &regs);
  if ((regs.ecx >> 25) & 1U) {
    // Bit 25: AES.
    ctx->method = METHOD_X86_AES;
  }
#endif

#if defined(USE_AARCH64_AES)
#if defined(USE_GETAUXVAL)
  // Use getauxval() to detect actual support for AES.
  if (getauxval(AT_HWCAP) & HWCAP_AES) {
    ctx->method = METHOD_AARCH64_AES;
  }
#else
  // Assume that the CPU supports the AES extensions.
  ctx->method = METHOD_AARCH64_AES;
#endif  // USE_GETAUXVAL
#endif  // USE_AARCH64_AES
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void aes_256_init_ex(aes256_ctx_t* ctx,
                     const uint8_t key[AES_KEY_BYTES],
                     const uint8_t iv[AES_IV_BYTES],
                     bool force_generic) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);
  PRECONDITION(iv != NULL);

  if (force_generic) {
    ctx->method = METHOD_GENERIC;
  } else {
    detect_cpu_capabilities(ctx);
  }

  key_expansion(ctx->round_key, key);
  memcpy(ctx->iv, iv, AES_BLOCK_BYTES);
}

void aes_256_cbc_encrypt_buffer(aes256_ctx_t* ctx,
                                uint8_t* buf,
                                size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (AES_BLOCK_BYTES - 1)) == 0U);

  const uint8_t* iv = ctx->iv;
  for (size_t i = 0; i < length; i += AES_BLOCK_BYTES) {
    xor_with_iv(buf, iv);
#if defined(USE_X86_AES)
    if (ctx->method == METHOD_X86_AES) {
      encrypt_one_block_x86(buf, ctx->round_key);
    } else
#elif defined(USE_AARCH64_AES)
    if (ctx->method == METHOD_AARCH64_AES) {
      encrypt_one_block_aarch64(buf, ctx->round_key);
    } else
#endif
    {
      encrypt_one_block(buf, ctx->round_key);
    }
    iv = buf;
    buf += AES_BLOCK_BYTES;
  }
  // Store Iv in ctx for next call.
  memcpy(ctx->iv, iv, AES_BLOCK_BYTES);
}

void aes_256_cbc_decrypt_buffer(aes256_ctx_t* ctx,
                                uint8_t* buf,
                                size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (AES_BLOCK_BYTES - 1)) == 0U);

  uint8_t next_iv[AES_BLOCK_BYTES];
  for (size_t i = 0; i < length; i += AES_BLOCK_BYTES) {
    memcpy(next_iv, buf, AES_BLOCK_BYTES);
#if defined(USE_X86_AES)
    if (ctx->method == METHOD_X86_AES) {
      decrypt_one_block_x86(buf, ctx->round_key);
    } else
#elif defined(USE_AARCH64_AES)
    if (ctx->method == METHOD_AARCH64_AES) {
      decrypt_one_block_aarch64(buf, ctx->round_key);
    } else
#endif
    {
      decrypt_one_block(buf, ctx->round_key);
    }
    xor_with_iv(buf, ctx->iv);
    memcpy(ctx->iv, next_iv, AES_BLOCK_BYTES);
    buf += AES_BLOCK_BYTES;
  }
}
