// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "chacha.h"

#include "../error.h"

//------------------------------------------------------------------------------
// Configuration.
//------------------------------------------------------------------------------

#define ROUNDS 20

//------------------------------------------------------------------------------
// Compatibility helpers.
//------------------------------------------------------------------------------

#define U32V(v) ((uint32_t)(v))

static inline uint32_t get_uint32_le(const uint8_t* p) {
  return ((uint32_t)p[0]) | (((uint32_t)p[1]) << 8) | (((uint32_t)p[2]) << 16) |
         (((uint32_t)p[3]) << 24);
}

static inline void set_uint32_le(uint8_t* p, const uint32_t v) {
  p[0] = (uint8_t)(v);
  p[1] = (uint8_t)(v >> 8);
  p[2] = (uint8_t)(v >> 16);
  p[3] = (uint8_t)(v >> 24);
}

#define ROTL32(v, n) (U32V((v) << (n)) | ((v) >> (32 - (n))))
#define ROTATE(v, c) (ROTL32(v, c))
#define XOR(v, w) ((v) ^ (w))
#define PLUS(v, w) (U32V((v) + (w)))
#define PLUSONE(v) (PLUS((v), 1))

//------------------------------------------------------------------------------
// Implementation.
//------------------------------------------------------------------------------

static const uint8_t SIGMA[16] = "expand 32-byte k";

#define QUARTERROUND(a, b, c, d)      \
  x[a] = PLUS(x[a], x[b]);            \
  x[d] = ROTATE(XOR(x[d], x[a]), 16); \
  x[c] = PLUS(x[c], x[d]);            \
  x[b] = ROTATE(XOR(x[b], x[c]), 12); \
  x[a] = PLUS(x[a], x[b]);            \
  x[d] = ROTATE(XOR(x[d], x[a]), 8);  \
  x[c] = PLUS(x[c], x[d]);            \
  x[b] = ROTATE(XOR(x[b], x[c]), 7);

static void hchacha20(uint8_t subkey[CHACHA_KEY_BYTES],
                      const uint8_t key[CHACHA_KEY_BYTES],
                      const uint8_t iv[16]) {
  uint32_t x[16];

  // Constant setup.
  x[0] = get_uint32_le(&SIGMA[0]);
  x[1] = get_uint32_le(&SIGMA[4]);
  x[2] = get_uint32_le(&SIGMA[8]);
  x[3] = get_uint32_le(&SIGMA[12]);

  // Key setup.
  x[4] = get_uint32_le(&key[0]);
  x[5] = get_uint32_le(&key[4]);
  x[6] = get_uint32_le(&key[8]);
  x[7] = get_uint32_le(&key[12]);
  x[8] = get_uint32_le(&key[16]);
  x[9] = get_uint32_le(&key[20]);
  x[10] = get_uint32_le(&key[24]);
  x[11] = get_uint32_le(&key[28]);

  // IV setup.
  x[12] = get_uint32_le(&iv[0]);
  x[13] = get_uint32_le(&iv[4]);
  x[14] = get_uint32_le(&iv[8]);
  x[15] = get_uint32_le(&iv[12]);

  for (int i = ROUNDS; i > 0; i -= 2) {
    QUARTERROUND(0, 4, 8, 12);
    QUARTERROUND(1, 5, 9, 13);
    QUARTERROUND(2, 6, 10, 14);
    QUARTERROUND(3, 7, 11, 15);
    QUARTERROUND(0, 5, 10, 15);
    QUARTERROUND(1, 6, 11, 12);
    QUARTERROUND(2, 7, 8, 13);
    QUARTERROUND(3, 4, 9, 14);
  }

  set_uint32_le(&subkey[0], x[0]);
  set_uint32_le(&subkey[4], x[1]);
  set_uint32_le(&subkey[8], x[2]);
  set_uint32_le(&subkey[12], x[3]);
  set_uint32_le(&subkey[16], x[12]);
  set_uint32_le(&subkey[20], x[13]);
  set_uint32_le(&subkey[24], x[14]);
  set_uint32_le(&subkey[28], x[15]);
}

static void salsa20_wordtobyte(uint8_t output[64], const uint32_t input[16]) {
  uint32_t x[16];

  for (int i = 0; i < 16; ++i) {
    x[i] = input[i];
  }
  for (int i = ROUNDS; i > 0; i -= 2) {
    QUARTERROUND(0, 4, 8, 12)
    QUARTERROUND(1, 5, 9, 13)
    QUARTERROUND(2, 6, 10, 14)
    QUARTERROUND(3, 7, 11, 15)
    QUARTERROUND(0, 5, 10, 15)
    QUARTERROUND(1, 6, 11, 12)
    QUARTERROUND(2, 7, 8, 13)
    QUARTERROUND(3, 4, 9, 14)
  }
  for (int i = 0; i < 16; ++i) {
    x[i] = PLUS(x[i], input[i]);
  }
  for (int i = 0; i < 16; ++i) {
    const int i4 = i * 4;
    set_uint32_le(&output[i4], x[i]);
  }
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void chacha20_256_init(chacha256_ctx_t* ctx,
                       const uint8_t key[CHACHA_KEY_BYTES],
                       const uint8_t iv[CHACHA_IV_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);
  PRECONDITION(iv != NULL);

  // Constant setup.
  ctx->input[0] = get_uint32_le(&SIGMA[0]);
  ctx->input[1] = get_uint32_le(&SIGMA[4]);
  ctx->input[2] = get_uint32_le(&SIGMA[8]);
  ctx->input[3] = get_uint32_le(&SIGMA[12]);

  // Key setup.
  ctx->input[4] = get_uint32_le(&key[0]);
  ctx->input[5] = get_uint32_le(&key[4]);
  ctx->input[6] = get_uint32_le(&key[8]);
  ctx->input[7] = get_uint32_le(&key[12]);
  ctx->input[8] = get_uint32_le(&key[16]);
  ctx->input[9] = get_uint32_le(&key[20]);
  ctx->input[10] = get_uint32_le(&key[24]);
  ctx->input[11] = get_uint32_le(&key[28]);

  // Counter setup.
  ctx->input[12] = 0;
  ctx->input[13] = 0;

  // IV setup.
  ctx->input[14] = get_uint32_le(&iv[0]);
  ctx->input[15] = get_uint32_le(&iv[4]);
}

void xchacha20_256_init(chacha256_ctx_t* ctx,
                        const uint8_t key[CHACHA_KEY_BYTES],
                        const uint8_t iv[XCHACHA_IV_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);
  PRECONDITION(iv != NULL);

  // Generate the sub-key to use from the 256-bit key and 192-bit IV.
  uint8_t subkey[CHACHA_KEY_BYTES];
  hchacha20(subkey, key, &iv[0]);

  // Initialize the ChaCha20 context as usual, but with the sub-key instead.
  chacha20_256_init(ctx, subkey, &iv[16]);
}

void chacha20_256_encrypt_buffer(chacha256_ctx_t* ctx,
                                 uint8_t* buf,
                                 size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (CHACHA_BLOCK_BYTES - 1)) == 0U);

  uint8_t output[64];
  for (size_t i = 0U; i < length; i += 64U) {
    salsa20_wordtobyte(output, ctx->input);
    ctx->input[12] = PLUSONE(ctx->input[12]);
    if (ctx->input[12] == 0U) {
      // Note: This case (32-bit carry arithmetic) can only happen when
      // the stream length exceeds 256 GB.
      ctx->input[13] = PLUSONE(ctx->input[13]);
    }
    for (int k = 0; k < 64; ++k) {
      buf[k] = buf[k] ^ output[k];
    }
    buf += 64;
  }
}
