// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// An implementation of the ChaCha cipher.
// Based on the public domain reference implementation by D. J. Bernstein.
// See: https://cr.yp.to/chacha.html
//------------------------------------------------------------------------------

#ifndef CHACHA_H_
#define CHACHA_H_

#include <stddef.h>
#include <stdint.h>

// The number of bytes for a ChaCha-256 key (256 bits = 32 bytes).
#define CHACHA_KEY_BYTES 32

// The number of bytes for a ChaCha-256 IV (64 bits = 8 bytes).
#define CHACHA_IV_BYTES 8

// The number of bytes for a XChaCha-256 IV (192 bits = 24 bytes).
#define XCHACHA_IV_BYTES 24

// While ChaCha is a stream cipher it performs its operation in 64-byte chunks,
// and this implementation requires data to be en/decrypted in multiples of
// 64 bytes.
#define CHACHA_BLOCK_BYTES 64

typedef struct {
  uint32_t input[16];
} chacha256_ctx_t;

void chacha20_256_init(chacha256_ctx_t* ctx,
                       const uint8_t key[CHACHA_KEY_BYTES],
                       const uint8_t iv[CHACHA_IV_BYTES]);

void xchacha20_256_init(chacha256_ctx_t* ctx,
                        const uint8_t key[CHACHA_KEY_BYTES],
                        const uint8_t iv[XCHACHA_IV_BYTES]);

void chacha20_256_encrypt_buffer(chacha256_ctx_t* ctx,
                                 uint8_t* buf,
                                 size_t length);

static inline void chacha20_256_decrypt_buffer(chacha256_ctx_t* ctx,
                                               uint8_t* buf,
                                               size_t length) {
  // ChaCha decryption is identical to encryption.
  chacha20_256_encrypt_buffer(ctx, buf, length);
}

#endif  // CHACHA_H_
