// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "entropy.h"

#include "../debug.h"
#include "../error.h"
#include "../memory.h"
#include <stdbool.h>
#include <string.h>

#if defined(__APPLE__)
#include "TargetConditionals.h"
#endif

#if defined(_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
// Include order matters... Include <windows.h> first.
#include <bcrypt.h>
#include <winternl.h>
#define USE_BCRYPTGENRANDOM
#else
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#define USE_DEV_RANDOM
#endif

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

#if defined(USE_DEV_RANDOM)
static bool read_bytes(const char* file_name,
                       uint8_t* buf,
                       const size_t num_bytes) {
  int fd = open(file_name, O_RDONLY);
  if (fd == -1) {
    return false;
  }

  size_t bytes_left = num_bytes;
  while (bytes_left > 0U) {
    const ssize_t bytes_read = read(fd, buf, bytes_left);
    if (bytes_read <= 0) {
      break;
    }
    bytes_left -= (size_t)bytes_read;
  }
  close(fd);

  return bytes_left == 0U;
}
#endif  // USE_DEV_RANDOM

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void make_entropy(uint8_t* out_buffer, const size_t num_bytes) {
  PRECONDITION(out_buffer != NULL);

  bool success = false;

#if defined(USE_BCRYPTGENRANDOM)
  const NTSTATUS status = BCryptGenRandom(NULL,
                                          (BYTE*)out_buffer,
                                          (ULONG)num_bytes,
                                          BCRYPT_USE_SYSTEM_PREFERRED_RNG);
  if (NT_SUCCESS(status)) {
    success = true;
  }
#elif defined(USE_DEV_RANDOM)
  // We prefer /dev/urandom over /dev/random.
  // * https://security.stackexchange.com/a/3939
  // * https://www.2uo.de/myths-about-urandom/
  success = read_bytes("/dev/urandom", out_buffer, num_bytes);
  if (!success) {
    success = read_bytes("/dev/random", out_buffer, num_bytes);
  }
#else
// TODO: Implement for more platforms.
#error "Missing a CSPRNG implementation for this platform"
#endif

  if (!success) {
    FATAL_ERROR("Unable to get system entropy");
  }

  DBG_DUMP_HEX(out_buffer, num_bytes);
}
