// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// This is a Poly1305 implementation inspired by the public domain
// "Poly1305-donna" by Andrew Moon.
//
// This implementation uses 32 bit * 32 bit = 64 bit multiplication and 64 bit
// addition, and is almost ten times faster than the reference implementation by
// D. J. Bernstein that uses 8-bit arithmetic.
//------------------------------------------------------------------------------

#include "poly1305.h"

#include "../error.h"
#include <string.h>

//------------------------------------------------------------------------------
// Private helper functions.
//------------------------------------------------------------------------------

#if defined(__GNUC__) || defined(__clang__)
#define NOINLINE __attribute__((noinline))
#else
#define NOINLINE
#endif

// Interpret four 8 bit unsigned integers as a 32 bit unsigned integer in little
// endian.
static uint32_t load_u32_le(const uint8_t* p) {
  return ((uint32_t)p[0]) | (((uint32_t)p[1]) << 8) | (((uint32_t)p[2]) << 16) |
         (((uint32_t)p[3]) << 24);
}

// Store a 32 bit unsigned integer as four 8 bit unsigned integers in little
// endian.
static void store_u32_le(uint8_t* p, uint32_t v) {
  p[0] = (uint8_t)v;
  p[1] = (uint8_t)(v >> 8);
  p[2] = (uint8_t)(v >> 16);
  p[3] = (uint8_t)(v >> 24);
}

// Note: We appear to get slightly better performance on x86_64 with NOINLINE,
// and the code gets smaller, so win-win.
static NOINLINE void poly1305_do_blocks(poly1305_ctx_t* ctx,
                                        const uint8_t* message,
                                        size_t bytes) {
  const uint32_t r0 = ctx->r[0];
  const uint32_t r1 = ctx->r[1];
  const uint32_t r2 = ctx->r[2];
  const uint32_t r3 = ctx->r[3];
  const uint32_t r4 = ctx->r[4];

  const uint32_t s1 = r1 * 5U;
  const uint32_t s2 = r2 * 5U;
  const uint32_t s3 = r3 * 5U;
  const uint32_t s4 = r4 * 5U;

  uint32_t h0 = ctx->h[0];
  uint32_t h1 = ctx->h[1];
  uint32_t h2 = ctx->h[2];
  uint32_t h3 = ctx->h[3];
  uint32_t h4 = ctx->h[4];

  const uint32_t hibit = ctx->final ? 0 : (1U << 24);  // 1 << 128

  while (bytes >= POLY1305_BLOCK_BYTES) {
    // h += message[i]
    h0 += (load_u32_le(&message[0])) & 0x3ffffffU;
    h1 += (load_u32_le(&message[3]) >> 2) & 0x3ffffffU;
    h2 += (load_u32_le(&message[6]) >> 4) & 0x3ffffffU;
    h3 += (load_u32_le(&message[9]) >> 6) & 0x3ffffffU;
    h4 += (load_u32_le(&message[12]) >> 8) | hibit;

    // h *= r
    uint64_t d0 = ((uint64_t)h0 * r0) + ((uint64_t)h1 * s4) +
                  ((uint64_t)h2 * s3) + ((uint64_t)h3 * s2) +
                  ((uint64_t)h4 * s1);
    uint64_t d1 = ((uint64_t)h0 * r1) + ((uint64_t)h1 * r0) +
                  ((uint64_t)h2 * s4) + ((uint64_t)h3 * s3) +
                  ((uint64_t)h4 * s2);
    uint64_t d2 = ((uint64_t)h0 * r2) + ((uint64_t)h1 * r1) +
                  ((uint64_t)h2 * r0) + ((uint64_t)h3 * s4) +
                  ((uint64_t)h4 * s3);
    uint64_t d3 = ((uint64_t)h0 * r3) + ((uint64_t)h1 * r2) +
                  ((uint64_t)h2 * r1) + ((uint64_t)h3 * r0) +
                  ((uint64_t)h4 * s4);
    uint64_t d4 = ((uint64_t)h0 * r4) + ((uint64_t)h1 * r3) +
                  ((uint64_t)h2 * r2) + ((uint64_t)h3 * r1) +
                  ((uint64_t)h4 * r0);

    // (partial) h %= p
    uint32_t c = (uint32_t)(d0 >> 26);
    h0 = (uint32_t)d0 & 0x3ffffffU;
    d1 += c;
    c = (uint32_t)(d1 >> 26);
    h1 = (uint32_t)d1 & 0x3ffffffU;
    d2 += c;
    c = (uint32_t)(d2 >> 26);
    h2 = (uint32_t)d2 & 0x3ffffffU;
    d3 += c;
    c = (uint32_t)(d3 >> 26);
    h3 = (uint32_t)d3 & 0x3ffffffU;
    d4 += c;
    c = (uint32_t)(d4 >> 26);
    h4 = (uint32_t)d4 & 0x3ffffffU;
    h0 += c * 5U;
    c = (h0 >> 26);
    h0 = h0 & 0x3ffffffU;
    h1 += c;

    message += POLY1305_BLOCK_BYTES;
    bytes -= POLY1305_BLOCK_BYTES;
  }

  ctx->h[0] = h0;
  ctx->h[1] = h1;
  ctx->h[2] = h2;
  ctx->h[3] = h3;
  ctx->h[4] = h4;
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void poly1305_init(poly1305_ctx_t* ctx, const uint8_t key[32]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);

  // r = key[0-15] & 0xffffffc0ffffffc0ffffffc0fffffff
  ctx->r[0] = (load_u32_le(&key[0])) & 0x3ffffffU;
  ctx->r[1] = (load_u32_le(&key[3]) >> 2) & 0x3ffff03U;
  ctx->r[2] = (load_u32_le(&key[6]) >> 4) & 0x3ffc0ffU;
  ctx->r[3] = (load_u32_le(&key[9]) >> 6) & 0x3f03fffU;
  ctx->r[4] = (load_u32_le(&key[12]) >> 8) & 0x00fffffU;

  // h = 0
  ctx->h[0] = 0;
  ctx->h[1] = 0;
  ctx->h[2] = 0;
  ctx->h[3] = 0;
  ctx->h[4] = 0;

  // s = key[16-31]
  ctx->s[0] = load_u32_le(&key[16]);
  ctx->s[1] = load_u32_le(&key[20]);
  ctx->s[2] = load_u32_le(&key[24]);
  ctx->s[3] = load_u32_le(&key[28]);

  memset(&ctx->buffer[0], 0, sizeof(ctx->buffer));
  ctx->leftover = 0;
  ctx->final = false;
}

void poly1305_append(poly1305_ctx_t* ctx,
                     const uint8_t* message,
                     size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(message != NULL);

  // Handle leftover.
  if (ctx->leftover) {
    size_t want = (POLY1305_BLOCK_BYTES - ctx->leftover);
    if (want > length) {
      want = length;
    }
    for (size_t i = 0; i < want; i++) {
      ctx->buffer[ctx->leftover + i] = message[i];
    }
    length -= want;
    message += want;
    ctx->leftover += want;
    if (ctx->leftover < POLY1305_BLOCK_BYTES) {
      return;
    }
    poly1305_do_blocks(ctx, ctx->buffer, POLY1305_BLOCK_BYTES);
    ctx->leftover = 0;
  }

  // Process full blocks.
  if (length >= POLY1305_BLOCK_BYTES) {
    size_t want = (length & ~(POLY1305_BLOCK_BYTES - 1));
    poly1305_do_blocks(ctx, message, want);
    message += want;
    length -= want;
  }

  // Store leftover.
  if (length != 0U) {
    for (size_t i = 0; i < length; i++) {
      ctx->buffer[ctx->leftover + i] = message[i];
    }
    ctx->leftover += length;
  }
}

void poly1305_finalize(poly1305_ctx_t* ctx, uint8_t hash[POLY1305_HASH_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(hash != NULL);

  // Process the remaining block (if any).
  if (ctx->leftover) {
    size_t i = ctx->leftover;
    ctx->buffer[i++] = 1;
    for (; i < POLY1305_BLOCK_BYTES; i++) {
      ctx->buffer[i] = 0;
    }
    ctx->final = true;
    poly1305_do_blocks(ctx, ctx->buffer, POLY1305_BLOCK_BYTES);
  }

  // Fully carry h.
  uint32_t h0 = ctx->h[0];
  uint32_t h1 = ctx->h[1];
  uint32_t h2 = ctx->h[2];
  uint32_t h3 = ctx->h[3];
  uint32_t h4 = ctx->h[4];

  uint32_t c = h1 >> 26;
  h1 = h1 & 0x3ffffffU;
  h2 += c;
  c = h2 >> 26;
  h2 = h2 & 0x3ffffffU;
  h3 += c;
  c = h3 >> 26;
  h3 = h3 & 0x3ffffffU;
  h4 += c;
  c = h4 >> 26;
  h4 = h4 & 0x3ffffffU;
  h0 += c * 5U;
  c = h0 >> 26;
  h0 = h0 & 0x3ffffffU;
  h1 += c;

  // compute h + -p
  uint32_t g0 = h0 + 5;
  c = g0 >> 26;
  g0 &= 0x3ffffffU;
  uint32_t g1 = h1 + c;
  c = g1 >> 26;
  g1 &= 0x3ffffffU;
  uint32_t g2 = h2 + c;
  c = g2 >> 26;
  g2 &= 0x3ffffffU;
  uint32_t g3 = h3 + c;
  c = g3 >> 26;
  g3 &= 0x3ffffffU;
  uint32_t g4 = h4 + c - (1U << 26);

  // Select h if h < p, or h + -p if h >= p.
  uint32_t mask = (g4 >> ((sizeof(uint32_t) * 8U) - 1)) - 1;
  g0 &= mask;
  g1 &= mask;
  g2 &= mask;
  g3 &= mask;
  g4 &= mask;
  mask = ~mask;
  h0 = (h0 & mask) | g0;
  h1 = (h1 & mask) | g1;
  h2 = (h2 & mask) | g2;
  h3 = (h3 & mask) | g3;
  h4 = (h4 & mask) | g4;

  // h = h % (2^128)
  h0 = ((h0) | (h1 << 26)) & 0xffffffffU;
  h1 = ((h1 >> 6) | (h2 << 20)) & 0xffffffffU;
  h2 = ((h2 >> 12) | (h3 << 14)) & 0xffffffffU;
  h3 = ((h3 >> 18) | (h4 << 8)) & 0xffffffffU;

  // hash = (h + s) % (2^128)
  uint64_t f = (uint64_t)h0 + ctx->s[0];
  h0 = (uint32_t)f;
  f = (uint64_t)h1 + ctx->s[1] + (f >> 32);
  h1 = (uint32_t)f;
  f = (uint64_t)h2 + ctx->s[2] + (f >> 32);
  h2 = (uint32_t)f;
  f = (uint64_t)h3 + ctx->s[3] + (f >> 32);
  h3 = (uint32_t)f;

  store_u32_le(&hash[0], h0);
  store_u32_le(&hash[4], h1);
  store_u32_le(&hash[8], h2);
  store_u32_le(&hash[12], h3);
}

void poly1305(const uint8_t key[POLY1305_KEY_BYTES],
              const uint8_t* message,
              size_t length,
              uint8_t hash[POLY1305_HASH_BYTES]) {
  poly1305_ctx_t ctx;
  poly1305_init(&ctx, key);
  poly1305_append(&ctx, message, length);
  poly1305_finalize(&ctx, hash);
}
