// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "serpent.h"

#include "../error.h"
#include <ctype.h>
#include <string.h>

// Can we use x86 SSE intrinsics?
#if defined(__x86_64__) || defined(_M_AMD64) || defined(_M_X64) || \
    defined(__amd64__) || defined(__amd64)
#include <emmintrin.h>
#define USE_X86_SSE
#endif

//------------------------------------------------------------------------------
// Private functions & macros.
//------------------------------------------------------------------------------

#define ROUNDS 32       // # of rounds.
#define PHI 0x9e3779b9  // Magic number used in key schedule initialisation.

static inline uint32_t get_uint32_le(const uint8_t* b) {
  return (((uint32_t)b[0]) << 0) | (((uint32_t)b[1]) << 8) |
         (((uint32_t)b[2]) << 16) | (((uint32_t)b[3]) << 24);
}

static inline void set_uint32_le(uint8_t* b, const uint32_t w) {
  b[0] = (uint8_t)(w >> 0);
  b[1] = (uint8_t)(w >> 8);
  b[2] = (uint8_t)(w >> 16);
  b[3] = (uint8_t)(w >> 24);
}

static inline void bytes_to_uint32_le(uint32_t* words,
                                      const uint8_t* bytes,
                                      const int num_words) {
  const uint8_t* src = &bytes[0];
  uint32_t* dst = &words[0];
  for (int i = 0; i < num_words; ++i) {
    *dst++ = get_uint32_le(src);
    src += 4;
  }
}

static inline uint32_t rotlConstant(const uint32_t x, const unsigned r) {
  return (x << r) | (x >> (32 - r));
}

static inline uint32_t rotrConstant(const uint32_t x, const unsigned r) {
  return (x >> r) | (x << (32 - r));
}

// Linear transformation.
#define LT(i, a, b, c, d, e)                        \
  {                                                 \
    (a) = rotlConstant(a, 13);                      \
    (c) = rotlConstant(c, 3);                       \
    (d) = rotlConstant((d) ^ (c) ^ ((a) << 3), 7);  \
    (b) = rotlConstant((b) ^ (a) ^ (c), 1);         \
    (a) = rotlConstant((a) ^ (b) ^ (d), 5);         \
    (c) = rotlConstant((c) ^ (d) ^ ((b) << 7), 22); \
  }

// Inverse linear transformation.
#define ILT(i, a, b, c, d, e)                    \
  {                                              \
    (c) = rotrConstant(c, 22);                   \
    (a) = rotrConstant(a, 5);                    \
    (c) ^= (d) ^ ((b) << 7);                     \
    (a) ^= (b) ^ (d);                            \
    (b) = rotrConstant(b, 1);                    \
    (d) = rotrConstant(d, 7) ^ (c) ^ ((a) << 3); \
    (b) ^= (a) ^ (c);                            \
    (c) = rotrConstant(c, 3);                    \
    (a) = rotrConstant(a, 13);                   \
  }

// Order of output from S-box functions.
#define beforeS0(f) f(0, a, b, c, d, e)
#define afterS0(f) f(1, b, e, c, a, d)
#define afterS1(f) f(2, c, b, a, e, d)
#define afterS2(f) f(3, a, e, b, d, c)
#define afterS3(f) f(4, e, b, d, c, a)
#define afterS4(f) f(5, b, a, e, c, d)
#define afterS5(f) f(6, a, c, b, e, d)
#define afterS6(f) f(7, a, c, d, b, e)
#define afterS7(f) f(8, d, e, b, a, c)

// Order of output from inverse S-box functions.
#define beforeI7(f) f(8, a, b, c, d, e)
#define afterI7(f) f(7, d, a, b, e, c)
#define afterI6(f) f(6, a, b, c, e, d)
#define afterI5(f) f(5, b, d, e, c, a)
#define afterI4(f) f(4, b, c, e, a, d)
#define afterI3(f) f(3, a, b, e, c, d)
#define afterI2(f) f(2, b, d, e, c, a)
#define afterI1(f) f(1, a, b, c, e, d)
#define afterI0(f) f(0, a, d, b, e, c)

// The instruction sequences for the S-box functions come from Dag Arne Osvik's
// paper "Speeding up Serpent": https://www.ii.uib.no/~osvik/pub/aes3.pdf
#define S0(i, r0, r1, r2, r3, r4) \
  {                               \
    (r3) ^= (r0);                 \
    (r4) = r1;                    \
    (r1) &= (r3);                 \
    (r4) ^= (r2);                 \
    (r1) ^= (r0);                 \
    (r0) |= (r3);                 \
    (r0) ^= (r4);                 \
    (r4) ^= (r3);                 \
    (r3) ^= (r2);                 \
    (r2) |= (r1);                 \
    (r2) ^= (r4);                 \
    (r4) = ~(r4);                 \
    (r4) |= (r1);                 \
    (r1) ^= (r3);                 \
    (r1) ^= (r4);                 \
    (r3) |= (r0);                 \
    (r1) ^= (r3);                 \
    (r4) ^= (r3);                 \
  }

#define I0(i, r0, r1, r2, r3, r4) \
  {                               \
    (r2) = ~(r2);                 \
    (r4) = r1;                    \
    (r1) |= (r0);                 \
    (r4) = ~(r4);                 \
    (r1) ^= (r2);                 \
    (r2) |= (r4);                 \
    (r1) ^= (r3);                 \
    (r0) ^= (r4);                 \
    (r2) ^= (r0);                 \
    (r0) &= (r3);                 \
    (r4) ^= (r0);                 \
    (r0) |= (r1);                 \
    (r0) ^= (r2);                 \
    (r3) ^= (r4);                 \
    (r2) ^= (r1);                 \
    (r3) ^= (r0);                 \
    (r3) ^= (r1);                 \
    (r2) &= (r3);                 \
    (r4) ^= (r2);                 \
  }

#define S1(i, r0, r1, r2, r3, r4) \
  {                               \
    (r0) = ~(r0);                 \
    (r2) = ~(r2);                 \
    (r4) = r0;                    \
    (r0) &= (r1);                 \
    (r2) ^= (r0);                 \
    (r0) |= (r3);                 \
    (r3) ^= (r2);                 \
    (r1) ^= (r0);                 \
    (r0) ^= (r4);                 \
    (r4) |= (r1);                 \
    (r1) ^= (r3);                 \
    (r2) |= (r0);                 \
    (r2) &= (r4);                 \
    (r0) ^= (r1);                 \
    (r1) &= (r2);                 \
    (r1) ^= (r0);                 \
    (r0) &= (r2);                 \
    (r0) ^= (r4);                 \
  }

#define I1(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r1;                    \
    (r1) ^= (r3);                 \
    (r3) &= (r1);                 \
    (r4) ^= (r2);                 \
    (r3) ^= (r0);                 \
    (r0) |= (r1);                 \
    (r2) ^= (r3);                 \
    (r0) ^= (r4);                 \
    (r0) |= (r2);                 \
    (r1) ^= (r3);                 \
    (r0) ^= (r1);                 \
    (r1) |= (r3);                 \
    (r1) ^= (r0);                 \
    (r4) = ~(r4);                 \
    (r4) ^= (r1);                 \
    (r1) |= (r0);                 \
    (r1) ^= (r0);                 \
    (r1) |= (r4);                 \
    (r3) ^= (r1);                 \
  }

#define S2(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r0;                    \
    (r0) &= (r2);                 \
    (r0) ^= (r3);                 \
    (r2) ^= (r1);                 \
    (r2) ^= (r0);                 \
    (r3) |= (r4);                 \
    (r3) ^= (r1);                 \
    (r4) ^= (r2);                 \
    (r1) = r3;                    \
    (r3) |= (r4);                 \
    (r3) ^= (r0);                 \
    (r0) &= (r1);                 \
    (r4) ^= (r0);                 \
    (r1) ^= (r3);                 \
    (r1) ^= (r4);                 \
    (r4) = ~(r4);                 \
  }

#define I2(i, r0, r1, r2, r3, r4) \
  {                               \
    (r2) ^= (r3);                 \
    (r3) ^= (r0);                 \
    (r4) = r3;                    \
    (r3) &= (r2);                 \
    (r3) ^= (r1);                 \
    (r1) |= (r2);                 \
    (r1) ^= (r4);                 \
    (r4) &= (r3);                 \
    (r2) ^= (r3);                 \
    (r4) &= (r0);                 \
    (r4) ^= (r2);                 \
    (r2) &= (r1);                 \
    (r2) |= (r0);                 \
    (r3) = ~(r3);                 \
    (r2) ^= (r3);                 \
    (r0) ^= (r3);                 \
    (r0) &= (r1);                 \
    (r3) ^= (r4);                 \
    (r3) ^= (r0);                 \
  }

#define S3(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r0;                    \
    (r0) |= (r3);                 \
    (r3) ^= (r1);                 \
    (r1) &= (r4);                 \
    (r4) ^= (r2);                 \
    (r2) ^= (r3);                 \
    (r3) &= (r0);                 \
    (r4) |= (r1);                 \
    (r3) ^= (r4);                 \
    (r0) ^= (r1);                 \
    (r4) &= (r0);                 \
    (r1) ^= (r3);                 \
    (r4) ^= (r2);                 \
    (r1) |= (r0);                 \
    (r1) ^= (r2);                 \
    (r0) ^= (r3);                 \
    (r2) = r1;                    \
    (r1) |= (r3);                 \
    (r1) ^= (r0);                 \
  }

#define I3(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r2;                    \
    (r2) ^= (r1);                 \
    (r1) &= (r2);                 \
    (r1) ^= (r0);                 \
    (r0) &= (r4);                 \
    (r4) ^= (r3);                 \
    (r3) |= (r1);                 \
    (r3) ^= (r2);                 \
    (r0) ^= (r4);                 \
    (r2) ^= (r0);                 \
    (r0) |= (r3);                 \
    (r0) ^= (r1);                 \
    (r4) ^= (r2);                 \
    (r2) &= (r3);                 \
    (r1) |= (r3);                 \
    (r1) ^= (r2);                 \
    (r4) ^= (r0);                 \
    (r2) ^= (r4);                 \
  }

#define S4(i, r0, r1, r2, r3, r4) \
  {                               \
    (r1) ^= (r3);                 \
    (r3) = ~(r3);                 \
    (r2) ^= (r3);                 \
    (r3) ^= (r0);                 \
    (r4) = r1;                    \
    (r1) &= (r3);                 \
    (r1) ^= (r2);                 \
    (r4) ^= (r3);                 \
    (r0) ^= (r4);                 \
    (r2) &= (r4);                 \
    (r2) ^= (r0);                 \
    (r0) &= (r1);                 \
    (r3) ^= (r0);                 \
    (r4) |= (r1);                 \
    (r4) ^= (r0);                 \
    (r0) |= (r3);                 \
    (r0) ^= (r2);                 \
    (r2) &= (r3);                 \
    (r0) = ~(r0);                 \
    (r4) ^= (r2);                 \
  }

#define I4(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r2;                    \
    (r2) &= (r3);                 \
    (r2) ^= (r1);                 \
    (r1) |= (r3);                 \
    (r1) &= (r0);                 \
    (r4) ^= (r2);                 \
    (r4) ^= (r1);                 \
    (r1) &= (r2);                 \
    (r0) = ~(r0);                 \
    (r3) ^= (r4);                 \
    (r1) ^= (r3);                 \
    (r3) &= (r0);                 \
    (r3) ^= (r2);                 \
    (r0) ^= (r1);                 \
    (r2) &= (r0);                 \
    (r3) ^= (r0);                 \
    (r2) ^= (r4);                 \
    (r2) |= (r3);                 \
    (r3) ^= (r0);                 \
    (r2) ^= (r1);                 \
  }

#define S5(i, r0, r1, r2, r3, r4) \
  {                               \
    (r0) ^= (r1);                 \
    (r1) ^= (r3);                 \
    (r3) = ~(r3);                 \
    (r4) = r1;                    \
    (r1) &= (r0);                 \
    (r2) ^= (r3);                 \
    (r1) ^= (r2);                 \
    (r2) |= (r4);                 \
    (r4) ^= (r3);                 \
    (r3) &= (r1);                 \
    (r3) ^= (r0);                 \
    (r4) ^= (r1);                 \
    (r4) ^= (r2);                 \
    (r2) ^= (r0);                 \
    (r0) &= (r3);                 \
    (r2) = ~(r2);                 \
    (r0) ^= (r4);                 \
    (r4) |= (r3);                 \
    (r2) ^= (r4);                 \
  }

#define I5(i, r0, r1, r2, r3, r4) \
  {                               \
    (r1) = ~(r1);                 \
    (r4) = r3;                    \
    (r2) ^= (r1);                 \
    (r3) |= (r0);                 \
    (r3) ^= (r2);                 \
    (r2) |= (r1);                 \
    (r2) &= (r0);                 \
    (r4) ^= (r3);                 \
    (r2) ^= (r4);                 \
    (r4) |= (r0);                 \
    (r4) ^= (r1);                 \
    (r1) &= (r2);                 \
    (r1) ^= (r3);                 \
    (r4) ^= (r2);                 \
    (r3) &= (r4);                 \
    (r4) ^= (r1);                 \
    (r3) ^= (r0);                 \
    (r3) ^= (r4);                 \
    (r4) = ~(r4);                 \
  }

#define S6(i, r0, r1, r2, r3, r4) \
  {                               \
    (r2) = ~(r2);                 \
    (r4) = r3;                    \
    (r3) &= (r0);                 \
    (r0) ^= (r4);                 \
    (r3) ^= (r2);                 \
    (r2) |= (r4);                 \
    (r1) ^= (r3);                 \
    (r2) ^= (r0);                 \
    (r0) |= (r1);                 \
    (r2) ^= (r1);                 \
    (r4) ^= (r0);                 \
    (r0) |= (r3);                 \
    (r0) ^= (r2);                 \
    (r4) ^= (r3);                 \
    (r4) ^= (r0);                 \
    (r3) = ~(r3);                 \
    (r2) &= (r4);                 \
    (r2) ^= (r3);                 \
  }

#define I6(i, r0, r1, r2, r3, r4) \
  {                               \
    (r0) ^= (r2);                 \
    (r4) = r2;                    \
    (r2) &= (r0);                 \
    (r4) ^= (r3);                 \
    (r2) = ~(r2);                 \
    (r3) ^= (r1);                 \
    (r2) ^= (r3);                 \
    (r4) |= (r0);                 \
    (r0) ^= (r2);                 \
    (r3) ^= (r4);                 \
    (r4) ^= (r1);                 \
    (r1) &= (r3);                 \
    (r1) ^= (r0);                 \
    (r0) ^= (r3);                 \
    (r0) |= (r2);                 \
    (r3) ^= (r1);                 \
    (r4) ^= (r0);                 \
  }

#define S7(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r2;                    \
    (r2) &= (r1);                 \
    (r2) ^= (r3);                 \
    (r3) &= (r1);                 \
    (r4) ^= (r2);                 \
    (r2) ^= (r1);                 \
    (r1) ^= (r0);                 \
    (r0) |= (r4);                 \
    (r0) ^= (r2);                 \
    (r3) ^= (r1);                 \
    (r2) ^= (r3);                 \
    (r3) &= (r0);                 \
    (r3) ^= (r4);                 \
    (r4) ^= (r2);                 \
    (r2) &= (r0);                 \
    (r4) = ~(r4);                 \
    (r2) ^= (r4);                 \
    (r4) &= (r0);                 \
    (r1) ^= (r3);                 \
    (r4) ^= (r1);                 \
  }

#define I7(i, r0, r1, r2, r3, r4) \
  {                               \
    (r4) = r2;                    \
    (r2) ^= (r0);                 \
    (r0) &= (r3);                 \
    (r2) = ~(r2);                 \
    (r4) |= (r3);                 \
    (r3) ^= (r1);                 \
    (r1) |= (r0);                 \
    (r0) ^= (r2);                 \
    (r2) &= (r4);                 \
    (r1) ^= (r2);                 \
    (r2) ^= (r0);                 \
    (r0) |= (r2);                 \
    (r3) &= (r4);                 \
    (r0) ^= (r3);                 \
    (r4) ^= (r1);                 \
    (r3) ^= (r4);                 \
    (r4) |= (r0);                 \
    (r3) ^= (r2);                 \
    (r4) ^= (r2);                 \
  }

// Key xor.
#define KX(r, a, b, c, d, e) \
  {                          \
    (a) ^= k[(4 * (r)) + 0]; \
    (b) ^= k[(4 * (r)) + 1]; \
    (c) ^= k[(4 * (r)) + 2]; \
    (d) ^= k[(4 * (r)) + 3]; \
  }

#define LK(r, a, b, c, d, e)      \
  {                               \
    (a) = k[((8 - (r)) * 4) + 0]; \
    (b) = k[((8 - (r)) * 4) + 1]; \
    (c) = k[((8 - (r)) * 4) + 2]; \
    (d) = k[((8 - (r)) * 4) + 3]; \
  }

#define SK(r, a, b, c, d, e)    \
  {                             \
    k[((8 - (r)) * 4) + 4] = a; \
    k[((8 - (r)) * 4) + 5] = b; \
    k[((8 - (r)) * 4) + 6] = c; \
    k[((8 - (r)) * 4) + 7] = d; \
  }

static void key_schedule(uint32_t* k,
                         const uint8_t user_key[SERPENT_KEY_BYTES]) {
  uint32_t k0[8];
  bytes_to_uint32_le(k0, user_key, 8);

  uint32_t t = k0[7];
  unsigned int i = 0;
  for (i = 0; i < 8; ++i) {
    k[i] = k0[i] = t = rotlConstant(
        k0[i] ^ k0[(i + 3) % 8] ^ k0[(i + 5) % 8] ^ t ^ PHI ^ i, 11);
  }
  for (i = 8; i < 4 * (ROUNDS + 1); ++i) {
    k[i] = t = rotlConstant(k[i - 8] ^ k[i - 5] ^ k[i - 3] ^ t ^ PHI ^ i, 11);
  }
  k -= 20;

  uint32_t a = 0;
  uint32_t b = 0;
  uint32_t c = 0;
  uint32_t d = 0;
  uint32_t e = 0;
  for (i = 0; i < ROUNDS / 8; i++) {
    afterS2(LK);
    afterS2(S3);
    afterS3(SK);
    afterS1(LK);
    afterS1(S2);
    afterS2(SK);
    afterS0(LK);
    afterS0(S1);
    afterS1(SK);
    beforeS0(LK);
    beforeS0(S0);
    afterS0(SK);
    k += 8 * 4;
    afterS6(LK);
    afterS6(S7);
    afterS7(SK);
    afterS5(LK);
    afterS5(S6);
    afterS6(SK);
    afterS4(LK);
    afterS4(S5);
    afterS5(SK);
    afterS3(LK);
    afterS3(S4);
    afterS4(SK);
  }
  afterS2(LK);
  afterS2(S3);
  afterS3(SK);
}

static void encrypt_one_block(const serpent256_ctx_t* ctx, uint8_t* block) {
  uint32_t a = get_uint32_le(&block[0]);
  uint32_t b = get_uint32_le(&block[4]);
  uint32_t c = get_uint32_le(&block[8]);
  uint32_t d = get_uint32_le(&block[12]);
  uint32_t e = 0U;

  const uint32_t* k = ctx->expkey;
  unsigned int i = 1;

  do {
    beforeS0(KX);
    beforeS0(S0);
    afterS0(LT);
    afterS0(KX);
    afterS0(S1);
    afterS1(LT);
    afterS1(KX);
    afterS1(S2);
    afterS2(LT);
    afterS2(KX);
    afterS2(S3);
    afterS3(LT);
    afterS3(KX);
    afterS3(S4);
    afterS4(LT);
    afterS4(KX);
    afterS4(S5);
    afterS5(LT);
    afterS5(KX);
    afterS5(S6);
    afterS6(LT);
    afterS6(KX);
    afterS6(S7);

    if (i == 4) {
      break;
    }

    ++i;
    c = b;
    b = e;
    e = d;
    d = a;
    a = e;
    k += 32;
    beforeS0(LT);
  } while (1);

  afterS7(KX);

  set_uint32_le(&block[0], d);
  set_uint32_le(&block[4], e);
  set_uint32_le(&block[8], b);
  set_uint32_le(&block[12], a);
}

static void decrypt_one_block(const serpent256_ctx_t* ctx, uint8_t* block) {
  uint32_t a = get_uint32_le(&block[0]);
  uint32_t b = get_uint32_le(&block[4]);
  uint32_t c = get_uint32_le(&block[8]);
  uint32_t d = get_uint32_le(&block[12]);
  uint32_t e = 0U;

  const uint32_t* k = ctx->expkey + 96;
  unsigned int i = 4;

  beforeI7(KX);
  goto start;

  do {
    c = b;
    b = d;
    d = e;
    k -= 32;
    beforeI7(ILT);
  start:
    beforeI7(I7);
    afterI7(KX);
    afterI7(ILT);
    afterI7(I6);
    afterI6(KX);
    afterI6(ILT);
    afterI6(I5);
    afterI5(KX);
    afterI5(ILT);
    afterI5(I4);
    afterI4(KX);
    afterI4(ILT);
    afterI4(I3);
    afterI3(KX);
    afterI3(ILT);
    afterI3(I2);
    afterI2(KX);
    afterI2(ILT);
    afterI2(I1);
    afterI1(KX);
    afterI1(ILT);
    afterI1(I0);
    afterI0(KX);
  } while (--i != 0);

  set_uint32_le(&block[0], a);
  set_uint32_le(&block[4], d);
  set_uint32_le(&block[8], b);
  set_uint32_le(&block[12], e);
}

static void xor_block_and_iv(const uint8_t* iv, uint8_t* buf) {
#if defined(USE_X86_SSE)
  // SERPENT_BLOCK_BYTES = 16, i.e. 128 bits.
  _mm_storeu_si128((__m128i*)buf,
                   _mm_xor_si128(_mm_loadu_si128((__m128i*)buf),
                                 _mm_loadu_si128((const __m128i*)iv)));
#else
  for (int i = 0; i < SERPENT_BLOCK_BYTES; ++i) {
    buf[i] ^= iv[i];
  }
#endif
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void serpent_256_init(serpent256_ctx_t* ctx,
                      const uint8_t key[SERPENT_KEY_BYTES],
                      const uint8_t iv[SERPENT_IV_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);
  PRECONDITION(iv != NULL);

  // Copy IV.
  memcpy(ctx->iv, iv, SERPENT_BLOCK_BYTES);

  // Prepare keys.
  key_schedule(ctx->expkey, key);
}

void serpent_256_cbc_encrypt_buffer(serpent256_ctx_t* ctx,
                                    uint8_t* buf,
                                    size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (SERPENT_BLOCK_BYTES - 1)) == 0U);

  const uint8_t* iv = ctx->iv;
  for (size_t i = 0; i < length; i += SERPENT_BLOCK_BYTES) {
    xor_block_and_iv(iv, buf);
    encrypt_one_block(ctx, buf);
    iv = buf;
    buf += SERPENT_BLOCK_BYTES;
  }
  memcpy(ctx->iv, iv, SERPENT_BLOCK_BYTES);
}

void serpent_256_cbc_decrypt_buffer(serpent256_ctx_t* ctx,
                                    uint8_t* buf,
                                    size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (SERPENT_BLOCK_BYTES - 1)) == 0U);

  uint8_t tmp_iv[SERPENT_BLOCK_BYTES];
  for (size_t i = 0; i < length; i += SERPENT_BLOCK_BYTES) {
    memcpy(tmp_iv, buf, SERPENT_BLOCK_BYTES);
    decrypt_one_block(ctx, buf);
    xor_block_and_iv(ctx->iv, buf);
    memcpy(ctx->iv, tmp_iv, SERPENT_BLOCK_BYTES);
    buf += SERPENT_BLOCK_BYTES;
  }
}
