// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "sha2.h"

#include "../error.h"
#include <string.h>

//------------------------------------------------------------------------------
// Private helper functions.
//------------------------------------------------------------------------------

// Read a big endian 32-bit word from a byte array.
static inline uint32_t get_uint32_be(const uint8_t* ptr) {
  return (((uint32_t)(ptr[0])) << 24) | (((uint32_t)(ptr[1])) << 16) |
         (((uint32_t)(ptr[2])) << 8) | (uint32_t)(ptr[3]);
}

// Write a big endian 32-bit word to a byte array.
static inline void set_uint32_be(const uint32_t x, uint8_t* ptr) {
  ptr[0] = (uint8_t)(x >> 24);
  ptr[1] = (uint8_t)(x >> 16);
  ptr[2] = (uint8_t)(x >> 8);
  ptr[3] = (uint8_t)(x);
}

// Write a big endian 64-bit word to a byte array.
static inline void set_uint64_be(const uint64_t x, uint8_t* ptr) {
  ptr[0] = (uint8_t)(x >> 56);
  ptr[1] = (uint8_t)(x >> 48);
  ptr[2] = (uint8_t)(x >> 40);
  ptr[3] = (uint8_t)(x >> 32);
  ptr[4] = (uint8_t)(x >> 24);
  ptr[5] = (uint8_t)(x >> 16);
  ptr[6] = (uint8_t)(x >> 8);
  ptr[7] = (uint8_t)(x >> 0);
}

// Right rotate a 32-bit word.
static inline uint32_t ror(const uint32_t x, const uint32_t shift) {
  return (x >> shift) | (x << (32U - shift));
}

// Minimum value, size_t.
static inline size_t min_size(const size_t a, const size_t b) {
  return a < b ? a : b;
}

//------------------------------------------------------------------------------
// Implementation.
//------------------------------------------------------------------------------

static void hash_one_chunk(sha2_256_ctx_t* ctx) {
  // Array of round constants.
  static const uint32_t k[64] = {
      0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
      0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
      0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
      0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
      0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
      0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
      0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
      0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
      0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
      0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
      0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2};

  // Work buffer for each chunk.
  uint32_t w[64];

  // Extract the chunk as sixteen 32-bit words.
  for (size_t i = 0U; i < 16U; ++i) {
    w[i] = get_uint32_be(&ctx->buffer[i * 4U]);
  }

  // Extend the sixteen 32-bit words into eighty 32-bit words.
  for (size_t i = 16U; i < 64U; ++i) {
    uint32_t s0 = ror(w[i - 15], 7) ^ ror(w[i - 15], 18) ^ (w[i - 15] >> 3);
    uint32_t s1 = ror(w[i - 2], 17) ^ ror(w[i - 2], 19) ^ (w[i - 2] >> 10);
    w[i] = w[i - 16] + s0 + w[i - 7] + s1;
  }

  // Initialize working variables to current hash value.
  uint32_t a = ctx->h[0];
  uint32_t b = ctx->h[1];
  uint32_t c = ctx->h[2];
  uint32_t d = ctx->h[3];
  uint32_t e = ctx->h[4];
  uint32_t f = ctx->h[5];
  uint32_t g = ctx->h[6];
  uint32_t h = ctx->h[7];

  // Compression function main loop.
  for (size_t i = 0U; i < 64U; ++i) {
    const uint32_t S1 = ror(e, 6) ^ ror(e, 11) ^ ror(e, 25);
    const uint32_t ch = (e & f) ^ ((~e) & g);
    const uint32_t temp1 = h + S1 + ch + k[i] + w[i];
    const uint32_t S0 = ror(a, 2) ^ ror(a, 13) ^ ror(a, 22);
    const uint32_t maj = (a & b) ^ (a & c) ^ (b & c);
    const uint32_t temp2 = S0 + maj;

    h = g;
    g = f;
    f = e;
    e = d + temp1;
    d = c;
    c = b;
    b = a;
    a = temp1 + temp2;
  }

  // Add the compressed chunk to the current hash value.
  ctx->h[0] += a;
  ctx->h[1] += b;
  ctx->h[2] += c;
  ctx->h[3] += d;
  ctx->h[4] += e;
  ctx->h[5] += f;
  ctx->h[6] += g;
  ctx->h[7] += h;
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void sha2_256_init(sha2_256_ctx_t* ctx) {
  PRECONDITION(ctx != NULL);

  // Initial state of the hash.
  ctx->h[0] = 0x6a09e667;
  ctx->h[1] = 0xbb67ae85;
  ctx->h[2] = 0x3c6ef372;
  ctx->h[3] = 0xa54ff53a;
  ctx->h[4] = 0x510e527f;
  ctx->h[5] = 0x9b05688c;
  ctx->h[6] = 0x1f83d9ab;
  ctx->h[7] = 0x5be0cd19;

  ctx->total_length = 0U;
}

void sha2_256_append(sha2_256_ctx_t* ctx, const uint8_t* msg, size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(msg != NULL);

  size_t buf_pos = ctx->total_length & (SHA256_CHUNK_BYTES - 1);
  ctx->total_length += length;

  while (length != 0U) {
    // Try to fill up the internal buffer.
    const size_t bytes_to_copy = min_size(SHA256_CHUNK_BYTES - buf_pos, length);
    memcpy(&ctx->buffer[buf_pos], msg, bytes_to_copy);
    msg += bytes_to_copy;
    buf_pos += bytes_to_copy;
    length -= bytes_to_copy;

    // If we filled up the buffer, hash the chunk.
    if (buf_pos == SHA256_CHUNK_BYTES) {
      hash_one_chunk(ctx);
      buf_pos = 0U;
    }
  }
}

void sha2_256_finalize(sha2_256_ctx_t* ctx, uint8_t hash[SHA256_HASH_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(hash != NULL);

  // The original message size, in bits.
  const uint64_t original_length_bits = 8U * (uint64_t)ctx->total_length;

  size_t buf_pos = ctx->total_length & (SHA256_CHUNK_BYTES - 1);

  // Add a 1-bit.
  ctx->buffer[buf_pos++] = 0x80U;
  if (buf_pos == SHA256_CHUNK_BYTES) {
    hash_one_chunk(ctx);
    buf_pos = 0U;
  }

  // Zero-fill + add original bit length. This may require an extra chunk if the
  // 64-bit length specification does not fit in the current chunk.
  int done = 0;
  while (!done) {
    size_t bytes_left = SHA256_CHUNK_BYTES - buf_pos;
    memset(&ctx->buffer[buf_pos], 0, bytes_left);
    if (bytes_left >= 8U) {
      // Store the 64-bit word at the end of the buffer.
      set_uint64_be(original_length_bits,
                    &ctx->buffer[SHA256_CHUNK_BYTES - 8U]);
      done = 1;
    }
    hash_one_chunk(ctx);
    buf_pos = 0U;
  }

  // Write the hash to the output buffer.
  set_uint32_be(ctx->h[0], hash);
  set_uint32_be(ctx->h[1], hash + 4);
  set_uint32_be(ctx->h[2], hash + 8);
  set_uint32_be(ctx->h[3], hash + 12);
  set_uint32_be(ctx->h[4], hash + 16);
  set_uint32_be(ctx->h[5], hash + 20);
  set_uint32_be(ctx->h[6], hash + 24);
  set_uint32_be(ctx->h[7], hash + 28);

  // After finalization, the context is cleared.
  sha2_256_init(ctx);
}

void sha2_256(const uint8_t* msg,
              size_t length,
              uint8_t hash[SHA256_HASH_BYTES]) {
  PRECONDITION(msg != NULL);
  PRECONDITION(hash != NULL);

  sha2_256_ctx_t ctx;
  sha2_256_init(&ctx);
  sha2_256_append(&ctx, msg, length);
  sha2_256_finalize(&ctx, hash);
}
