// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "twofish.h"

#include "../error.h"
#include <string.h>

// Can we use x86 SSE intrinsics?
#if defined(__x86_64__) || defined(_M_AMD64) || defined(_M_X64) || \
    defined(__amd64__) || defined(__amd64)
#include <emmintrin.h>
#define USE_X86_SSE
#endif

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

#define RS_MOD 0x14D
#define RHO 0x01010101L

static const uint8_t RS[4][8] = {
    {
        0x01,
        0xA4,
        0x55,
        0x87,
        0x5A,
        0x58,
        0xDB,
        0x9E,
    },
    {
        0xA4,
        0x56,
        0x82,
        0xF3,
        0x1E,
        0xC6,
        0x68,
        0xE5,
    },
    {
        0x02,
        0xA1,
        0xFC,
        0xC1,
        0x47,
        0xAE,
        0x3D,
        0x19,
    },
    {
        0xA4,
        0x55,
        0x87,
        0x5A,
        0x58,
        0xDB,
        0x9E,
        0x03,
    },
};

static const uint8_t Q0[] = {
    0xA9, 0x67, 0xB3, 0xE8, 0x04, 0xFD, 0xA3, 0x76, 0x9A, 0x92, 0x80, 0x78,
    0xE4, 0xDD, 0xD1, 0x38, 0x0D, 0xC6, 0x35, 0x98, 0x18, 0xF7, 0xEC, 0x6C,
    0x43, 0x75, 0x37, 0x26, 0xFA, 0x13, 0x94, 0x48, 0xF2, 0xD0, 0x8B, 0x30,
    0x84, 0x54, 0xDF, 0x23, 0x19, 0x5B, 0x3D, 0x59, 0xF3, 0xAE, 0xA2, 0x82,
    0x63, 0x01, 0x83, 0x2E, 0xD9, 0x51, 0x9B, 0x7C, 0xA6, 0xEB, 0xA5, 0xBE,
    0x16, 0x0C, 0xE3, 0x61, 0xC0, 0x8C, 0x3A, 0xF5, 0x73, 0x2C, 0x25, 0x0B,
    0xBB, 0x4E, 0x89, 0x6B, 0x53, 0x6A, 0xB4, 0xF1, 0xE1, 0xE6, 0xBD, 0x45,
    0xE2, 0xF4, 0xB6, 0x66, 0xCC, 0x95, 0x03, 0x56, 0xD4, 0x1C, 0x1E, 0xD7,
    0xFB, 0xC3, 0x8E, 0xB5, 0xE9, 0xCF, 0xBF, 0xBA, 0xEA, 0x77, 0x39, 0xAF,
    0x33, 0xC9, 0x62, 0x71, 0x81, 0x79, 0x09, 0xAD, 0x24, 0xCD, 0xF9, 0xD8,
    0xE5, 0xC5, 0xB9, 0x4D, 0x44, 0x08, 0x86, 0xE7, 0xA1, 0x1D, 0xAA, 0xED,
    0x06, 0x70, 0xB2, 0xD2, 0x41, 0x7B, 0xA0, 0x11, 0x31, 0xC2, 0x27, 0x90,
    0x20, 0xF6, 0x60, 0xFF, 0x96, 0x5C, 0xB1, 0xAB, 0x9E, 0x9C, 0x52, 0x1B,
    0x5F, 0x93, 0x0A, 0xEF, 0x91, 0x85, 0x49, 0xEE, 0x2D, 0x4F, 0x8F, 0x3B,
    0x47, 0x87, 0x6D, 0x46, 0xD6, 0x3E, 0x69, 0x64, 0x2A, 0xCE, 0xCB, 0x2F,
    0xFC, 0x97, 0x05, 0x7A, 0xAC, 0x7F, 0xD5, 0x1A, 0x4B, 0x0E, 0xA7, 0x5A,
    0x28, 0x14, 0x3F, 0x29, 0x88, 0x3C, 0x4C, 0x02, 0xB8, 0xDA, 0xB0, 0x17,
    0x55, 0x1F, 0x8A, 0x7D, 0x57, 0xC7, 0x8D, 0x74, 0xB7, 0xC4, 0x9F, 0x72,
    0x7E, 0x15, 0x22, 0x12, 0x58, 0x07, 0x99, 0x34, 0x6E, 0x50, 0xDE, 0x68,
    0x65, 0xBC, 0xDB, 0xF8, 0xC8, 0xA8, 0x2B, 0x40, 0xDC, 0xFE, 0x32, 0xA4,
    0xCA, 0x10, 0x21, 0xF0, 0xD3, 0x5D, 0x0F, 0x00, 0x6F, 0x9D, 0x36, 0x42,
    0x4A, 0x5E, 0xC1, 0xE0,
};

static const uint8_t Q1[] = {
    0x75, 0xF3, 0xC6, 0xF4, 0xDB, 0x7B, 0xFB, 0xC8, 0x4A, 0xD3, 0xE6, 0x6B,
    0x45, 0x7D, 0xE8, 0x4B, 0xD6, 0x32, 0xD8, 0xFD, 0x37, 0x71, 0xF1, 0xE1,
    0x30, 0x0F, 0xF8, 0x1B, 0x87, 0xFA, 0x06, 0x3F, 0x5E, 0xBA, 0xAE, 0x5B,
    0x8A, 0x00, 0xBC, 0x9D, 0x6D, 0xC1, 0xB1, 0x0E, 0x80, 0x5D, 0xD2, 0xD5,
    0xA0, 0x84, 0x07, 0x14, 0xB5, 0x90, 0x2C, 0xA3, 0xB2, 0x73, 0x4C, 0x54,
    0x92, 0x74, 0x36, 0x51, 0x38, 0xB0, 0xBD, 0x5A, 0xFC, 0x60, 0x62, 0x96,
    0x6C, 0x42, 0xF7, 0x10, 0x7C, 0x28, 0x27, 0x8C, 0x13, 0x95, 0x9C, 0xC7,
    0x24, 0x46, 0x3B, 0x70, 0xCA, 0xE3, 0x85, 0xCB, 0x11, 0xD0, 0x93, 0xB8,
    0xA6, 0x83, 0x20, 0xFF, 0x9F, 0x77, 0xC3, 0xCC, 0x03, 0x6F, 0x08, 0xBF,
    0x40, 0xE7, 0x2B, 0xE2, 0x79, 0x0C, 0xAA, 0x82, 0x41, 0x3A, 0xEA, 0xB9,
    0xE4, 0x9A, 0xA4, 0x97, 0x7E, 0xDA, 0x7A, 0x17, 0x66, 0x94, 0xA1, 0x1D,
    0x3D, 0xF0, 0xDE, 0xB3, 0x0B, 0x72, 0xA7, 0x1C, 0xEF, 0xD1, 0x53, 0x3E,
    0x8F, 0x33, 0x26, 0x5F, 0xEC, 0x76, 0x2A, 0x49, 0x81, 0x88, 0xEE, 0x21,
    0xC4, 0x1A, 0xEB, 0xD9, 0xC5, 0x39, 0x99, 0xCD, 0xAD, 0x31, 0x8B, 0x01,
    0x18, 0x23, 0xDD, 0x1F, 0x4E, 0x2D, 0xF9, 0x48, 0x4F, 0xF2, 0x65, 0x8E,
    0x78, 0x5C, 0x58, 0x19, 0x8D, 0xE5, 0x98, 0x57, 0x67, 0x7F, 0x05, 0x64,
    0xAF, 0x63, 0xB6, 0xFE, 0xF5, 0xB7, 0x3C, 0xA5, 0xCE, 0xE9, 0x68, 0x44,
    0xE0, 0x4D, 0x43, 0x69, 0x29, 0x2E, 0xAC, 0x15, 0x59, 0xA8, 0x0A, 0x9E,
    0x6E, 0x47, 0xDF, 0x34, 0x35, 0x6A, 0xCF, 0xDC, 0x22, 0xC9, 0xC0, 0x9B,
    0x89, 0xD4, 0xED, 0xAB, 0x12, 0xA2, 0x0D, 0x52, 0xBB, 0x02, 0x2F, 0xA9,
    0xD7, 0x61, 0x1E, 0xB4, 0x50, 0x04, 0xF6, 0xC2, 0x16, 0x25, 0x86, 0x56,
    0x55, 0x09, 0xBE, 0x91,
};

static const uint8_t mult5B[] = {
    0x00, 0x5B, 0xB6, 0xED, 0x05, 0x5E, 0xB3, 0xE8, 0x0A, 0x51, 0xBC, 0xE7,
    0x0F, 0x54, 0xB9, 0xE2, 0x14, 0x4F, 0xA2, 0xF9, 0x11, 0x4A, 0xA7, 0xFC,
    0x1E, 0x45, 0xA8, 0xF3, 0x1B, 0x40, 0xAD, 0xF6, 0x28, 0x73, 0x9E, 0xC5,
    0x2D, 0x76, 0x9B, 0xC0, 0x22, 0x79, 0x94, 0xCF, 0x27, 0x7C, 0x91, 0xCA,
    0x3C, 0x67, 0x8A, 0xD1, 0x39, 0x62, 0x8F, 0xD4, 0x36, 0x6D, 0x80, 0xDB,
    0x33, 0x68, 0x85, 0xDE, 0x50, 0x0B, 0xE6, 0xBD, 0x55, 0x0E, 0xE3, 0xB8,
    0x5A, 0x01, 0xEC, 0xB7, 0x5F, 0x04, 0xE9, 0xB2, 0x44, 0x1F, 0xF2, 0xA9,
    0x41, 0x1A, 0xF7, 0xAC, 0x4E, 0x15, 0xF8, 0xA3, 0x4B, 0x10, 0xFD, 0xA6,
    0x78, 0x23, 0xCE, 0x95, 0x7D, 0x26, 0xCB, 0x90, 0x72, 0x29, 0xC4, 0x9F,
    0x77, 0x2C, 0xC1, 0x9A, 0x6C, 0x37, 0xDA, 0x81, 0x69, 0x32, 0xDF, 0x84,
    0x66, 0x3D, 0xD0, 0x8B, 0x63, 0x38, 0xD5, 0x8E, 0xA0, 0xFB, 0x16, 0x4D,
    0xA5, 0xFE, 0x13, 0x48, 0xAA, 0xF1, 0x1C, 0x47, 0xAF, 0xF4, 0x19, 0x42,
    0xB4, 0xEF, 0x02, 0x59, 0xB1, 0xEA, 0x07, 0x5C, 0xBE, 0xE5, 0x08, 0x53,
    0xBB, 0xE0, 0x0D, 0x56, 0x88, 0xD3, 0x3E, 0x65, 0x8D, 0xD6, 0x3B, 0x60,
    0x82, 0xD9, 0x34, 0x6F, 0x87, 0xDC, 0x31, 0x6A, 0x9C, 0xC7, 0x2A, 0x71,
    0x99, 0xC2, 0x2F, 0x74, 0x96, 0xCD, 0x20, 0x7B, 0x93, 0xC8, 0x25, 0x7E,
    0xF0, 0xAB, 0x46, 0x1D, 0xF5, 0xAE, 0x43, 0x18, 0xFA, 0xA1, 0x4C, 0x17,
    0xFF, 0xA4, 0x49, 0x12, 0xE4, 0xBF, 0x52, 0x09, 0xE1, 0xBA, 0x57, 0x0C,
    0xEE, 0xB5, 0x58, 0x03, 0xEB, 0xB0, 0x5D, 0x06, 0xD8, 0x83, 0x6E, 0x35,
    0xDD, 0x86, 0x6B, 0x30, 0xD2, 0x89, 0x64, 0x3F, 0xD7, 0x8C, 0x61, 0x3A,
    0xCC, 0x97, 0x7A, 0x21, 0xC9, 0x92, 0x7F, 0x24, 0xC6, 0x9D, 0x70, 0x2B,
    0xC3, 0x98, 0x75, 0x2E,
};

static const uint8_t multEF[] = {
    0x00, 0xEF, 0xB7, 0x58, 0x07, 0xE8, 0xB0, 0x5F, 0x0E, 0xE1, 0xB9, 0x56,
    0x09, 0xE6, 0xBE, 0x51, 0x1C, 0xF3, 0xAB, 0x44, 0x1B, 0xF4, 0xAC, 0x43,
    0x12, 0xFD, 0xA5, 0x4A, 0x15, 0xFA, 0xA2, 0x4D, 0x38, 0xD7, 0x8F, 0x60,
    0x3F, 0xD0, 0x88, 0x67, 0x36, 0xD9, 0x81, 0x6E, 0x31, 0xDE, 0x86, 0x69,
    0x24, 0xCB, 0x93, 0x7C, 0x23, 0xCC, 0x94, 0x7B, 0x2A, 0xC5, 0x9D, 0x72,
    0x2D, 0xC2, 0x9A, 0x75, 0x70, 0x9F, 0xC7, 0x28, 0x77, 0x98, 0xC0, 0x2F,
    0x7E, 0x91, 0xC9, 0x26, 0x79, 0x96, 0xCE, 0x21, 0x6C, 0x83, 0xDB, 0x34,
    0x6B, 0x84, 0xDC, 0x33, 0x62, 0x8D, 0xD5, 0x3A, 0x65, 0x8A, 0xD2, 0x3D,
    0x48, 0xA7, 0xFF, 0x10, 0x4F, 0xA0, 0xF8, 0x17, 0x46, 0xA9, 0xF1, 0x1E,
    0x41, 0xAE, 0xF6, 0x19, 0x54, 0xBB, 0xE3, 0x0C, 0x53, 0xBC, 0xE4, 0x0B,
    0x5A, 0xB5, 0xED, 0x02, 0x5D, 0xB2, 0xEA, 0x05, 0xE0, 0x0F, 0x57, 0xB8,
    0xE7, 0x08, 0x50, 0xBF, 0xEE, 0x01, 0x59, 0xB6, 0xE9, 0x06, 0x5E, 0xB1,
    0xFC, 0x13, 0x4B, 0xA4, 0xFB, 0x14, 0x4C, 0xA3, 0xF2, 0x1D, 0x45, 0xAA,
    0xF5, 0x1A, 0x42, 0xAD, 0xD8, 0x37, 0x6F, 0x80, 0xDF, 0x30, 0x68, 0x87,
    0xD6, 0x39, 0x61, 0x8E, 0xD1, 0x3E, 0x66, 0x89, 0xC4, 0x2B, 0x73, 0x9C,
    0xC3, 0x2C, 0x74, 0x9B, 0xCA, 0x25, 0x7D, 0x92, 0xCD, 0x22, 0x7A, 0x95,
    0x90, 0x7F, 0x27, 0xC8, 0x97, 0x78, 0x20, 0xCF, 0x9E, 0x71, 0x29, 0xC6,
    0x99, 0x76, 0x2E, 0xC1, 0x8C, 0x63, 0x3B, 0xD4, 0x8B, 0x64, 0x3C, 0xD3,
    0x82, 0x6D, 0x35, 0xDA, 0x85, 0x6A, 0x32, 0xDD, 0xA8, 0x47, 0x1F, 0xF0,
    0xAF, 0x40, 0x18, 0xF7, 0xA6, 0x49, 0x11, 0xFE, 0xA1, 0x4E, 0x16, 0xF9,
    0xB4, 0x5B, 0x03, 0xEC, 0xB3, 0x5C, 0x04, 0xEB, 0xBA, 0x55, 0x0D, 0xE2,
    0xBD, 0x52, 0x0A, 0xE5,
};

static inline uint32_t ROL(const uint32_t x, const unsigned n) {
  return (x << (n & 0x1F)) | (x >> (32 - (n & 0x1F)));
}

static inline uint32_t ROR(const uint32_t x, const unsigned n) {
  return (x >> (n & 0x1F)) | (x << (32 - (n & 0x1F)));
}

// Swap bytes on big endian machines.
#if defined(__BYTE_ORDER__) && defined(__ORDER_BIG_ENDIAN__) && \
    (__BYTE_ORDER__ == __ORDER_BIG_ENDIAN__)
#if defined(__GNUC__)
#define BSWAP(x) (__builtin_bswap32(x))
#else
#define BSWAP(x) (((ROR(x, 8) & 0xFF00FF00) | (ROL(x, 8) & 0x00FF00FF)))
#endif
#else
#define BSWAP(x) (x)
#endif

// Extract byte number.
static inline uint32_t bN(const uint32_t x, const unsigned N) {
  return (x >> (N * 8U)) & 0xFFU;
}

#define b0(x) ((uint8_t)(x))
#define b1(x) ((uint8_t)((x) >> 8))
#define b2(x) (uint8_t)((x) >> 16)
#define b3(x) (uint8_t)((x) >> 24)

#define BYTEARRAY_TO_U32(r)                                  \
  ((((uint32_t)(r)[0]) << 24) ^ (((uint32_t)(r)[1]) << 16) ^ \
   (((uint32_t)(r)[2]) << 8) ^ ((uint32_t)(r)[3]))
#define BYTES_TO_U32(r0, r1, r2, r3)                     \
  ((((uint32_t)(r0)) << 24) ^ (((uint32_t)(r1)) << 16) ^ \
   (((uint32_t)(r2)) << 8) ^ ((uint32_t)(r3)))

// Multiply two polynomials represented as uint32_t's, actually called with
// bytes, but since I'm not really going to too much work to optimize key setup
// (since raw encryption speed is what I'm after), big deal.
static uint32_t poly_mult(uint32_t a, uint32_t b) {
  uint32_t t = 0;
  while (a) {
    if (a & 1) {
      t ^= b;
    }
    b <<= 1;
    a >>= 1;
  }
  return t;
}

// Take the polynomial t and return the t % modulus in GF(256).
static uint32_t gf_mod(uint32_t t, uint32_t modulus) {
  modulus <<= 7;
  for (int i = 0; i < 8; ++i) {
    const uint32_t tt = t ^ modulus;
    if (tt < t) {
      t = tt;
    }
    modulus >>= 1;
  }
  return t;
}

// Multiply a and b and return the modulus.
#define gf_mult(a, b, modulus) gf_mod(poly_mult(a, b), modulus)

static uint32_t rs_matrix_multiply(uint8_t sd[8]) {
  uint8_t result[4];
  for (int j = 0; j < 4; ++j) {
    uint8_t t = 0U;
    for (int k = 0; k < 8; ++k) {
      t ^= gf_mult(RS[j][k], sd[k], RS_MOD);
    }
    result[3 - j] = t;
  }
  return BYTEARRAY_TO_U32(result);
}

// The zero-keyed h function (used by the key setup routine).
static uint32_t h(uint32_t X, const uint32_t L[4], int k) {
  uint8_t y0 = b0(X);
  uint8_t y1 = b1(X);
  uint8_t y2 = b2(X);
  uint8_t y3 = b3(X);

  switch (k) {
    case 4:
      y0 = Q1[y0] ^ b0(L[3]);
      y1 = Q0[y1] ^ b1(L[3]);
      y2 = Q0[y2] ^ b2(L[3]);
      y3 = Q1[y3] ^ b3(L[3]);
      // Fall through
    case 3:
      y0 = Q1[y0] ^ b0(L[2]);
      y1 = Q1[y1] ^ b1(L[2]);
      y2 = Q0[y2] ^ b2(L[2]);
      y3 = Q0[y3] ^ b3(L[2]);
      // Fall through
    case 2:
    default:
      y0 = Q1[Q0[Q0[y0] ^ b0(L[1])] ^ b0(L[0])];
      y1 = Q0[Q0[Q1[y1] ^ b1(L[1])] ^ b1(L[0])];
      y2 = Q1[Q1[Q0[y2] ^ b2(L[1])] ^ b2(L[0])];
      y3 = Q0[Q1[Q1[y3] ^ b3(L[1])] ^ b3(L[0])];
  }

  // Inline the MDS matrix multiply.
  const uint8_t z0 = multEF[y0] ^ y1 ^ multEF[y2] ^ mult5B[y3];
  const uint8_t z1 = multEF[y0] ^ mult5B[y1] ^ y2 ^ multEF[y3];
  const uint8_t z2 = mult5B[y0] ^ multEF[y1] ^ multEF[y2] ^ y3;
  const uint8_t z3 = y0 ^ multEF[y1] ^ mult5B[y2] ^ mult5B[y3];

  return BYTES_TO_U32(z0, z1, z2, z3);
}

static void key_sched_256(const uint8_t M[], uint32_t S[4], uint32_t K[40]) {
  static const int k = 4;  // For 256-bit keys.

  uint32_t Mo[4];
  uint32_t Me[4];
  for (int i = 0; i < k; ++i) {
    const int i2 = i * 2;
    Me[i] = BSWAP(((uint32_t*)M)[i2]);
    Mo[i] = BSWAP(((uint32_t*)M)[i2 + 1]);
  }

  uint8_t vector[8];
  for (int i = 0; i < k; ++i) {
    for (int j = 0; j < 4; ++j) {
      vector[j] = bN(Me[i], j);
    }
    for (int j = 0; j < 4; ++j) {
      vector[j + 4] = bN(Mo[i], j);
    }
    S[k - i - 1] = rs_matrix_multiply(vector);
  }

  for (int i = 0; i < 20; ++i) {
    const int i2 = i * 2;
    const uint32_t A = h(i2 * RHO, Me, k);
    const uint32_t B = ROL(h((i2 * RHO) + RHO, Mo, k), 8);
    K[i2] = A + B;
    K[i2 + 1] = ROL(A + (2 * B), 9);
  }
}

// Given the Sbox keys, create the fully keyed QF.
static void create_full_key_256(const uint32_t L[4], uint32_t QF[4][256]) {
  // For all input values to the Q permutations.
  for (int i = 0; i < 256; ++i) {
    // Run the Q permutations.
    uint8_t y0 = (uint8_t)i;
    uint8_t y1 = (uint8_t)i;
    uint8_t y2 = (uint8_t)i;
    uint8_t y3 = (uint8_t)i;

    y0 = Q1[y0] ^ b0(L[3]);
    y1 = Q0[y1] ^ b1(L[3]);
    y2 = Q0[y2] ^ b2(L[3]);
    y3 = Q1[y3] ^ b3(L[3]);

    y0 = Q1[y0] ^ b0(L[2]);
    y1 = Q1[y1] ^ b1(L[2]);
    y2 = Q0[y2] ^ b2(L[2]);
    y3 = Q0[y3] ^ b3(L[2]);

    y0 = Q1[Q0[Q0[y0] ^ b0(L[1])] ^ b0(L[0])];
    y1 = Q0[Q0[Q1[y1] ^ b1(L[1])] ^ b1(L[0])];
    y2 = Q1[Q1[Q0[y2] ^ b2(L[1])] ^ b2(L[0])];
    y3 = Q0[Q1[Q1[y3] ^ b3(L[1])] ^ b3(L[0])];

    // Now do the partial MDS matrix multiplies.
    QF[0][i] =
        ((((uint32_t)multEF[y0]) << 24) | (((uint32_t)multEF[y0]) << 16) |
         (((uint32_t)mult5B[y0]) << 8) | ((uint32_t)y0));
    QF[1][i] = ((((uint32_t)y1) << 24) | (((uint32_t)mult5B[y1]) << 16) |
                (((uint32_t)multEF[y1]) << 8) | ((uint32_t)multEF[y1]));
    QF[2][i] = ((((uint32_t)multEF[y2]) << 24) | (((uint32_t)y2) << 16) |
                (((uint32_t)multEF[y2]) << 8) | ((uint32_t)mult5B[y2]));
    QF[3][i] =
        ((((uint32_t)mult5B[y3]) << 24) | (((uint32_t)multEF[y3]) << 16) |
         (((uint32_t)y3) << 8) | ((uint32_t)mult5B[y3]));
  }
}

// Fully keyed h (aka g) function.
#define fkh(X) (S[0][b0(X)] ^ S[1][b1(X)] ^ S[2][b2(X)] ^ S[3][b3(X)])

// One encryption round.
#define ENC_ROUND(R0, R1, R2, R3, round)                  \
  T0 = fkh(R0);                                           \
  T1 = fkh(ROL((R1), 8));                                 \
  (R2) = ROR((R2) ^ (T1 + T0 + K[(2 * (round)) + 8]), 1); \
  (R3) = ROL((R3), 1) ^ (2 * T1 + T0 + K[(2 * (round)) + 9]);

static void encrypt_block(const uint32_t K[40],
                          uint32_t S[4][256],
                          uint8_t PT[TWOFISH_BLOCK_BYTES]) {
  // Load/byteswap/whiten input.
  uint32_t R3 = K[3] ^ BSWAP(((uint32_t*)PT)[3]);
  uint32_t R2 = K[2] ^ BSWAP(((uint32_t*)PT)[2]);
  uint32_t R1 = K[1] ^ BSWAP(((uint32_t*)PT)[1]);
  uint32_t R0 = K[0] ^ BSWAP(((uint32_t*)PT)[0]);

  uint32_t T0;  // NOLINT(cppcoreguidelines-init-variables): Macro magic.
  uint32_t T1;  // NOLINT(cppcoreguidelines-init-variables): Macro magic.
  ENC_ROUND(R0, R1, R2, R3, 0);
  ENC_ROUND(R2, R3, R0, R1, 1);
  ENC_ROUND(R0, R1, R2, R3, 2);
  ENC_ROUND(R2, R3, R0, R1, 3);
  ENC_ROUND(R0, R1, R2, R3, 4);
  ENC_ROUND(R2, R3, R0, R1, 5);
  ENC_ROUND(R0, R1, R2, R3, 6);
  ENC_ROUND(R2, R3, R0, R1, 7);
  ENC_ROUND(R0, R1, R2, R3, 8);
  ENC_ROUND(R2, R3, R0, R1, 9);
  ENC_ROUND(R0, R1, R2, R3, 10);
  ENC_ROUND(R2, R3, R0, R1, 11);
  ENC_ROUND(R0, R1, R2, R3, 12);
  ENC_ROUND(R2, R3, R0, R1, 13);
  ENC_ROUND(R0, R1, R2, R3, 14);
  ENC_ROUND(R2, R3, R0, R1, 15);

  // Load/byteswap/whiten output.
  ((uint32_t*)PT)[3] = BSWAP(R1 ^ K[7]);
  ((uint32_t*)PT)[2] = BSWAP(R0 ^ K[6]);
  ((uint32_t*)PT)[1] = BSWAP(R3 ^ K[5]);
  ((uint32_t*)PT)[0] = BSWAP(R2 ^ K[4]);
}

// One decryption round.
#define DEC_ROUND(R0, R1, R2, R3, round)                  \
  T0 = fkh(R0);                                           \
  T1 = fkh(ROL((R1), 8));                                 \
  (R2) = ROL((R2), 1) ^ (T0 + T1 + K[(2 * (round)) + 8]); \
  (R3) = ROR((R3) ^ (T0 + (2 * T1) + K[(2 * (round)) + 9]), 1);

static void decrypt_block(const uint32_t K[40],
                          uint32_t S[4][256],
                          uint8_t PT[TWOFISH_BLOCK_BYTES]) {
  // Load/byteswap/whiten input.
  uint32_t R3 = K[7] ^ BSWAP(((uint32_t*)PT)[3]);
  uint32_t R2 = K[6] ^ BSWAP(((uint32_t*)PT)[2]);
  uint32_t R1 = K[5] ^ BSWAP(((uint32_t*)PT)[1]);
  uint32_t R0 = K[4] ^ BSWAP(((uint32_t*)PT)[0]);

  uint32_t T0;  // NOLINT(cppcoreguidelines-init-variables): Macro magic.
  uint32_t T1;  // NOLINT(cppcoreguidelines-init-variables): Macro magic.
  DEC_ROUND(R0, R1, R2, R3, 15);
  DEC_ROUND(R2, R3, R0, R1, 14);
  DEC_ROUND(R0, R1, R2, R3, 13);
  DEC_ROUND(R2, R3, R0, R1, 12);
  DEC_ROUND(R0, R1, R2, R3, 11);
  DEC_ROUND(R2, R3, R0, R1, 10);
  DEC_ROUND(R0, R1, R2, R3, 9);
  DEC_ROUND(R2, R3, R0, R1, 8);
  DEC_ROUND(R0, R1, R2, R3, 7);
  DEC_ROUND(R2, R3, R0, R1, 6);
  DEC_ROUND(R0, R1, R2, R3, 5);
  DEC_ROUND(R2, R3, R0, R1, 4);
  DEC_ROUND(R0, R1, R2, R3, 3);
  DEC_ROUND(R2, R3, R0, R1, 2);
  DEC_ROUND(R0, R1, R2, R3, 1);
  DEC_ROUND(R2, R3, R0, R1, 0);

  // Load/byteswap/whiten output.
  ((uint32_t*)PT)[3] = BSWAP(R1 ^ K[3]);
  ((uint32_t*)PT)[2] = BSWAP(R0 ^ K[2]);
  ((uint32_t*)PT)[1] = BSWAP(R3 ^ K[1]);
  ((uint32_t*)PT)[0] = BSWAP(R2 ^ K[0]);
}

static void xor_block_and_iv(const uint8_t* iv, uint8_t* buf) {
#if defined(USE_X86_SSE)
  // TWOFISH_BLOCK_BYTES = 16, i.e. 128 bits.
  _mm_storeu_si128((__m128i*)buf,
                   _mm_xor_si128(_mm_loadu_si128((__m128i*)buf),
                                 _mm_loadu_si128((const __m128i*)iv)));
#else
  for (int i = 0; i < TWOFISH_BLOCK_BYTES; ++i) {
    buf[i] ^= iv[i];
  }
#endif
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void twofish_256_init(twofish256_ctx_t* ctx,
                      const uint8_t key[TWOFISH_KEY_BYTES],
                      const uint8_t iv[TWOFISH_IV_BYTES]) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(key != NULL);
  PRECONDITION(iv != NULL);

  // Copy IV.
  memcpy(ctx->iv, iv, TWOFISH_BLOCK_BYTES);

  // Prepare keys.
  uint32_t KS[4];
  key_sched_256(key, KS, ctx->K);
  create_full_key_256(KS, ctx->QF);
}

void twofish_256_cbc_encrypt_buffer(twofish256_ctx_t* ctx,
                                    uint8_t* buf,
                                    const size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (TWOFISH_BLOCK_BYTES - 1)) == 0U);

  const uint8_t* iv = ctx->iv;
  for (size_t i = TWOFISH_BLOCK_BYTES - 1; i < length;
       i += TWOFISH_BLOCK_BYTES) {
    xor_block_and_iv(iv, buf);
    encrypt_block(ctx->K, ctx->QF, buf);
    iv = buf;
    buf += TWOFISH_BLOCK_BYTES;
  }
  memcpy(ctx->iv, iv, TWOFISH_BLOCK_BYTES);
}

void twofish_256_cbc_decrypt_buffer(twofish256_ctx_t* ctx,
                                    uint8_t* buf,
                                    const size_t length) {
  PRECONDITION(ctx != NULL);
  PRECONDITION(buf != NULL);
  PRECONDITION((length & (TWOFISH_BLOCK_BYTES - 1)) == 0U);

  uint8_t tmp_iv[TWOFISH_BLOCK_BYTES];
  for (size_t i = TWOFISH_BLOCK_BYTES - 1; i < length;
       i += TWOFISH_BLOCK_BYTES) {
    memcpy(tmp_iv, buf, TWOFISH_BLOCK_BYTES);
    decrypt_block(ctx->K, ctx->QF, buf);
    xor_block_and_iv(ctx->iv, buf);
    memcpy(ctx->iv, tmp_iv, TWOFISH_BLOCK_BYTES);
    buf += TWOFISH_BLOCK_BYTES;
  }
}
