// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#ifndef U128_H_
#define U128_H_

#include <stdint.h>

#if defined(_MSC_VER)
#include <intrin.h>
#endif

// Use a true 128-bit integer type on compilers that support it, and fall back
// to a struct of two 64-bit integers for other compilers.
#if (defined(__GNUC__) || defined(__clang__)) && defined(__SIZEOF_INT128__) && \
    (__SIZEOF_INT128__ == 16)
#if defined(__GNUC__)
// __int128 is non-standard, so we need to turn off pedantic warnings.
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#endif
typedef unsigned __int128 u128;
#if defined(__GNUC__)
#pragma GCC diagnostic pop
#endif
#define HAS_INT128
#else
typedef struct {
  uint64_t l;
  uint64_t h;
} u128;
#endif

static inline u128 to_u128(uint64_t ah, uint64_t al) {
#if defined(HAS_INT128)
  return (((u128)ah) << 64) | (u128)al;
#else
  u128 r;
  r.l = al;
  r.h = ah;
  return r;
#endif
}

static inline uint64_t hi64(const u128 x) {
#if defined(HAS_INT128)
  return (uint64_t)(x >> 64);
#else
  return x.h;
#endif
}

static inline uint64_t lo64(const u128 x) {
#if defined(HAS_INT128)
  return (uint64_t)x;
#else
  return x.l;
#endif
}

static inline u128 and128(const u128 a, const u128 b) {
#if defined(HAS_INT128)
  return a & b;
#else
  u128 c;
  c.l = a.l & b.l;
  c.h = a.h & b.h;
  return c;
#endif
}

static inline u128 or128(const u128 a, const u128 b) {
#if defined(HAS_INT128)
  return a | b;
#else
  u128 c;
  c.l = a.l | b.l;
  c.h = a.h | b.h;
  return c;
#endif
}

static inline u128 xor128(const u128 a, const u128 b) {
#if defined(HAS_INT128)
  return a ^ b;
#else
  u128 c;
  c.l = a.l ^ b.l;
  c.h = a.h ^ b.h;
  return c;
#endif
}

static inline u128 add128(const u128 a, const u128 b) {
#if defined(HAS_INT128)
  return a + b;
#else
  u128 c;
  c.l = a.l + b.l;
  c.h = a.h + b.h + (c.l < a.l ? 1U : 0U);
  return c;
#endif
}

static inline u128 mul128(const u128 a, const u128 b) {
#if defined(HAS_INT128)
  return a * b;
#elif defined(_MSC_VER) && defined(_M_X64)
  // For MSVC we can use the _umul128() intrinsic (64x64->128). This produces
  // roughly the same code as GCC/Clang with direct __int128 support.
  uint64_t ch;
  uint64_t cl = _umul128(a.l, b.l, &ch);
  ch += (a.l * b.h) + (a.h * b.l);
  return to_u128(ch, cl);
#else
  // In the generic, portable C solution, we have to do crosswise 32-bit
  // multiplications of the lower 64 bits in combinaton with crosswise 64-bit
  // multiplications. Inspired by:
  // https://www.codeproject.com/Tips/618570/UInt-Multiplication-Squaring
  const uint64_t al32h = a.l >> 32;
  const uint64_t al32l = (a.l & 0xffffffffU);
  const uint64_t bl32h = b.l >> 32;
  const uint64_t bl32l = (b.l & 0xffffffffU);

  uint64_t t = (al32l * bl32l);
  const uint64_t wl = (t & 0xffffffffU);
  t = (al32h * bl32l) + (t >> 32);
  const uint64_t wh = (t >> 32);
  t = (al32l * bl32h) + (t & 0xffffffffU);

  u128 c;
  c.l = (t << 32) + wl;
  c.h = (al32h * bl32h) + (a.h * b.l) + (a.l * b.h) + (t >> 32) + wh;
  return c;
#endif
}

#endif  // U128_H_
