// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "zkdf.h"

#include <stddef.h>
#include <string.h>

#include "../debug.h"
#include "../error.h"
#include "../memory.h"
#include "sha2.h"
#include "u128.h"
#include "xoshiro.h"

// The number of bytes in the garble buffer.
// Note: The size of the buffer is tuned so that it fits in the L2 cache of most
// CPU:s. Since the cache latency of GPU:s is usually at least an order of
// magnitude higher than that of CPU:s, this prevents efficient implementations
// on GPU:s. It also makes ASIC/FPGA solutions costly, as each KDF core would
// need substantial amounts of RAM.
#define GARBLE_BYTES (256 * 1024)  // Number of bytes to generate (must be 2^k)

// Tune the number of shuffle operations per iteration.
// Note: For balanced operation, SHUFFLE_ITERS x SHUFFLE_SUB_STEPS should be
// close to GARBLE_BYTES.
#define SHUFFLE_ITERS 13267   // Number of shuffle indirection iterations
#define SHUFFLE_SUB_STEPS 19  // Number of shuffle indirection sub-steps

//------------------------------------------------------------------------------
// Helpers.
//------------------------------------------------------------------------------

// Bitwise XOR all bytes of an unsigned 128-bit integer into a single byte.
static inline uint8_t xor_u128_bytes(const u128 x) {
  const uint64_t xh = hi64(x);
  uint32_t yh = ((uint32_t)xh) ^ (uint32_t)(xh >> 32);
  yh ^= yh >> 16;
  yh ^= yh >> 8;

  const uint64_t xl = lo64(x);
  uint32_t yl = ((uint32_t)xl) ^ (uint32_t)(xl >> 32);
  yl ^= yl >> 16;
  yl ^= yl >> 8;

  return (uint8_t)(yh ^ yl);
}

static inline uint32_t ror32(const uint32_t x, const unsigned r) {
  return (x >> r) | (x << (32U - r));
}

static inline uint8_t ror8(const uint8_t x, const unsigned r) {
  return (x >> r) | (x << (8U - r));
}

//------------------------------------------------------------------------------
// Garble logic.
//------------------------------------------------------------------------------

static inline void generate_garble(const uint8_t* passphrase,
                                   const uint32_t len,
                                   uint32_t* src_idx_p,
                                   xoshiro256_state_t* state,
                                   uint8_t* garble) {
  // We XOR the repeated passphrase and a random sequencey into the buffer.
  uint32_t src_idx = *src_idx_p;
  u128 rnd_sum = to_u128(0U, 1U);
  for (int i = 0; i < GARBLE_BYTES; ++i) {
    // Get the next byte from the passhprase.
    const uint8_t byte_from_pass = passphrase[src_idx];
    src_idx = (src_idx + 1U) % len;

    // Use the PRNG to generate 4 x 64 random bits.
    const uint64_t r1h = xoshiro256ss(state);
    const uint64_t r1l = xoshiro256ss(state);
    const uint64_t r2h = xoshiro256ss(state);
    const uint64_t r2l = xoshiro256ss(state);

    // Calculate rnd_sum = r2 * rnd_sum + (r1 ^ byte_from_pass), using 128-bit
    // arithmetic.
    const u128 r1_xor_pass = to_u128(r1h, r1l ^ ((uint64_t)byte_from_pass));
    const u128 r2 = to_u128(r2h, r2l);
    rnd_sum = add128(mul128(r2, rnd_sum), r1_xor_pass);

    // Convert the 128-bit sum into a single byte.
    const uint8_t rnd_byte = xor_u128_bytes(rnd_sum);

    // Mix the byte into the garble buffer.
    const uint8_t garble_byte = garble[i];
    garble[i] = ror8(garble_byte, 7) ^ rnd_byte;
  }

  *src_idx_p = src_idx;
}

//------------------------------------------------------------------------------
// Shuffle logic.
//------------------------------------------------------------------------------

// Byte offsets relative to the base index when reading/writing "garble words".
#define GWORD_BYTE1_OFFS 97
#define GWORD_BYTE2_OFFS 179

// Load an unsigned "word", 24 bits wide, assembled from three bytes at
// different locations in the buffer, starting at IDX.
static inline uint32_t get_garble_word(const uint8_t* garble,
                                       const uint32_t idx) {
  return ((uint32_t)garble[idx]) |
         (((uint32_t)garble[(idx + GWORD_BYTE1_OFFS) & (GARBLE_BYTES - 1)])
          << 8) |
         (((uint32_t)garble[(idx + GWORD_BYTE2_OFFS) & (GARBLE_BYTES - 1)])
          << 16);
}

// Store an unsigned "word", 24 bits wide, into three bytes at different
// locations in the buffer, starting at IDX.
static inline void set_garble_word(uint8_t* garble,
                                   const uint32_t idx,
                                   const uint32_t word) {
  garble[idx] = (uint8_t)word;
  garble[(idx + GWORD_BYTE1_OFFS) & (GARBLE_BYTES - 1)] = (uint8_t)(word >> 8);
  garble[(idx + GWORD_BYTE2_OFFS) & (GARBLE_BYTES - 1)] = (uint8_t)(word >> 16);
}

static inline void shuffle_garble(xoshiro256_state_t* state, uint8_t* garble) {
  uint32_t prev_word = get_garble_word(garble, GARBLE_BYTES - 1);
  uint32_t idx = prev_word & (GARBLE_BYTES - 1);

  // Walk the garble buffer, using the buffer contents as indices.
  for (int i = 0; i < SHUFFLE_ITERS; ++i) {
    // Get a random mutator for the next sub-steps.
    uint32_t mutator = (uint32_t)xoshiro256ss(state);

    // Do a few sub-steps that use a common mutator.
    for (int j = 0; j < SHUFFLE_SUB_STEPS; ++j) {
      // Read a new word from the current index.
      const uint32_t word = get_garble_word(garble, idx);

      // Rotate the mutator.
      // Note: We introduce a data load dependency in order to penalize machines
      // with high cache/memory latencies (e.g. GPU:s).
      const unsigned shift = 1U + (((word + 0x123456U) >> 15) & 7U);
      mutator = ror32(mutator, shift);

      // Mutate the previous word.
      prev_word ^= mutator;

      // Write the mutated previous word to the current index.
      set_garble_word(garble, idx, prev_word);

      // Calculate the next index based on the current word.
      idx = (idx + word) & (GARBLE_BYTES - 1);

      prev_word = word;
    }
  }
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

void zkdf_make_key(const char* passphrase,
                   const uint8_t salt[ZKDF_SALT_BYTES],
                   int num_passes,
                   uint8_t key[ZKDF_KEY_BYTES]) {
  PRECONDITION(passphrase != NULL);
  PRECONDITION(salt != NULL);
  PRECONDITION(num_passes >= 1);
  PRECONDITION(key != NULL);

  DBG_DUMP_HEX(salt, ZKDF_SALT_BYTES);

  // Initialize the xoshiro state from the salt.
  xoshiro256_state_t state;
  xoshiro256_init(salt, &state);

  // Allocate memory for the garble buffer.
  // Note: secure_alloc() is guaranteed to clears all bytes of the garble buffer
  // to zero, so the first generate_garble() pass is well defined.
  uint8_t* garble = (uint8_t*)secure_alloc(GARBLE_BYTES);

  // Get the length of the passphrase.
  // Note: We intentionally include the null termination byte, in order to get a
  // separator token between repeated passphrase strings (generate_garble()
  // repeats the passphrase across the garble buffer). That way we can
  // differentiate between passphrases "foo" and "foofoo", for instance.
  const uint32_t len = (uint32_t)strlen(passphrase) + 1U;

  // Fill out the garble buffer with a unique random number sequence for the
  // given passphrase and salt.
  uint32_t src_idx = 0U;
  for (int i = 0; i < num_passes; ++i) {
    generate_garble((const uint8_t*)passphrase, len, &src_idx, &state, garble);
    shuffle_garble(&state, garble);
  }

  // Generate the key from the garble buffer using a SHA2-256 hash.
  sha2_256(garble, GARBLE_BYTES, key);
  DBG_DUMP_HEX(key, ZKDF_KEY_BYTES);

  // Free the garble buffer.
  secure_free(garble);
}
