// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Error handling.
//
// Most unexpected errors are treated as unrecoverable exceptions that terminate
// the program.
//------------------------------------------------------------------------------

#ifndef ERROR_H_
#define ERROR_H_

#include <stdio.h>
#include <stdlib.h>

#if defined(__FILE_NAME__)
// At least Clang 9+ and GCC 12+ provide __FILE_NAME__, which excludes the
// full path of the file.
#define PRECONDITION(cond)                                 \
  do {                                                     \
    if (!(cond)) {                                         \
      print_precondition_failed_(__FILE_NAME__, __LINE__); \
      exit(1);                                             \
    }                                                      \
  } while (0)
#else
// Fall back to the C standard __FILE__ macro, which may include the full
// path (which we trim out at runtime).
#define PRECONDITION(cond)                            \
  do {                                                \
    if (!(cond)) {                                    \
      print_precondition_failed_(__FILE__, __LINE__); \
      exit(1);                                        \
    }                                                 \
  } while (0)
#endif

void print_precondition_failed_(const char* file, int line);

#define FATAL_ERROR(msg)                    \
  do {                                      \
    fflush(stdout);                         \
    fprintf(stderr, "***Error: %s\n", msg); \
    exit(1);                                \
  } while (0)

// TODO: Make a more portable solution for variadic argument handling.
#if defined(__clang__)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wgnu-zero-variadic-macro-arguments"
#endif
// NOLINTBEGIN(clang-diagnostic-gnu-zero-variadic-macro-arguments)
#define FATAL_ERRORF(fmt, ...)                             \
  do {                                                     \
    fflush(stdout);                                        \
    fprintf(stderr, "***Error: " fmt "\n", ##__VA_ARGS__); \
    exit(1);                                               \
  } while (0)
// NOLINTEND(clang-diagnostic-gnu-zero-variadic-macro-arguments)
#if defined(__clang__)
#pragma clang diagnostic pop
#endif

#endif  // ERROR_H_
