// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "fileutils.h"

#include "crypto/entropy.h"
#include "crypto/sha2.h"
#include "error.h"
#include "memory.h"
#include "stringutils.h"
#include <stdint.h>
#include <string.h>

#if defined(_WIN32)
#undef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#undef NOMINMAX
#define NOMINMAX
#include <windows.h>
#undef ERROR
#undef log
#include <accctrl.h>
#include <aclapi.h>
#include <fcntl.h>
#include <io.h>
#include <shlwapi.h>
#include <stdio.h>
#include <sys/types.h>
// types.h must be included before stat.h according to MS.
#include <sys/stat.h>
#else
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#endif

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

#if defined(_WIN32)
#define DIR_SEPARATOR '\\'
#define PATH_SEPARATOR ';'
#else
#define DIR_SEPARATOR '/'
#define PATH_SEPARATOR ':'
#endif
#define NUM_TEMPFILE_HEX_CHARS 32  // Must be a multiple of 2.

static char to_hex(const uint8_t b) {
  if (b < 10U) {
    return (char)(((unsigned)b) + (unsigned)'0');
  }
  return (char)(((unsigned)b) - 10U + (unsigned)'a');
}

#if defined(_WIN32)
static bool get_sid_for_current_user(PSID* user_sid) {
  // Get the current process token.
  HANDLE process_token = NULL;
  if (!OpenProcessToken(
          GetCurrentProcess(), TOKEN_READ | TOKEN_QUERY, &process_token)) {
    return false;
  }

  // Get the user token from the current process.
  DWORD token_info_size = 0;
  if (!GetTokenInformation(
          process_token, TokenUser, NULL, 0, &token_info_size)) {
    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
      return false;
    }
  }
  PTOKEN_USER user = (PTOKEN_USER)LocalAlloc(GPTR, token_info_size);
  if (user == NULL) {
    return false;
  }
  if (!GetTokenInformation(
          process_token, TokenUser, user, token_info_size, &token_info_size)) {
    LocalFree(user);
    return false;
  }

  // Make a copy of the user SID.
  DWORD sid_len = GetLengthSid(user->User.Sid);
  *user_sid = (SID*)LocalAlloc(GPTR, sid_len);
  CopySid(sid_len, *user_sid, user->User.Sid);
  LocalFree(user);

  return true;
}
#endif

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

bool file_or_dir_exists(const char* path) {
  PRECONDITION(path != NULL);

#if defined(_WIN32)
  struct __stat64 s;
  return _stat64(path, &s) == 0;
#else
  struct stat s;
  return stat(path, &s) == 0;
#endif
}

bool file_exists(const char* path) {
  PRECONDITION(path != NULL);

#if defined(_WIN32)
  struct __stat64 s;
  if (_stat64(path, &s) != 0) {
    return false;
  }
  return (s.st_mode & _S_IFREG) != 0;
#else
  struct stat s;
  if (stat(path, &s) != 0) {
    return false;
  }
  return S_ISREG(s.st_mode);
#endif
}

bool dir_exists(const char* path) {
  PRECONDITION(path != NULL);

#if defined(_WIN32)
  struct __stat64 s;
  if (_stat64(path, &s) != 0) {
    return false;
  }
  return (s.st_mode & _S_IFDIR) != 0;
#else
  struct stat s;
  if (stat(path, &s) != 0) {
    return false;
  }
  return S_ISDIR(s.st_mode);
#endif
}

bool delete_file(const char* path) {
  PRECONDITION(path != NULL);

#if defined(_WIN32)
  // Use DeleteFileW() to delete the file.
  uint16_t* path_utf16 = utf8_to_utf16_alloc(path);
  const bool success = (bool)DeleteFileW(path_utf16);
  free(path_utf16);
  return success;
#else
  return unlink(path) == 0;
#endif
}

FILE* create_private_file(const char* path) {
  PRECONDITION(path != NULL);

#if defined(_WIN32)
  PSID user_sid = NULL;
  PACL acl = NULL;
  PSECURITY_DESCRIPTOR sd = NULL;
  FILE* f = NULL;

  // Get the SID for the current user.
  if (!get_sid_for_current_user(&user_sid)) {
    fprintf(stderr, "***Error: Unable to get SID for current user\n");
    goto cleanup;
  }

  // Create an ACL.
  EXPLICIT_ACCESSA ea;
  ZeroMemory(&ea, sizeof(EXPLICIT_ACCESSA));
  ea.grfAccessPermissions =
      ACCESS_SYSTEM_SECURITY | READ_CONTROL | WRITE_DAC | GENERIC_ALL;
  ea.grfAccessMode = GRANT_ACCESS;
  ea.grfInheritance = NO_INHERITANCE;
  ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
  ea.Trustee.ptstrName = (char*)user_sid;
  if (SetEntriesInAclA(1, &ea, NULL, &acl) != ERROR_SUCCESS) {
    fprintf(stderr, "***Error: Unable to create ACL\n");
    goto cleanup;
  }

  // Initialize a security descriptor.
  sd = (PSECURITY_DESCRIPTOR)LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
  if (sd == NULL ||
      !InitializeSecurityDescriptor(sd, SECURITY_DESCRIPTOR_REVISION)) {
    fprintf(stderr, "***Error: Unable to initialize security descriptor\n");
    goto cleanup;
  }

  // Add the ACL to the security descriptor.
  if (!SetSecurityDescriptorDacl(sd, TRUE, acl, FALSE)) {
    fprintf(stderr, "***Error: Unable to set security descriptor\n");
    goto cleanup;
  }

  // Initialize a security attributes structure.
  SECURITY_ATTRIBUTES sa;
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = sd;
  sa.bInheritHandle = FALSE;

  // Delete the file, if any, so that we do not inherit permissions.
  // TODO: Can this be done as part of CreateFile to make it atomic?
  if (file_exists(path)) {
    if (!delete_file(path)) {
      fprintf(stderr, "***Error: Unable to delete existing file: %s\n", path);
      goto cleanup;
    }
  }

  // Open the file.
  // TODO: Use CreateFileW().
  HANDLE file_handle = CreateFileA(
      path, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (file_handle != INVALID_HANDLE_VALUE) {
    int nHandle = _open_osfhandle((intptr_t)file_handle, _O_RDWR);
    if (nHandle != -1) {
      f = _fdopen(nHandle, "wb");
    }
    if (f == NULL) {
      CloseHandle(file_handle);
    }
  }

cleanup:
  // Free allocated resources.
  if (sd != NULL) {
    LocalFree(sd);
  }
  if (acl != NULL) {
    LocalFree(acl);
  }
  if (user_sid != NULL) {
    LocalFree(user_sid);
  }
#else
  // We use umask() to ensure that the output file is not accessible by other
  // users.
  mode_t old_mode = umask(S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

  // Open the file for writing.
  FILE* f = fopen(path, "wb");

  // Restore the old umask.
  (void)umask(old_mode);

  // Just in case the file already existed and had a different file mode.
  if (f != NULL) {
    int fd = fileno(f);
    if (fd >= 0) {
      (void)fchmod(fd, S_IRUSR | S_IWUSR);
    }
  }
#endif

  return f;
}

void make_safe_tmp_filename(char* path,
                            const size_t max_path_len,
                            const char* file_ext) {
  PRECONDITION(path != NULL);
  PRECONDITION(max_path_len > 0U);
  PRECONDITION(file_ext != NULL);

  // Get the temporary file directory. Prefer a folder with volatile content
  // that only the current user has access to.
  const char* temp_dir = NULL;
#if defined(_WIN32)
  // On Windows we use GetTempPath() to query the temporary dir.
  char temp_path_buf[1024];
  DWORD len = GetTempPathA(sizeof(temp_path_buf), temp_path_buf);
  if (len > 0 && ((size_t)(len + 1)) < sizeof(temp_path_buf)) {
    temp_dir = &temp_path_buf[0];
  }
#else
  // First try $XDG_RUNTIME_DIR - it's guaranteed to be private (when defined
  // correctly and following the XDG standard).
  temp_dir = getenv("XDG_RUNTIME_DIR");

  if (temp_dir == NULL || !dir_exists(temp_dir)) {
    // Try $TMPDIR.
    temp_dir = getenv("TMPDIR");
  }

  if (temp_dir == NULL || !dir_exists(temp_dir)) {
    // Try "/tmp".
    temp_dir = "/tmp";
  }
#endif

  // If we could not find a temporary directory - bail now.
  if (temp_dir == NULL || !dir_exists(temp_dir)) {
    FATAL_ERROR("Could not find a vaild temporary directory");
  }

  // Detrmine the total size of the complete path.
  const size_t dir_name_len = strlen(temp_dir);
  const size_t ext_len = strlen(file_ext);
  size_t path_len = dir_name_len + NUM_TEMPFILE_HEX_CHARS + ext_len + 1;
  const int add_dir_sep = (temp_dir[dir_name_len - 1] != DIR_SEPARATOR);
  if (add_dir_sep) {
    path_len += 1;
  }
  if (path_len > max_path_len) {
    FATAL_ERROR("Not enough room for the path");
  }

  // Fill out all the parts of the path except for the random name part.
  memcpy(path, temp_dir, dir_name_len);
  size_t file_part_start = dir_name_len;
  if (add_dir_sep) {
    path[file_part_start++] = DIR_SEPARATOR;
  }
  memcpy(&path[file_part_start + NUM_TEMPFILE_HEX_CHARS], file_ext, ext_len);
  path[path_len - 1] = 0;
  char* file_part = &path[file_part_start];

  // Try random names until we get a file that does not already exist.
  bool success = false;
  for (int n = 0; n < 100; ++n) {
    // Use system entropy to generate a random file name.
    uint8_t rnd[NUM_TEMPFILE_HEX_CHARS / 2];
    make_entropy(rnd, sizeof(rnd));
    for (int i = 0; i < NUM_TEMPFILE_HEX_CHARS / 2; ++i) {
      const uint8_t b = rnd[i];
      file_part[(i * 2) + 0] = to_hex(b >> 4);
      file_part[(i * 2) + 1] = to_hex(b & 15U);
    }

    // If the file does not already exist we are done (this should be the common
    // case).
    if (!file_or_dir_exists(path)) {
      success = true;
      break;
    }
  }

  // This should never happen in practice...
  if (!success) {
    FATAL_ERROR("Failed to create a unique temporary file name");
  }
}

bool is_exe_in_path(const char* program) {
  PRECONDITION(program != NULL);

#if defined(_WIN32)
  char path[MAX_PATH + 1];

  // Check if the program is in the CWD.
  char cwd[MAX_PATH + 1];
  const DWORD cwd_buf_size = (DWORD)sizeof(cwd);
  const DWORD cwd_len = GetCurrentDirectoryA(cwd_buf_size, cwd);
  if (cwd_len > 0U && (cwd_len + 1U) < cwd_buf_size) {
    const int path_len = snprintf(path, sizeof(path), "%s\\%s", cwd, program);
    if (path_len > 0 && (path_len + 1) < (int)sizeof(path)) {
      if (file_exists(path)) {
        return true;
      }
    }
  }

  // Use PathFindOnPath() to find the executable.
  if (string_copy(program, path, sizeof(path))) {
    if (PathFindOnPathA(path, NULL)) {
      return true;
    }
  }

  return false;
#else
  // Get the PATH environment variable.
  const char* path_env = getenv("PATH");
  if (path_env == NULL) {
    return false;
  }

  // Walk the PATH string, splitting it at each path separator.
  const size_t path_env_len = strlen(path_env);
  const size_t program_len = strlen(program);
  for (size_t pos = 0U; pos < path_env_len;) {
    // Find the next path separator.
    size_t end = pos;
    for (; end < path_env_len && path_env[end] != PATH_SEPARATOR; ++end) {
    }
    size_t path_len = end - pos;

    if (path_len > 0U && (path_len + 1U + program_len) < 1024U) {
      // Copy the path to a new string.
      char path[1024];
      memcpy(path, &path_env[pos], path_len);

      // Append the program name.
      path[path_len] = DIR_SEPARATOR;
      memcpy(&path[path_len + 1U], program, program_len);
      path[path_len + 1U + program_len] = 0;

      // Does the program file exist?
      if (file_exists(path)) {
        return true;
      }

      pos += path_len + 1U;
    }
  }

  return false;
#endif
}

int run_command(const char* cmd) {
#if defined(_WIN32)
  // Initialize startup info.
  STARTUPINFOW si;
  ZeroMemory(&si, sizeof(si));
  si.cb = sizeof(si);

  // Run the command as a new process.
  int exit_code = 1;
  PROCESS_INFORMATION pi;
  uint16_t* cmd_utf16 = utf8_to_utf16_alloc(cmd);
  if (CreateProcessW(NULL, cmd_utf16, NULL, NULL, FALSE, 0, 0, 0, &si, &pi)) {
    // Wait for the process to finish.
    if (WaitForSingleObject(pi.hProcess, INFINITE) != WAIT_FAILED) {
      // Get the exit code.
      DWORD exit_code_dw = 1;
      if (GetExitCodeProcess(pi.hProcess, &exit_code_dw)) {
        exit_code = (int)exit_code_dw;
      }
    }
    CloseHandle(pi.hThread);
    CloseHandle(pi.hProcess);
  }
  free(cmd_utf16);

  return exit_code;
#else
  return system(cmd);
#endif
}

void compute_file_sha256(const char* path, uint8_t hash[SHA256_HASH_BYTES]) {
  // Open the file.
  FILE* f = fopen(path, "rb");
  if (f == NULL) {
    FATAL_ERRORF("Unable to hash file %s", path);
  }

  sha2_256_ctx_t ctx;
  sha2_256_init(&ctx);

  // Read and hash blocks of 1 KiB at a time.
  uint8_t buf[1024];
  while (true) {
    size_t bytes_read = fread(buf, 1, sizeof(buf), f);
    sha2_256_append(&ctx, buf, bytes_read);
    if (bytes_read < sizeof(buf)) {
      if (ferror(f)) {
        FATAL_ERRORF("Error occurred when reading %s", path);
      }
      break;
    }
  }

  // Get the final hash.
  sha2_256_finalize(&ctx, hash);

  // Close the file.
  fclose(f);
}
