// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "memory.h"

#include "error.h"

#if defined(_WIN32)
// For Windows.
#undef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#undef NOMINMAX
#define NOMINMAX
#include <windows.h>
#undef ERROR
#undef log
#else
#include <string.h>
#endif

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

#if !defined(_WIN32)
// We use a volatile static pointer to memset(), so that the compiler does not
// optimize away the call.
typedef void* (*memset_fun_t)(void*, int, size_t);
static volatile memset_fun_t s_memset_fun = memset;
#endif

//------------------------------------------------------------------------------
// Public methods.
//------------------------------------------------------------------------------

void secure_zero_memory(void* ptr, const size_t cnt) {
  PRECONDITION(ptr != NULL);

#if defined(_WIN32)
  SecureZeroMemory(ptr, cnt);
#else
  (void)s_memset_fun(ptr, 0, cnt);
#endif
}

void* secure_alloc(size_t size) {
  PRECONDITION(size > 0U);

  // Allocate memory for the buffer + our meta data.
  const size_t full_size = sizeof(size_t) + size;
  void* base_ptr = malloc(full_size);
  if (base_ptr == NULL) {
    // This should really never happen on most systems (e.g. Linux).
    FATAL_ERROR("Out of memory");
  }

  // Remember the size.
  *((size_t*)base_ptr) = size;
  void* ptr = (void*)&((size_t*)base_ptr)[1];

  // Always start with a zero-initialized buffer.
  secure_zero_memory(ptr, size);

  return ptr;
}

void secure_free(void* ptr) {
  PRECONDITION(ptr != NULL);

  // Retrieve the size.
  void* base_ptr = (void*)&((size_t*)ptr)[-1];
  size_t* size = (size_t*)base_ptr;

  // Ensure that the memory is cleared before freeing it (hide any secrets).
  secure_zero_memory(ptr, *size);
  *size = 0U;

  // Free the memory.
  free(base_ptr);
}
