// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#if !defined(_WIN32)
// Define this macro to get struct sigaction from signal.h.
// NOLINTNEXTLINE(bugprone-reserved-identifier,cert-dcl37-c,cert-dcl51-cpp)
#define _XOPEN_SOURCE 700
#endif

#include "passphrase.h"

#include "arguments.h"
#include "error.h"
#include "memory.h"
#include "passphrase_common.h"
#include "passphrase_gtk.h"
#include "stringutils.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Detect if we're building for macos.
#ifdef __APPLE__
#include "TargetConditionals.h"  // May define TARGET_OS_MAC
#endif

// Select terminal prompt implementation.
#if defined(_WIN32)
// For Windows.
#undef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#undef NOMINMAX
#define NOMINMAX
#include <windows.h>
#undef ERROR
#undef log
#include <io.h>
#define USE_WIN32_PROMPT
#else
// For POSIX systems.
#include <fcntl.h>
#include <paths.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>
#define USE_POSIX_PROMPT
#endif

// Select GUI dialog implementation.
#if defined(_WIN32)
#include <wincred.h>
#define USE_WIN32_DIALOG
#elif defined(TARGET_OS_MAC)
#define USE_COCOA_DIALOG
#else
#include <sys/wait.h>  // For WEXITSTATUS
#define USE_ASKPASS_DIALOG
#endif

//------------------------------------------------------------------------------
// Private.
//------------------------------------------------------------------------------

static void terminate_str_at_newline(char* str, const size_t max_size) {
  // Terminate the string at the first newline character (if any).
  str[max_size - 1] = 0;
  for (size_t i = 0U; i < max_size && str[i] != 0U; ++i) {
    if (str[i] == '\n' || str[i] == '\r') {
      str[i] = 0;
      break;
    }
  }
}

#if defined(USE_POSIX_PROMPT)
// Signals that we need to handle / silence during user input.
#define NUM_HANDLE_SIGNALS 9
static const int HANDLE_SIGNALS[NUM_HANDLE_SIGNALS] = {SIGALRM,
                                                       SIGHUP,
                                                       SIGINT,
                                                       SIGPIPE,
                                                       SIGQUIT,
                                                       SIGTERM,
                                                       SIGTSTP,
                                                       SIGTTIN,
                                                       SIGTTOU};

static volatile sig_atomic_t s_caught_signals[NUM_HANDLE_SIGNALS];

static void signal_catcher(int signal_id) {
  for (int i = 0; i < NUM_HANDLE_SIGNALS; ++i) {
    if (signal_id == HANDLE_SIGNALS[i]) {
      s_caught_signals[i] = 1;
      break;
    }
  }
}
#endif  // USE_POSIX_PROMPT

static bool has_graphical_session(void) {
#if !(defined(_WIN32) || defined(__APPLE__))
  // Don't try anything if we don't have a DISPLAY.
  {
    char* display = getenv("DISPLAY");
    if (display == NULL || strlen(display) == 0) {
      return false;
    }
  }

  // In some situations we may be on a text mode TTY without a graphical display
  // (e.g. CTRL+ALT+F3 on most Linux distros), but the user has a graphical
  // session open in another TTY. We do not want to open the dialog in that
  // graphical session even if we could, as it would be hidden from the user.
  {
    const char* tty = ttyname(STDIN_FILENO);
    if (tty) {
      // Known TTY names for terminals without a graphical session:
      if (strncmp(tty, "/dev/tty", 8) == 0) {
        // Virtual text mode tty or serial port tty (mostly anything that isn't
        // a graphical /dev/pts* pseudo-terminal).
        return false;
      }
      if (strcmp(tty, "/dev/console") == 0) {
        // "System console" (e.g. FreeBSD single user).
        return false;
      }
    }
  }
#endif
  return true;
}

//------------------------------------------------------------------------------
// Implementations of different passphrase querying methods.
//------------------------------------------------------------------------------

#if defined(USE_WIN32_DIALOG)
static int get_from_win32_dialog(const char* prompt,
                                 char* passphrase,
                                 size_t max_size) {
  // Convert the prompt string to UTF-16.
  const size_t max_prompt_utf16_len = (numchars_utf8(prompt) * 2U) + 1U;
  uint16_t* prompt_utf16 =
      (uint16_t*)secure_alloc(max_prompt_utf16_len * sizeof(uint16_t));
  if (!utf8_to_utf16(prompt, prompt_utf16, max_prompt_utf16_len)) {
    secure_free(prompt_utf16);
    return PASSPHRASE_FAIL;
  }

  // Prepare dialog captions.
  CREDUI_INFOW cui;
  memset(&cui, 0, sizeof(cui));
  cui.cbSize = sizeof(cui);
  cui.pszCaptionText = L"ezcrypt";
  cui.pszMessageText = prompt_utf16;

  // Prepare username and password buffers.
  // NOTE: The username appears to be displayed no matter what, so we set it to
  // an empty string (" ") and disallow the user from changing it
  // (CREDUI_FLAGS_KEEP_USERNAME).
  uint16_t username[] = L" ";
  ULONG pass_utf16_size = (ULONG)((max_size < CREDUI_MAX_PASSWORD_LENGTH)
                                      ? max_size
                                      : CREDUI_MAX_PASSWORD_LENGTH);
  uint16_t* pass_utf16 =
      (uint16_t*)secure_alloc(((size_t)pass_utf16_size) * sizeof(uint16_t));

  // Show the dialog.
  BOOL save = FALSE;
  DWORD err = CredUIPromptForCredentialsW(
      &cui,        // CREDUI_INFO structure
      L"ezcrypt",  // Target for credentials
      NULL,        // Reserved
      0,           // Reason (none)
      username,    // User name
      sizeof(username) /
          sizeof(username[0]),  // Max number of char for user name
      pass_utf16,               // Password
      pass_utf16_size,          // Max number of char for password
      &save,                    // State of save check box
      CREDUI_FLAGS_GENERIC_CREDENTIALS | CREDUI_FLAGS_ALWAYS_SHOW_UI |
          CREDUI_FLAGS_DO_NOT_PERSIST | CREDUI_FLAGS_KEEP_USERNAME |
          CREDUI_FLAGS_PASSWORD_ONLY_OK | CREDUI_FLAGS_EXCLUDE_CERTIFICATES);

  // Handle the dialog result.
  int result = PASSPHRASE_FAIL;
  if (err == NO_ERROR) {
    if (utf16_to_utf8(pass_utf16, passphrase, max_size)) {
      result = PASSPHRASE_OK;
    } else {
      result = PASSPHRASE_FAIL;
    }
  } else {
    passphrase[0] = 0;
    result = (err == ERROR_CANCELLED) ? PASSPHRASE_ABORT : PASSPHRASE_FAIL;
  }

  // Free strings that we no longer need.
  secure_free(pass_utf16);
  secure_free(prompt_utf16);

  return result;
}
#endif  // USE_WIN32_DIALOG

#if defined(USE_ASKPASS_DIALOG)
static int get_from_askpass_tool(const char* program,
                                 const char* args_fmt_str,
                                 const char* prompt,
                                 char* passphrase,
                                 size_t max_size) {
  char cmd[1024];

  // Check if the tool exists.
  {
    const int cmd_len =
        snprintf(cmd, sizeof(cmd), "which %s > /dev/null 2>&1", program);
    if (cmd_len < 0 || (size_t)cmd_len >= sizeof(cmd)) {
      return PASSPHRASE_FAIL;
    }
  }
  if (system(cmd) != 0) {
    // The tool does not appear to be installed on this system.
    return PASSPHRASE_FAIL;
  }

  // Start the tool.
  {
    const int cmd_len =
        snprintf(cmd, sizeof(cmd), args_fmt_str, program, prompt);
    if (cmd_len < 0 || (size_t)cmd_len >= sizeof(cmd)) {
      return PASSPHRASE_FAIL;
    }
  }
  FILE* p = popen(cmd, "r");
  if (p == NULL) {
    return PASSPHRASE_FAIL;
  }

  // Read the passphrase from stdout via the pipe.
  if (fgets(passphrase, (int)max_size, p) == NULL) {
    passphrase[0] = 0;
  }
  const int exit_status = WEXITSTATUS(pclose(p));
  if (exit_status != 0) {
    // TODO: At this point we don't really know if the tool failed (e.g. due to
    // a missing X11 session) or if the user pressed the "Cancel" button in the
    // dialog (or similar). The former case should result in FAIL rather than
    // ABORT.
    return PASSPHRASE_ABORT;
  }
  terminate_str_at_newline(passphrase, max_size);

  return passphrase[0] != 0 ? PASSPHRASE_OK : PASSPHRASE_ABORT;
}
#endif  // USE_ASKPASS_DIALOG

static int get_from_dialog(const char* prompt,
                           char* passphrase,
                           size_t max_size) {
  if (!has_graphical_session()) {
    return PASSPHRASE_FAIL;
  }

  int result = PASSPHRASE_FAIL;
#if defined(USE_WIN32_DIALOG)
  if (result == PASSPHRASE_FAIL) {
    result = get_from_win32_dialog(prompt, passphrase, max_size);
  }
#endif
#if defined(USE_COCOA_DIALOG)
  if (result == PASSPHRASE_FAIL) {
    // TODO: Add a macos cocoa GUI (NSSecureTextField?)
    (void)prompt;
    (void)max_size;
    passphrase[0] = 0;
  }
#endif
#if defined(HAS_GTK3)
  if (result == PASSPHRASE_FAIL) {
    result = get_passphrase_from_gtk_dialog(prompt, passphrase, max_size);
  }
#endif
#if defined(USE_ASKPASS_DIALOG)
  // Try zenity (https://help.gnome.org/users/zenity/stable/).
  if (result == PASSPHRASE_FAIL) {
    result =
        get_from_askpass_tool("zenity",
                              "\"%s\" --password --title=\"%s\" 2>/dev/null",
                              prompt,
                              passphrase,
                              max_size);
  }

  // Now try kdialog (https://develop.kde.org/deploy/kdialog/).
  if (result == PASSPHRASE_FAIL) {
    result = get_from_askpass_tool(
        "kdialog",
        "\"%s\" --password \"%s\" --title=ezcrypt 2>/dev/null",
        prompt,
        passphrase,
        max_size);
  }

  // TODO: We could add other askpass tools too, such as "yad".
#endif

  return result;
}

static int get_from_terminal(const char* prompt,
                             char* passphrase,
                             size_t max_size) {
  // Avoid silly buffer overruns.
  if (max_size <= 1U) {
    if (max_size == 1U) {
      passphrase[0] = 0;
    }
    return PASSPHRASE_ABORT;
  }

  bool too_long_passphrase = false;

#if defined(USE_POSIX_PROMPT)
  bool retry = true;
  while (retry) {
    // Clear the caught signals array.
    for (int i = 0; i < NUM_HANDLE_SIGNALS; ++i) {
      s_caught_signals[i] = 0;
    }

    // Handle signals that would mess up user input & terminal state.
    struct sigaction saved_sigactions[NUM_HANDLE_SIGNALS];
    struct sigaction sa;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = signal_catcher;
    for (int i = 0; i < NUM_HANDLE_SIGNALS; ++i) {
      (void)sigaction(HANDLE_SIGNALS[i], &sa, &saved_sigactions[i]);
    }

    // Use /dev/tty for I/O if possible.
    int in_fd = open(_PATH_TTY, O_RDWR);
    int out_fd = in_fd;
    if (in_fd == -1) {
      // Fall back to stdin/stderr.
      in_fd = STDIN_FILENO;
      out_fd = STDERR_FILENO;
    }

    // Turn off ECHO for the input handle.
    static struct termios old_terminal;
    tcgetattr(in_fd, &old_terminal);
    static struct termios new_terminal;
    new_terminal = old_terminal;
    new_terminal.c_lflag &= ~(ECHO);
    (void)tcsetattr(in_fd, TCSANOW, &new_terminal);

    // Print the prompt.
    const ssize_t prompt_len = (ssize_t)strlen(prompt);
    if (write(out_fd, prompt, strlen(prompt)) != prompt_len) {
      FATAL_ERROR("Unable to show passphrase prompt");
    }
    if (write(out_fd, ": ", 2) != 2) {
      FATAL_ERROR("Unable to show passphrase prompt");
    }

    // Get the passphrase, one character at a time.
    too_long_passphrase = false;
    size_t count = 0U;
    while (true) {
      char c = '\n';
      const ssize_t num_bytes = read(in_fd, &c, 1);
      if (num_bytes != 1 || c == '\n' || c == '\r') {
        break;
      }
      if ((count + 1U) >= max_size) {
        too_long_passphrase = true;
        break;
      }
      passphrase[count++] = c;
    }
    passphrase[count] = 0;

    // Print a new-line character.
    if (write(out_fd, "\n", 1) != 1) {
      FATAL_ERROR("Unable to show passphrase prompt");
    }

    // Restore old terminal settings, and close the tty file if relevant.
    tcsetattr(in_fd, TCSANOW, &old_terminal);
    if (in_fd != STDIN_FILENO) {
      (void)close(in_fd);
    }

    // Restore old signal handlers.
    for (int i = 0; i < NUM_HANDLE_SIGNALS; ++i) {
      (void)sigaction(HANDLE_SIGNALS[i], &saved_sigactions[i], NULL);
    }

    // Replay any caught signals to ourselves, and determine if we need to
    // make another round (which can happen if our process was not in the
    // foreground).
    retry = false;
    for (int i = 0; i < NUM_HANDLE_SIGNALS; ++i) {
      if (s_caught_signals[i] != 0) {
        const int signal_id = HANDLE_SIGNALS[i];
        kill(getpid(), signal_id);
        if (signal_id == SIGTSTP || signal_id == SIGTTIN ||
            signal_id == SIGTTOU) {
          retry = true;
        }
      }
    }
  }
#elif defined(USE_WIN32_PROMPT)
  // Use stdin/stderr for I/O.
  int in_fd = _fileno(stdin);
  int out_fd = _fileno(stderr);
  HANDLE in_handle = (HANDLE)_get_osfhandle(in_fd);

  // Turn off ECHO for stdin.
  bool echo_is_off = false;
  DWORD old_console_mode = 0;
  if (GetConsoleMode(in_handle, &old_console_mode)) {
    const DWORD new_console_mode =
        (old_console_mode & ~ENABLE_ECHO_INPUT) | ENABLE_LINE_INPUT;
    if (SetConsoleMode(in_handle, new_console_mode)) {
      echo_is_off = true;
    }
  }
  if (!echo_is_off) {
    fprintf(stderr, "***Error: Unable to turn off ECHO for console\n");
    return PASSPHRASE_FAIL;
  }

  // Print the prompt.
  (void)_write(out_fd, prompt, (unsigned)strlen(prompt));
  (void)_write(out_fd, ": ", 2);

  // Get the passphrase, one character at a time.
  too_long_passphrase = false;
  size_t count = 0U;
  while (true) {
    char c = '\n';
    int num_bytes = _read(in_fd, &c, 1);
    if (num_bytes != 1 || c == '\n' || c == '\r') {
      break;
    }
    if ((count + 1U) >= max_size) {
      too_long_passphrase = true;
      break;
    }
    passphrase[count++] = c;
  }
  passphrase[count] = 0;

  // Print a new-line character.
  (void)_write(out_fd, "\n", 1);

  // Restore ECHO mode for stdin.
  SetConsoleMode(in_handle, old_console_mode);
#else
  // TODO: Implement me!
  (void)prompt;
  passphrase[0] = 0;
#endif

  // Was the passphrase length valid?
  if (too_long_passphrase) {
    fprintf(stderr, "***Error: Too long passphrase.\n");
    return PASSPHRASE_FAIL;
  }
  return passphrase[0] != 0 ? PASSPHRASE_OK : PASSPHRASE_ABORT;
}

static int get_from_dialog_or_terminal(const char* prompt,
                                       char* passphrase,
                                       size_t max_size,
                                       const int disallow_gui) {
  int result = PASSPHRASE_FAIL;

  // First try a GUI dialog.
  if (!disallow_gui) {
    result = get_from_dialog(prompt, passphrase, max_size);
  }

  // Fall back to a terminal prompt.
  if (result == PASSPHRASE_FAIL) {
    result = get_from_terminal(prompt, passphrase, max_size);
  }

  return result;
}

static int get_from_env(const char* env, char* passphrase, size_t max_size) {
#if defined(_WIN32)
  bool success = false;
  uint16_t env_utf16[1024];
  if (utf8_to_utf16(env, env_utf16, sizeof(env_utf16) / sizeof(uint16_t))) {
    uint16_t* buf = (uint16_t*)secure_alloc(max_size * sizeof(uint16_t));
    const DWORD len = GetEnvironmentVariableW(env_utf16, buf, (DWORD)max_size);
    if (len == 0U) {
      fprintf(stderr, "***Error: No such environment variable: \"%s\"\n", env);
    } else if ((len + 1U) >= max_size) {
      fprintf(stderr, "***Error: Too long passphrase.\n");
    } else {
      success = utf16_to_utf8(buf, passphrase, max_size);
    }
    secure_free(buf);
  }
  if (!success) {
    passphrase[0] = 0;
    return PASSPHRASE_FAIL;
  }
#else
  const char* text = getenv(env);
  if (text == NULL) {
    fprintf(stderr, "***Error: No such environment variable: \"%s\"\n", env);
    passphrase[0] = 0;
    return PASSPHRASE_FAIL;
  }
  if (!string_copy(text, passphrase, max_size)) {
    fprintf(stderr, "***Error: Too long passphrase.\n");
    passphrase[0] = 0;
    return PASSPHRASE_FAIL;
  }
#endif

  // Once we have made a copy of the passphrase, delete the environment variable
  // from the scope of this process and potential child processes.
  // Rationale: Avoid that other processes can get access to it (even if
  // accidentally - e.g. in debug/crash logs).
#if defined(_WIN32)
  (void)SetEnvironmentVariableA(env, NULL);
#else
  (void)unsetenv(env);
#endif

  return PASSPHRASE_OK;
}

static int get_from_file(const char* path, char* passphrase, size_t max_size) {
  // Read the contents of the file into the passphrase string.
  FILE* f = fopen(path, "rb");
  if (f == NULL) {
    fprintf(stderr, "***Error: No such file: \"%s\"\n", path);
    passphrase[0] = 0;
    return PASSPHRASE_FAIL;
  }
  if (fgets(passphrase, (int)max_size, f) == NULL) {
    fprintf(stderr, "***Error: Failed to read file: \"%s\"\n", path);
    fclose(f);
    passphrase[0] = 0;
    return PASSPHRASE_FAIL;
  }
  fclose(f);
  terminate_str_at_newline(passphrase, max_size);

  return PASSPHRASE_OK;
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

bool get_passphrase(const args_t* args,
                    bool verify,
                    char* passphrase,
                    size_t max_size) {
  PRECONDITION(args != NULL);
  PRECONDITION(passphrase != NULL);
  PRECONDITION(max_size > 0U);

  // Is the passphrase provided via an environment variable?
  if (args->passphrase_env != NULL) {
    int result = get_from_env(args->passphrase_env, passphrase, max_size);
    return result == PASSPHRASE_OK;
  }

  // Is the passphrase provided via a file?
  if (args->passphrase_file != NULL) {
    int result = get_from_file(args->passphrase_file, passphrase, max_size);
    return result == PASSPHRASE_OK;
  }

  // Otherwise, get the passphrase from a dialog or the terminal.
  int result = get_from_dialog_or_terminal(
      "Enter passphrase", passphrase, max_size, args->disallow_gui);
  if (result != PASSPHRASE_OK) {
    return false;
  }
  if (!verify) {
    return true;
  }

  // Verify the password by asking for it again.
  char* passphrase_again = (char*)secure_alloc(max_size);
  bool success = false;
  result = get_from_dialog_or_terminal("Please repeat the passphrase",
                                       passphrase_again,
                                       max_size,
                                       args->disallow_gui);
  if (result == PASSPHRASE_OK) {
    if (strcmp(passphrase, passphrase_again) == 0) {
      success = true;
    } else {
      fprintf(stderr, "***Error: Passphrases do not match\n");
    }
  }
  secure_free(passphrase_again);

  return success;
}

void kill_passphrase_child_processes(void) {
#if defined(HAS_GTK3)
  kill_passphrase_child_processes_gtk();
#endif
}
