// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "pepper.h"

#include "crypto/entropy.h"
#include "crypto/sha2.h"
#include "memory.h"
#include <assert.h>

static uint8_t s_data_buffer[4096];

bool make_pepper_file(FILE* pepper_file, size_t length) {
  bool success = true;

  // Write the number of requested random bytes to the file.
  size_t bytes_left = length;
  while (bytes_left > 0U) {
    // Generate more entropy.
    const size_t bytes_to_write =
        bytes_left < sizeof(s_data_buffer) ? bytes_left : sizeof(s_data_buffer);
    make_entropy(s_data_buffer, bytes_to_write);

    // Write the random data to the pepper file.
    const size_t bytes_written =
        fwrite(&s_data_buffer[0], 1, bytes_to_write, pepper_file);
    if (bytes_written != bytes_to_write) {
      // This should only happen if there was a write error.
      success = false;
      break;
    }
    bytes_left -= bytes_written;
  }

  // Nuke the entropy buffer.
  secure_zero_memory(&s_data_buffer[0], sizeof(s_data_buffer));

  return success;
}

bool derive_pepper_from_file(FILE* pepper_file, uint8_t pepper[PEPPER_BYTES]) {
  static_assert(PEPPER_BYTES == SHA256_HASH_BYTES,
                "Pepper size must match SHA2-256 hash size");

  bool success = true;

  // The pepper is the SHA2-256 hash of the given file.
  sha2_256_ctx_t ctx;
  sha2_256_init(&ctx);
  while (true) {
    // Read more bytes from the pepper file.
    const size_t bytes_read =
        fread(&s_data_buffer[0], 1, sizeof(s_data_buffer), pepper_file);

    // Update the SHA2-256 has with the newly read bytes.
    sha2_256_append(&ctx, &s_data_buffer[0], bytes_read);

    if (ferror(pepper_file)) {
      success = false;
    }
    if (feof(pepper_file)) {
      break;
    }
  }
  sha2_256_finalize(&ctx, &pepper[0]);

  // Nuke the data buffer.
  secure_zero_memory(&s_data_buffer[0], sizeof(s_data_buffer));

  return success;
}
