// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "stringutils.h"

#include "error.h"
#include <string.h>

//------------------------------------------------------------------------------
// Private functions.
//------------------------------------------------------------------------------

static bool utf8_char_to_codepoint(const uint8_t* token,
                                   uint32_t* codepoint,
                                   int* token_len) {
  if (token[0] <= 0x7fU) {
    // 1 byte.
    *codepoint = (uint32_t)token[0];
    *token_len = 1;
  } else if ((token[0] & 0xe0U) == 0xc0U) {
    // 2 bytes.
    if ((token[1] & 0xc0U) != 0x80U) {
      return false;
    }
    *codepoint = (((uint32_t)(token[0] & 0x1fU)) << 6) |
                 (((uint32_t)(token[1] & 0x3fU)) << 0);
    *token_len = 2;
  } else if ((token[0] & 0xf0U) == 0xe0U) {
    // 3 bytes.
    if (((token[1] & 0xc0U) != 0x80U) || ((token[2] & 0xc0U) != 0x80U)) {
      return false;
    }
    *codepoint = (((uint32_t)(token[0] & 0x0fU)) << 12) |
                 (((uint32_t)(token[1] & 0x3fU)) << 6) |
                 (((uint32_t)(token[2] & 0x3fU)) << 0);
    *token_len = 3;
  } else if ((token[0] & 0xf8U) == 0xf0U) {
    // 4 bytes.
    if (((token[1] & 0xc0U) != 0x80U) || ((token[2] & 0xc0U) != 0x80U) ||
        ((token[3] & 0xc0U) != 0x80U)) {
      return false;
    }
    *codepoint = (((uint32_t)(token[0] & 0x07U)) << 18) |
                 (((uint32_t)(token[1] & 0x3fU)) << 12) |
                 (((uint32_t)(token[2] & 0x3fU)) << 6) |
                 (((uint32_t)(token[3] & 0x3fU)) << 0);
    *token_len = 4;
  } else {
    return false;
  }

  return true;
}

static bool codepoint_to_utf8_char(uint32_t codepoint,
                                   uint8_t* token,
                                   int* token_len) {
  if (codepoint <= 0x7fU) {
    // 1 byte.
    token[0] = (uint8_t)codepoint;
    *token_len = 1;
  } else if (codepoint <= 0x7ffU) {
    // 2 bytes.
    token[0] = (uint8_t)(0xc0U | (codepoint >> 6));
    token[1] = (uint8_t)(0x80U | (codepoint & 0x3fU));
    *token_len = 2;
  } else if (codepoint <= 0xffffU) {
    // 3 bytes.
    token[0] = (uint8_t)(0xe0U | (codepoint >> 12));
    token[1] = (uint8_t)(0x80U | ((codepoint >> 6) & 0x3fU));
    token[2] = (uint8_t)(0x80U | (codepoint & 0x3fU));
    *token_len = 3;
  } else if (codepoint <= 0x10ffffU) {
    // 4 bytes.
    token[0] = (uint8_t)(0xf0U | (codepoint >> 18));
    token[1] = (uint8_t)(0x80U | ((codepoint >> 12) & 0x3fU));
    token[2] = (uint8_t)(0x80U | ((codepoint >> 6) & 0x3fU));
    token[3] = (uint8_t)(0x80U | (codepoint & 0x3fU));
    *token_len = 4;
  } else {
    // Note: This can never happen when the code point comes from UTF-16 or
    // UTF-8.
    return false;
  }

  return true;
}

static bool utf16_char_to_codepoint(const uint16_t* token,
                                    uint32_t* codepoint,
                                    int* token_len) {
  if (token[0] <= 0xd7ffU || token[0] >= 0xe000U) {
    // A single 16-bit word.
    *codepoint = (uint32_t)token[0];
    *token_len = 1;
  } else if ((token[0] & 0xfc00U) == 0xd800U) {
    // A surrogate pair.
    if ((token[1] & 0xfc00U) != 0xdc00U) {
      return false;
    }
    *codepoint = 0x10000U + ((((uint32_t)(token[0] & 0x3ffU)) << 10) |
                             (((uint32_t)(token[1] & 0x3ffU)) << 0));
    *token_len = 2;
  } else {
    return false;
  }

  return true;
}

static bool codepoint_to_utf16_char(uint32_t codepoint,
                                    uint16_t* token,
                                    int* token_len) {
  if (codepoint <= 0xd7ffU || (codepoint >= 0xe000U && codepoint <= 0xffffU)) {
    // A single 16-bit word.
    token[0] = (uint16_t)codepoint;
    *token_len = 1;
  } else if (codepoint >= 0x10000U && codepoint <= 0x10ffffU) {
    // A surrogate pair.
    codepoint -= 0x10000U;
    token[0] = 0xd800U | (uint16_t)(codepoint >> 10);
    token[1] = 0xdc00U | (uint16_t)(codepoint & 0x3ffU);
    *token_len = 2;
  } else {
    return false;
  }

  return true;
}

//------------------------------------------------------------------------------
// Public functions.
//------------------------------------------------------------------------------

bool string_copy(const char* src, char* dst, size_t dst_size) {
  PRECONDITION(src != NULL);
  PRECONDITION(dst != NULL);
  PRECONDITION(dst_size >= 1U);

  // strncpy() has odd semantics, so we implement our own string copy routine
  // that better matches our usage patterns.
  const size_t src_len = strlen(src) + 1U;
  if (src_len > dst_size) {
    return false;
  }
  memcpy(dst, src, src_len);
  return true;
}

size_t numchars_utf8(const char* str) {
  size_t result = 0U;
  while (*str != 0) {
    int token_len = 1;
    uint32_t cp = 0;
    if (!utf8_char_to_codepoint((const uint8_t*)str, &cp, &token_len)) {
      return 0;
    }
    str += token_len;
    ++result;
  }
  return result;
}

size_t numchars_utf16(const uint16_t* str) {
  size_t result = 0U;
  while (*str != 0) {
    int token_len = 1;
    uint32_t cp = 0;
    if (!utf16_char_to_codepoint(str, &cp, &token_len)) {
      return 0;
    }
    str += token_len;
    ++result;
  }
  return result;
}

bool utf8_to_utf16(const char* src, uint16_t* dst, size_t dst_size) {
  PRECONDITION(src != NULL);
  PRECONDITION(dst != NULL);
  PRECONDITION(dst_size >= 1U);

  // Convert one token at a time.
  const uint16_t* dst_end = dst + dst_size;
  while (*src != 0 && dst < dst_end) {
    int utf8_token_len = 1;
    uint32_t cp = 0;
    if (!utf8_char_to_codepoint((const uint8_t*)src, &cp, &utf8_token_len)) {
      return false;
    }
    src += utf8_token_len;

    int utf16_token_len = 1;
    uint16_t utf16_token[2];
    if (!codepoint_to_utf16_char(cp, utf16_token, &utf16_token_len)) {
      return false;
    }
    if ((dst + utf16_token_len) > dst_end) {
      return false;
    }

    memcpy(dst, utf16_token, sizeof(uint16_t) * utf16_token_len);
    dst += utf16_token_len;
  }

  // Null terminate the string.
  if (dst >= dst_end) {
    return false;
  }
  *dst = 0;

  return true;
}

bool utf16_to_utf8(const uint16_t* src, char* dst, size_t dst_size) {
  PRECONDITION(src != NULL);
  PRECONDITION(dst != NULL);
  PRECONDITION(dst_size >= 1U);

  // Convert one token at a time.
  const char* dst_end = dst + dst_size;
  while (*src != 0 && dst < dst_end) {
    int utf16_token_len = 1;
    uint32_t cp = 0;
    if (!utf16_char_to_codepoint(src, &cp, &utf16_token_len)) {
      return false;
    }
    src += utf16_token_len;

    int utf8_token_len = 1;
    uint8_t utf8_token[4];
    if (!codepoint_to_utf8_char(cp, utf8_token, &utf8_token_len)) {
      // Note: This can never happen since the source is valid UTF-16.
      return false;
    }
    if ((dst + utf8_token_len) > dst_end) {
      return false;
    }

    memcpy(dst, utf8_token, utf8_token_len);
    dst += utf8_token_len;
  }

  // Null terminate the string.
  if (dst >= dst_end) {
    return false;
  }
  *dst = 0;

  return true;
}

uint16_t* utf8_to_utf16_alloc(const char* str) {
  // Worst case size.
  const size_t str_utf16_size = (2U * numchars_utf8(str)) + 1U;

  // Allocate memory for the buffer.
  uint16_t* str_utf16 = (uint16_t*)malloc(str_utf16_size * sizeof(uint16_t));
  if (str_utf16 == NULL) {
    FATAL_ERROR("Out of memory");
  }

  // Perform the conversion.
  if (!utf8_to_utf16(str, str_utf16, str_utf16_size)) {
    FATAL_ERROR("Invalid UTF-8 string?");
  }

  // Note: The caller is responsible for freeing the buffer.
  return str_utf16;
}

char* utf16_to_utf8_alloc(const uint16_t* str) {
  // Worst case size.
  const size_t str_utf8_size = (4U * numchars_utf16(str)) + 1U;

  // Allocate memory for the buffer.
  char* str_utf8 = (char*)malloc(str_utf8_size);
  if (str_utf8 == NULL) {
    FATAL_ERROR("Out of memory");
  }

  // Perform the conversion.
  if (!utf16_to_utf8(str, str_utf8, str_utf8_size)) {
    FATAL_ERROR("Invalid UTF-16 string?");
  }

  // Note: The caller is responsible for freeing the buffer.
  return str_utf8;
}
