// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/aes.h"
#include "minitest.h"

#include <stdint.h>
#include <string.h>

// AES acceleration method.
#define AES_METHOD_GENERIC 0
#define AES_METHOD_FASTEST 1
#define AES_METHOD_FIRST AES_METHOD_GENERIC
#define AES_METHOD_LAST AES_METHOD_FASTEST

// Key and IV from: NIST Special Publication 800-38A 2001 Edition, Appendix F
// Key: 603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4
// IV:  000102030405060708090a0b0c0d0e0f
static const uint8_t NIST_KEY[32] = {
    0x60, 0x3d, 0xeb, 0x10, 0x15, 0xca, 0x71, 0xbe, 0x2b, 0x73, 0xae,
    0xf0, 0x85, 0x7d, 0x77, 0x81, 0x1f, 0x35, 0x2c, 0x07, 0x3b, 0x61,
    0x08, 0xd7, 0x2d, 0x98, 0x10, 0xa3, 0x09, 0x14, 0xdf, 0xf4};
static const uint8_t NIST_IV[16] = {0x00,
                                    0x01,
                                    0x02,
                                    0x03,
                                    0x04,
                                    0x05,
                                    0x06,
                                    0x07,
                                    0x08,
                                    0x09,
                                    0x0a,
                                    0x0b,
                                    0x0c,
                                    0x0d,
                                    0x0e,
                                    0x0f};

TEST_MAIN("AES-CBC-256 cipher") {
  TEST_CASE("NIST 800-38a F.2.5 CBC-AES256.Encrypt example") {
    for (int method = AES_METHOD_FIRST; method <= AES_METHOD_LAST; ++method) {
      // Given: A plaintext.
      const uint8_t plaintext[64] = {
          0x6b, 0xc1, 0xbe, 0xe2, 0x2e, 0x40, 0x9f, 0x96, 0xe9, 0x3d, 0x7e,
          0x11, 0x73, 0x93, 0x17, 0x2a, 0xae, 0x2d, 0x8a, 0x57, 0x1e, 0x03,
          0xac, 0x9c, 0x9e, 0xb7, 0x6f, 0xac, 0x45, 0xaf, 0x8e, 0x51, 0x30,
          0xc8, 0x1c, 0x46, 0xa3, 0x5c, 0xe4, 0x11, 0xe5, 0xfb, 0xc1, 0x19,
          0x1a, 0x0a, 0x52, 0xef, 0xf6, 0x9f, 0x24, 0x45, 0xdf, 0x4f, 0x9b,
          0x17, 0xad, 0x2b, 0x41, 0x7b, 0xe6, 0x6c, 0x37, 0x10};

      // When: Encrypt the message.
      uint8_t buffer[64];
      memcpy(buffer, plaintext, sizeof(buffer));
      aes256_ctx_t ctx;
      aes_256_init_ex(&ctx, NIST_KEY, NIST_IV, method == AES_METHOD_GENERIC);
      aes_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

      // Then: The ciphertext should be correct.
      const uint8_t expected[64] = {
          0xf5, 0x8c, 0x4c, 0x04, 0xd6, 0xe5, 0xf1, 0xba, 0x77, 0x9e, 0xab,
          0xfb, 0x5f, 0x7b, 0xfb, 0xd6, 0x9c, 0xfc, 0x4e, 0x96, 0x7e, 0xdb,
          0x80, 0x8d, 0x67, 0x9f, 0x77, 0x7b, 0xc6, 0x70, 0x2c, 0x7d, 0x39,
          0xf2, 0x33, 0x69, 0xa9, 0xd9, 0xba, 0xcf, 0xa5, 0x30, 0xe2, 0x63,
          0x04, 0x23, 0x14, 0x61, 0xb2, 0xeb, 0x05, 0xe2, 0xc3, 0x9b, 0xe9,
          0xfc, 0xda, 0x6c, 0x19, 0x07, 0x8c, 0x6a, 0x9d, 0x1b};
      CHECK_EQ_BUF(buffer, expected);
    }
  }

  TEST_CASE("NIST 800-38a F.2.5 CBC-AES256.Decrypt example") {
    for (int method = AES_METHOD_FIRST; method <= AES_METHOD_LAST; ++method) {
      // Given: A ciphertext.
      const uint8_t ciphertext[64] = {
          0xf5, 0x8c, 0x4c, 0x04, 0xd6, 0xe5, 0xf1, 0xba, 0x77, 0x9e, 0xab,
          0xfb, 0x5f, 0x7b, 0xfb, 0xd6, 0x9c, 0xfc, 0x4e, 0x96, 0x7e, 0xdb,
          0x80, 0x8d, 0x67, 0x9f, 0x77, 0x7b, 0xc6, 0x70, 0x2c, 0x7d, 0x39,
          0xf2, 0x33, 0x69, 0xa9, 0xd9, 0xba, 0xcf, 0xa5, 0x30, 0xe2, 0x63,
          0x04, 0x23, 0x14, 0x61, 0xb2, 0xeb, 0x05, 0xe2, 0xc3, 0x9b, 0xe9,
          0xfc, 0xda, 0x6c, 0x19, 0x07, 0x8c, 0x6a, 0x9d, 0x1b};

      // When: Decrypt the message.
      uint8_t buffer[64];
      memcpy(buffer, ciphertext, sizeof(buffer));
      aes256_ctx_t ctx;
      aes_256_init_ex(&ctx, NIST_KEY, NIST_IV, method == AES_METHOD_GENERIC);
      aes_256_cbc_decrypt_buffer(&ctx, buffer, sizeof(buffer));

      // Then: The plaintext should be correct.
      const uint8_t expected[64] = {
          0x6b, 0xc1, 0xbe, 0xe2, 0x2e, 0x40, 0x9f, 0x96, 0xe9, 0x3d, 0x7e,
          0x11, 0x73, 0x93, 0x17, 0x2a, 0xae, 0x2d, 0x8a, 0x57, 0x1e, 0x03,
          0xac, 0x9c, 0x9e, 0xb7, 0x6f, 0xac, 0x45, 0xaf, 0x8e, 0x51, 0x30,
          0xc8, 0x1c, 0x46, 0xa3, 0x5c, 0xe4, 0x11, 0xe5, 0xfb, 0xc1, 0x19,
          0x1a, 0x0a, 0x52, 0xef, 0xf6, 0x9f, 0x24, 0x45, 0xdf, 0x4f, 0x9b,
          0x17, 0xad, 0x2b, 0x41, 0x7b, 0xe6, 0x6c, 0x37, 0x10};
      CHECK_EQ_BUF(buffer, expected);
    }
  }

  TEST_CASE("Hello world... 128 bytes") {
    for (int method = AES_METHOD_FIRST; method <= AES_METHOD_LAST; ++method) {
      // Given: A plaintext, a key and an IV.
      const uint8_t plaintext[128] =
          "Hello world! This is a small text that we want to encrypt. We make "
          "it long enough to contain a multiple of 64 bytes...........!\n";
      const uint8_t key[32] = {0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef,
                               0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
                               0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88,
                               0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00};
      const uint8_t iv[16] = {0x00,
                              0xa0,
                              0x30,
                              0x40,
                              0x50,
                              0x60,
                              0x70,
                              0x89,
                              0xab,
                              0x12,
                              0x39,
                              0x87,
                              0x12,
                              0x3a,
                              0xbc,
                              0xde};

      // When: Encrypt the message.
      uint8_t buffer[128];
      memcpy(buffer, plaintext, sizeof(buffer));
      aes256_ctx_t ctx;
      aes_256_init_ex(&ctx, key, iv, method == AES_METHOD_GENERIC);
      aes_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

      // Then: The ciphertext should be correct.
      // This is the output of "openssl aes-256-cbc -K ... -iv ...".
      const uint8_t expected[128] = {
          0x70, 0x48, 0x75, 0x3c, 0x52, 0xf1, 0xb2, 0x91, 0x5d, 0xfa, 0x8d,
          0x44, 0x00, 0xcb, 0x3d, 0x07, 0x54, 0x6d, 0xf0, 0xda, 0xe0, 0x67,
          0x05, 0x9b, 0x49, 0xea, 0x1d, 0x90, 0x16, 0xda, 0x7d, 0x41, 0x4d,
          0x22, 0x2a, 0xae, 0x08, 0xfb, 0x82, 0xb4, 0x67, 0x01, 0x3e, 0x04,
          0x29, 0x17, 0x9e, 0xd7, 0x49, 0x50, 0x01, 0x81, 0x23, 0xa0, 0x5e,
          0xdb, 0xce, 0x11, 0xd2, 0x86, 0xd0, 0x3c, 0xa7, 0xfd, 0xff, 0x77,
          0xd0, 0xaf, 0x10, 0x0b, 0x46, 0x2c, 0xa2, 0x85, 0xc9, 0xa0, 0x06,
          0xcf, 0x4d, 0x06, 0xa8, 0x11, 0x0b, 0xfa, 0xdb, 0x63, 0xfd, 0x31,
          0xf4, 0x40, 0xb8, 0x67, 0x64, 0x43, 0x57, 0x3f, 0xb7, 0xc4, 0xf8,
          0xc3, 0x3b, 0xf4, 0xcf, 0x04, 0xcd, 0xf6, 0x9e, 0x5f, 0x8a, 0x44,
          0x28, 0x43, 0xda, 0x35, 0x62, 0x44, 0x55, 0xdb, 0xd6, 0x7f, 0x47,
          0x31, 0x2d, 0x85, 0x67, 0x6b, 0x80, 0x4e};
      CHECK_EQ_BUF(buffer, expected);
    }
  }

  TEST_CASE("Encrypt + decrypt 256 bytes") {
    for (int method = AES_METHOD_FIRST; method <= AES_METHOD_LAST; ++method) {
      // Given: A plaintext, a key and an IV.
      const uint8_t plaintext[256] =
          "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do "
          "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim "
          "ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
          "aliquip ex ea commodo consequat. Duis aute irure dolor in";
      const uint8_t key[32] = {0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef,
                               0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
                               0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88,
                               0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00};
      const uint8_t iv[16] = {0x00,
                              0xa0,
                              0x30,
                              0x40,
                              0x50,
                              0x60,
                              0x70,
                              0x89,
                              0xab,
                              0x12,
                              0x39,
                              0x87,
                              0x12,
                              0x3a,
                              0xbc,
                              0xde};

      // When: Encrypt the message and decrypt it again.
      uint8_t buffer[256];
      memcpy(buffer, plaintext, sizeof(buffer));
      aes256_ctx_t ctx;
      aes_256_init_ex(&ctx, key, iv, method == AES_METHOD_GENERIC);
      aes_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

      aes_256_init_ex(&ctx, key, iv, method == AES_METHOD_GENERIC);
      aes_256_cbc_decrypt_buffer(&ctx, buffer, sizeof(buffer));

      // Then: The plaintext should be identical to the original.
      CHECK_EQ_BUF(buffer, plaintext);
    }
  }
}
