// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/chacha.h"
#include "minitest.h"

#include <stdint.h>
#include <string.h>

// Test vectors from "Test Vectors for the Stream Cipher ChaCha" (draft).
// https://datatracker.ietf.org/doc/html/draft-strombergson-chacha-test-vectors-01
static const uint8_t TC8_KEY[32] = {
    0xc4, 0x6e, 0xc1, 0xb1, 0x8c, 0xe8, 0xa8, 0x78, 0x72, 0x5a, 0x37,
    0xe7, 0x80, 0xdf, 0xb7, 0x35, 0x1f, 0x68, 0xed, 0x2e, 0x19, 0x4c,
    0x79, 0xfb, 0xc6, 0xae, 0xbe, 0xe1, 0xa6, 0x67, 0x97, 0x5d};
static const uint8_t TC8_IV[8] =
    {0x1a, 0xda, 0x31, 0xd5, 0xcf, 0x68, 0x82, 0x21};

TEST_MAIN("ChaCha20-256 cipher") {
  TEST_CASE("TC8: All your base are belong to us! - All-zero plaintext") {
    // Given: An all-zero plaintext.
    const uint8_t plaintext[128] = {0};

    // When: Encrypt the message.
    uint8_t buffer[128];
    memcpy(buffer, plaintext, sizeof(buffer));
    chacha256_ctx_t ctx;
    chacha20_256_init(&ctx, TC8_KEY, TC8_IV);
    chacha20_256_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[128] = {
        0xf6, 0x3a, 0x89, 0xb7, 0x5c, 0x22, 0x71, 0xf9, 0x36, 0x88, 0x16, 0x54,
        0x2b, 0xa5, 0x2f, 0x06, 0xed, 0x49, 0x24, 0x17, 0x92, 0x30, 0x2b, 0x00,
        0xb5, 0xe8, 0xf8, 0x0a, 0xe9, 0xa4, 0x73, 0xaf, 0xc2, 0x5b, 0x21, 0x8f,
        0x51, 0x9a, 0xf0, 0xfd, 0xd4, 0x06, 0x36, 0x2e, 0x8d, 0x69, 0xde, 0x7f,
        0x54, 0xc6, 0x04, 0xa6, 0xe0, 0x0f, 0x35, 0x3f, 0x11, 0x0f, 0x77, 0x1b,
        0xdc, 0xa8, 0xab, 0x92, 0xe5, 0xfb, 0xc3, 0x4e, 0x60, 0xa1, 0xd9, 0xa9,
        0xdb, 0x17, 0x34, 0x5b, 0x0a, 0x40, 0x27, 0x36, 0x85, 0x3b, 0xf9, 0x10,
        0xb0, 0x60, 0xbd, 0xf1, 0xf8, 0x97, 0xb6, 0x29, 0x0f, 0x01, 0xd1, 0x38,
        0xae, 0x2c, 0x4c, 0x90, 0x22, 0x5b, 0xa9, 0xea, 0x14, 0xd5, 0x18, 0xf5,
        0x59, 0x29, 0xde, 0xa0, 0x98, 0xca, 0x7a, 0x6c, 0xcf, 0xe6, 0x12, 0x27,
        0x05, 0x3c, 0x84, 0xe4, 0x9a, 0x4a, 0x33, 0x32};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("TC8: All your base are belong to us! - ASCII plaintext") {
    // Given: A non-zero plaintext.
    const uint8_t plaintext[64] =
        "Hello world! This is a message that will be encrypted. Hold on!";

    // When: Encrypt the message.
    uint8_t buffer[64];
    memcpy(buffer, plaintext, sizeof(buffer));
    chacha256_ctx_t ctx;
    chacha20_256_init(&ctx, TC8_KEY, TC8_IV);
    chacha20_256_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[64] = {
        0xbe, 0x5f, 0xe5, 0xdb, 0x33, 0x02, 0x06, 0x96, 0x44, 0xe4, 0x72,
        0x75, 0x0b, 0xf1, 0x47, 0x6f, 0x9e, 0x69, 0x4d, 0x64, 0xb2, 0x51,
        0x0b, 0x6d, 0xd0, 0x9b, 0x8b, 0x6b, 0x8e, 0xc1, 0x53, 0xdb, 0xaa,
        0x3a, 0x55, 0xaf, 0x26, 0xf3, 0x9c, 0x91, 0xf4, 0x64, 0x53, 0x0e,
        0xe8, 0x07, 0xbd, 0x0d, 0x2d, 0xb6, 0x70, 0xc3, 0x84, 0x21, 0x15,
        0x77, 0x7e, 0x63, 0x13, 0x3b, 0xb3, 0xc6, 0x8a, 0x92};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("XChaCha20: IETF RFC draft xchacha-03 - A.2.1") {
    // Given: The A.2.1 test vector plaintext (padded to a multiple of 64
    // bytes), key, and nonce. See:
    // https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-xchacha-03.
    const uint8_t plaintext[320] =
        "The dhole (pronounced \"dole\") is also known as the Asiatic wild "
        "dog, red dog, and whistling dog. It is about the size of a German "
        "shepherd but looks more like a long-legged fox. This highly elusive "
        "and skilled jumper is classified with wolves, coyotes, jackals, and "
        "foxes in the taxonomic family "
        "Canidae.\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
    const uint8_t key[CHACHA_KEY_BYTES] = {
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a,
        0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95,
        0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f};
    const uint8_t iv[XCHACHA_IV_BYTES] = {
        0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b,
        0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x58};

    // When: Encrypt the message.
    uint8_t buffer[320];
    memcpy(buffer, plaintext, sizeof(buffer));
    chacha256_ctx_t ctx;
    xchacha20_256_init(&ctx, key, iv);
    chacha20_256_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    // Note: The last 16 bytes are not part of the original draft RFC test
    // vector, but we need to include them since the plaintext is not a multiple
    // of the block size.
    const uint8_t expected[320] = {
        0x45, 0x59, 0xab, 0xba, 0x4e, 0x48, 0xc1, 0x61, 0x02, 0xe8, 0xbb, 0x2c,
        0x05, 0xe6, 0x94, 0x7f, 0x50, 0xa7, 0x86, 0xde, 0x16, 0x2f, 0x9b, 0x0b,
        0x7e, 0x59, 0x2a, 0x9b, 0x53, 0xd0, 0xd4, 0xe9, 0x8d, 0x8d, 0x64, 0x10,
        0xd5, 0x40, 0xa1, 0xa6, 0x37, 0x5b, 0x26, 0xd8, 0x0d, 0xac, 0xe4, 0xfa,
        0xb5, 0x23, 0x84, 0xc7, 0x31, 0xac, 0xbf, 0x16, 0xa5, 0x92, 0x3c, 0x0c,
        0x48, 0xd3, 0x57, 0x5d, 0x4d, 0x0d, 0x2c, 0x67, 0x3b, 0x66, 0x6f, 0xaa,
        0x73, 0x10, 0x61, 0x27, 0x77, 0x01, 0x09, 0x3a, 0x6b, 0xf7, 0xa1, 0x58,
        0xa8, 0x86, 0x42, 0x92, 0xa4, 0x1c, 0x48, 0xe3, 0xa9, 0xb4, 0xc0, 0xda,
        0xec, 0xe0, 0xf8, 0xd9, 0x8d, 0x0d, 0x7e, 0x05, 0xb3, 0x7a, 0x30, 0x7b,
        0xbb, 0x66, 0x33, 0x31, 0x64, 0xec, 0x9e, 0x1b, 0x24, 0xea, 0x0d, 0x6c,
        0x3f, 0xfd, 0xdc, 0xec, 0x4f, 0x68, 0xe7, 0x44, 0x30, 0x56, 0x19, 0x3a,
        0x03, 0xc8, 0x10, 0xe1, 0x13, 0x44, 0xca, 0x06, 0xd8, 0xed, 0x8a, 0x2b,
        0xfb, 0x1e, 0x8d, 0x48, 0xcf, 0xa6, 0xbc, 0x0e, 0xb4, 0xe2, 0x46, 0x4b,
        0x74, 0x81, 0x42, 0x40, 0x7c, 0x9f, 0x43, 0x1a, 0xee, 0x76, 0x99, 0x60,
        0xe1, 0x5b, 0xa8, 0xb9, 0x68, 0x90, 0x46, 0x6e, 0xf2, 0x45, 0x75, 0x99,
        0x85, 0x23, 0x85, 0xc6, 0x61, 0xf7, 0x52, 0xce, 0x20, 0xf9, 0xda, 0x0c,
        0x09, 0xab, 0x6b, 0x19, 0xdf, 0x74, 0xe7, 0x6a, 0x95, 0x96, 0x74, 0x46,
        0xf8, 0xd0, 0xfd, 0x41, 0x5e, 0x7b, 0xee, 0x2a, 0x12, 0xa1, 0x14, 0xc2,
        0x0e, 0xb5, 0x29, 0x2a, 0xe7, 0xa3, 0x49, 0xae, 0x57, 0x78, 0x20, 0xd5,
        0x52, 0x0a, 0x1f, 0x3f, 0xb6, 0x2a, 0x17, 0xce, 0x6a, 0x7e, 0x68, 0xfa,
        0x7c, 0x79, 0x11, 0x1d, 0x88, 0x60, 0x92, 0x0b, 0xc0, 0x48, 0xef, 0x43,
        0xfe, 0x84, 0x48, 0x6c, 0xcb, 0x87, 0xc2, 0x5f, 0x0a, 0xe0, 0x45, 0xf0,
        0xcc, 0xe1, 0xe7, 0x98, 0x9a, 0x9a, 0xa2, 0x20, 0xa2, 0x8b, 0xdd, 0x48,
        0x27, 0xe7, 0x51, 0xa2, 0x4a, 0x6d, 0x5c, 0x62, 0xd7, 0x90, 0xa6, 0x63,
        0x93, 0xb9, 0x31, 0x11, 0xc1, 0xa5, 0x5d, 0xd7, 0x42, 0x1a, 0x10, 0x18,
        0x49, 0x74, 0xc7, 0xc5, 0x5c, 0x50, 0x48, 0x44, 0x51, 0xe5, 0x4e, 0x1b,
        0x4b, 0x99, 0x5b, 0xd8, 0xfd, 0xd9, 0x45, 0x97};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("Encrypt+decrypt a message") {
    // Given: A message, a key and an IV.
    const uint8_t plaintext[64] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    const uint8_t key[32] = {0x19, 0x29, 0x39, 0x49, 0x59, 0x69, 0x79, 0x89,
                             0xab, 0xcb, 0xdb, 0xeb, 0xfb, 0xeb, 0xdb, 0xcb,
                             0xab, 0x83, 0x94, 0x76, 0xaf, 0x83, 0x28, 0x74,
                             0xfd, 0x8a, 0x75, 0xc9, 0x00, 0xd0, 0x8c, 0x6d};
    const uint8_t iv[8] = {0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef};

    // When: Encrypt and decrypt the message.
    uint8_t buffer[64];
    memcpy(buffer, plaintext, sizeof(buffer));
    chacha256_ctx_t ctx;
    chacha20_256_init(&ctx, key, iv);
    chacha20_256_encrypt_buffer(&ctx, buffer, sizeof(buffer));
    chacha20_256_init(&ctx, key, iv);
    chacha20_256_decrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The decrypted ciphertext should equal the plaintext.
    CHECK_EQ_BUF(buffer, plaintext);
  }
}
