// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/poly1305.h"
#include "minitest.h"

#include <stdint.h>

// Some of the tests use the test vectors from RFC 8439,
// https://www.rfc-editor.org/rfc/rfc8439

// The test key material from RFC 8439.
static const uint8_t RFC8439_KEY[POLY1305_KEY_BYTES] = {
    0x85, 0xd6, 0xbe, 0x78, 0x57, 0x55, 0x6d, 0x33, 0x7f, 0x44, 0x52,
    0xfe, 0x42, 0xd5, 0x06, 0xa8, 0x01, 0x03, 0x80, 0x8a, 0xfb, 0x0d,
    0xb2, 0xfd, 0x4a, 0xbf, 0xf6, 0xaf, 0x41, 0x49, 0xf5, 0x1b};

TEST_MAIN("Poly1305 keyed hash function") {
  TEST_CASE("Zero key - empty message") {
    // Given: A test key and an empty message.
    const uint8_t key[POLY1305_KEY_BYTES] = {0};
    const char message[] = "";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(key, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00,
                                                   0x00};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("RFC 8439 key - empty message") {
    // Given: A test key and an empty message.
    const char message[] = "";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(RFC8439_KEY, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0x01,
                                                   0x03,
                                                   0x80,
                                                   0x8a,
                                                   0xfb,
                                                   0x0d,
                                                   0xb2,
                                                   0xfd,
                                                   0x4a,
                                                   0xbf,
                                                   0xf6,
                                                   0xaf,
                                                   0x41,
                                                   0x49,
                                                   0xf5,
                                                   0x1b};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("RFC 8439 key - RFC 8439 message") {
    // Given: A test key and the test vector message from RFC 8439.
    const char message[] = "Cryptographic Forum Research Group";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(RFC8439_KEY, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0xa8,
                                                   0x06,
                                                   0x1d,
                                                   0xc1,
                                                   0x30,
                                                   0x51,
                                                   0x36,
                                                   0xc6,
                                                   0xc2,
                                                   0x2b,
                                                   0x8b,
                                                   0xaf,
                                                   0x0c,
                                                   0x01,
                                                   0x27,
                                                   0xa9};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("RFC 8439 key - Short message (5 bytes)") {
    // Given: A test key and a message that is exactly one Poly1305 block long.
    const char message[] = "Hello";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(RFC8439_KEY, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0xf7,
                                                   0x4f,
                                                   0x69,
                                                   0x4d,
                                                   0xcd,
                                                   0xf0,
                                                   0xd5,
                                                   0x13,
                                                   0x1e,
                                                   0xd5,
                                                   0x9f,
                                                   0x4b,
                                                   0x4e,
                                                   0x76,
                                                   0x04,
                                                   0x95};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("RFC 8439 key - One block message (16 bytes)") {
    // Given: A test key and a message that is exactly one Poly1305 block long.
    const char message[] = "Hello world, sir";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(RFC8439_KEY, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0x0d,
                                                   0x61,
                                                   0x00,
                                                   0x26,
                                                   0x6c,
                                                   0xb0,
                                                   0x33,
                                                   0x24,
                                                   0xfc,
                                                   0xde,
                                                   0x43,
                                                   0xda,
                                                   0xf9,
                                                   0x2e,
                                                   0xbf,
                                                   0xaa};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("Another test key - Two blocks message (32 bytes)") {
    // Given: A test key and a message that is exactly two Poly1305 blocks long.
    static const uint8_t key[POLY1305_KEY_BYTES] = {
        0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b,
        0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
        0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20};
    const char message[] = "This message is thirtytwo bytes!";

    // When: Calculate the hash.
    uint8_t hash[POLY1305_HASH_BYTES];
    poly1305(key, (const uint8_t*)message, sizeof(message) - 1, hash);

    // Then: The hash should have the expected value.
    const uint8_t expected[POLY1305_HASH_BYTES] = {0x24,
                                                   0xf6,
                                                   0xd0,
                                                   0x8e,
                                                   0x3a,
                                                   0xa6,
                                                   0xed,
                                                   0xbf,
                                                   0x1e,
                                                   0xc9,
                                                   0x33,
                                                   0xae,
                                                   0x3e,
                                                   0x86,
                                                   0x80,
                                                   0xb3};
    CHECK_EQ_BUF(hash, expected);
  }
}
