// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/serpent.h"
#include "minitest.h"

#include <stdint.h>
#include <string.h>

TEST_MAIN("Serpent-CBC-256 cipher") {
  TEST_CASE("All zero plaintext, key and IV") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0};
    const uint8_t key[32] = {0};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0x49,
                                  0x67,
                                  0x2b,
                                  0xa8,
                                  0x98,
                                  0xd9,
                                  0x8d,
                                  0xf9,
                                  0x50,
                                  0x19,
                                  0x18,
                                  0x04,
                                  0x45,
                                  0x49,
                                  0x10,
                                  0x89};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ecb_tbl.txt - I=1") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0xD0,
                                   0x95,
                                   0x57,
                                   0x6F,
                                   0xCE,
                                   0xA3,
                                   0xE3,
                                   0xA7,
                                   0xED,
                                   0x98,
                                   0xD9,
                                   0xF2,
                                   0x90,
                                   0x73,
                                   0xD7,
                                   0x8E};
    const uint8_t key[32] = {0};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0xB9,
                                  0x0E,
                                  0xE5,
                                  0x86,
                                  0x2D,
                                  0xE6,
                                  0x91,
                                  0x68,
                                  0xF2,
                                  0xBD,
                                  0xD5,
                                  0x12,
                                  0x5B,
                                  0x45,
                                  0x47,
                                  0x2B};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ecb_iv.txt - 1st PT, Encrypt") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0x10,
                                   0x32,
                                   0x54,
                                   0x76,
                                   0x98,
                                   0xba,
                                   0xdc,
                                   0xfe,
                                   0xef,
                                   0xcd,
                                   0xab,
                                   0x89,
                                   0x67,
                                   0x45,
                                   0x23,
                                   0x01};
    const uint8_t key[32] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
                             0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,
                             0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88,
                             0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0x93,
                                  0xdf,
                                  0x9a,
                                  0x3c,
                                  0xaf,
                                  0xe3,
                                  0x87,
                                  0xbd,
                                  0x99,
                                  0x9e,
                                  0xeb,
                                  0xe3,
                                  0x93,
                                  0xa1,
                                  0x7f,
                                  0xca};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ecb_iv.txt - 1st CT, Decrypt") {
    // Given: A 16-byte ciphertext, a key and an all-zero IV (simulate ECB).
    const uint8_t ciphertext[16] = {0x10,
                                    0x32,
                                    0x54,
                                    0x76,
                                    0x98,
                                    0xba,
                                    0xdc,
                                    0xfe,
                                    0xef,
                                    0xcd,
                                    0xab,
                                    0x89,
                                    0x67,
                                    0x45,
                                    0x23,
                                    0x01};
    const uint8_t key[32] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
                             0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,
                             0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88,
                             0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, ciphertext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_decrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The plaintext should be correct.
    const uint8_t expected[16] = {0xe9,
                                  0x65,
                                  0xb8,
                                  0x01,
                                  0xe1,
                                  0xbf,
                                  0x5a,
                                  0x20,
                                  0x03,
                                  0x40,
                                  0x79,
                                  0x48,
                                  0xe0,
                                  0xca,
                                  0x83,
                                  0xc3};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("All zero plaintext & key, non-zero IV") {
    // Given: A 32-byte plaintext, a key and a non-zero IV.
    const uint8_t plaintext[32] = {0};
    const uint8_t key[32] = {0};
    const uint8_t iv[16] = {0x01,
                            0x21,
                            0x43,
                            0x65,
                            0x87,
                            0xa9,
                            0xcb,
                            0xed,
                            0xef,
                            0xcd,
                            0xab,
                            0x89,
                            0x67,
                            0x45,
                            0x23,
                            0x01};

    // When: Encrypt the message.
    uint8_t buffer[32];
    memcpy(buffer, plaintext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[32] = {
        0x01, 0x2a, 0x1c, 0x9e, 0x65, 0x9f, 0x4a, 0xd4, 0xf0, 0xbb, 0xf5,
        0xdc, 0xa5, 0xbd, 0x94, 0x87, 0x2c, 0x16, 0x73, 0x4c, 0x36, 0x89,
        0x65, 0x93, 0xfd, 0x6a, 0x07, 0xa9, 0x88, 0x71, 0x53, 0xa0};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("Hello world... 128 bytes") {
    // Given: A plaintext, a key and an IV.
    const uint8_t plaintext[128] =
        "Hello world! This is a small text that we want to encrypt. We make it "
        "long enough to contain a multiple of 64 bytes...........!\n";
    const uint8_t key[32] = {0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef,
                             0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
                             0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88,
                             0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00};
    const uint8_t iv[16] = {0x00,
                            0xa0,
                            0x30,
                            0x40,
                            0x50,
                            0x60,
                            0x70,
                            0x89,
                            0xab,
                            0x12,
                            0x39,
                            0x87,
                            0x12,
                            0x3a,
                            0xbc,
                            0xde};

    // When: Encrypt the message.
    uint8_t buffer[128];
    memcpy(buffer, plaintext, sizeof(buffer));
    serpent256_ctx_t ctx;
    serpent_256_init(&ctx, key, iv);
    serpent_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The plaintext should be correct.
    const uint8_t expected[128] = {
        0xb8, 0xbd, 0x51, 0x7a, 0xe5, 0x8a, 0x8a, 0x3c, 0x5b, 0x34, 0xcf, 0x8e,
        0xfd, 0xc7, 0x7d, 0x17, 0x7a, 0xb6, 0xd3, 0xa7, 0xad, 0x0e, 0x05, 0xf7,
        0x75, 0x68, 0x28, 0x60, 0x59, 0xe2, 0xff, 0x2e, 0xbe, 0x94, 0x19, 0x5c,
        0xe9, 0x65, 0x51, 0x69, 0xed, 0xdb, 0x18, 0xc8, 0x54, 0xb3, 0x4a, 0x64,
        0x28, 0x10, 0xfb, 0x04, 0x90, 0x65, 0xa9, 0x45, 0x04, 0xe0, 0x51, 0x6b,
        0x85, 0xba, 0x85, 0xb2, 0x92, 0x43, 0x81, 0x71, 0x22, 0x53, 0xf4, 0xda,
        0xbc, 0x40, 0xae, 0xfc, 0xfb, 0x41, 0x6d, 0xce, 0x41, 0xea, 0x3a, 0x37,
        0xad, 0x0f, 0x7e, 0xe5, 0xc8, 0xfb, 0x47, 0x0e, 0x13, 0x86, 0xa4, 0x31,
        0x7c, 0xe4, 0xfb, 0xc9, 0xcc, 0x1f, 0xf3, 0x43, 0xb9, 0x29, 0x19, 0x7c,
        0xcf, 0xd0, 0x6c, 0xe9, 0x73, 0x31, 0x89, 0xec, 0x3a, 0xa3, 0xd4, 0x98,
        0x90, 0xd7, 0x74, 0x06, 0x20, 0xe3, 0xa7, 0xbb};
    CHECK_EQ_BUF(buffer, expected);
  }
}
