// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/sha2.h"
#include "minitest.h"

#include <stdint.h>

TEST_MAIN("SHA2 hashing algorithm") {
  TEST_CASE("SHA2-256 - Empty message") {
    // Given: An empty (zero length) message string.
    const uint8_t msg[] = "";

    // When: Calculate the SHA2 hash.
    uint8_t hash[SHA256_HASH_BYTES];
    sha2_256(msg, sizeof(msg) - 1, hash);

    // Then: The hash should be correct.
    const uint8_t expected[SHA256_HASH_BYTES] = {
        0xe3, 0xb0, 0xc4, 0x42, 0x98, 0xfc, 0x1c, 0x14, 0x9a, 0xfb, 0xf4,
        0xc8, 0x99, 0x6f, 0xb9, 0x24, 0x27, 0xae, 0x41, 0xe4, 0x64, 0x9b,
        0x93, 0x4c, 0xa4, 0x95, 0x99, 0x1b, 0x78, 0x52, 0xb8, 0x55};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("SHA2-256 - Quick brown fox") {
    // Given: A short message string.
    const uint8_t msg[] = "The quick brown fox jumps over the lazy dog";

    // When: Calculate the SHA2 hash.
    uint8_t hash[SHA256_HASH_BYTES];
    sha2_256(msg, sizeof(msg) - 1, hash);

    // Then: The hash should be correct.
    const uint8_t expected[SHA256_HASH_BYTES] = {
        0xd7, 0xa8, 0xfb, 0xb3, 0x07, 0xd7, 0x80, 0x94, 0x69, 0xca, 0x9a,
        0xbc, 0xb0, 0x08, 0x2e, 0x4f, 0x8d, 0x56, 0x51, 0xe4, 0x6d, 0x3c,
        0xdb, 0x76, 0x2d, 0x02, 0xd0, 0xbf, 0x37, 0xc9, 0xe5, 0x92};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("SHA2-256 - Lorem ipsum") {
    // Given: A long message string.
    const uint8_t msg[] =
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam "
        "commodo venenatis justo vitae imperdiet. Phasellus vitae sapien quis "
        "sem faucibus volutpat. Duis convallis nisl magna, quis semper nunc "
        "vulputate molestie. Praesent quis mauris aliquet orci lobortis "
        "commodo. Nam eget purus purus. Fusce eget luctus purus. Aenean "
        "sollicitudin efficitur ornare. Morbi et dignissim elit, nec egestas "
        "felis. Ut ac eros augue. Duis ut fringilla mi. Mauris et justo nec "
        "turpis volutpat ullamcorper ut ac dolor. Vivamus varius, metus eget "
        "placerat pharetra, eros lectus luctus arcu, et tincidunt magna nisl "
        "sit amet turpis. Donec sodales in sapien in pharetra. Cras vel magna "
        "nec purus efficitur hendrerit. Nulla porta lacinia lorem sit amet "
        "eleifend. Quisque tempus quis nulla ac semper. Curabitur vehicula "
        "sapien non neque lobortis tempus. Etiam at fringilla justo. In vel "
        "ipsum pretium, sagittis purus sed, laoreet libero. Nullam dapibus "
        "tortor vitae eros gravida, a hendrerit elit gravida.";

    // When: Calculate the SHA2 hash.
    uint8_t hash[SHA256_HASH_BYTES];
    sha2_256(msg, sizeof(msg) - 1, hash);

    // Then: The hash should be correct.
    const uint8_t expected[SHA256_HASH_BYTES] = {
        0x5c, 0xac, 0xd4, 0xe3, 0x14, 0x11, 0x37, 0x8d, 0x5d, 0xcc, 0xbc,
        0x92, 0x03, 0x0f, 0x0b, 0xb9, 0xf8, 0x3e, 0x1f, 0x5b, 0x09, 0x5a,
        0xfa, 0x5b, 0x86, 0x11, 0x7b, 0xd7, 0x85, 0xe5, 0x8a, 0x84};
    CHECK_EQ_BUF(hash, expected);
  }

  TEST_CASE("SHA2-256 - Chunk size edge case") {
    // Given: A message that is one byte less than a full SHA2 chunk (64 bytes).
    const uint8_t msg[] =
        "This message is exactly sixtythree bytes long excluding the nil";

    // When: Calculate the SHA2 hash.
    uint8_t hash[SHA256_HASH_BYTES];
    sha2_256(msg, sizeof(msg) - 1, hash);

    // Then: The hash should be correct.
    const uint8_t expected[SHA256_HASH_BYTES] = {
        0xaa, 0xe5, 0x84, 0xfa, 0x53, 0x1d, 0xbe, 0x6a, 0x60, 0x13, 0xf1,
        0x80, 0xfb, 0xca, 0xbd, 0x21, 0x6a, 0x98, 0x24, 0xd6, 0x5b, 0x1d,
        0x05, 0xa9, 0xa8, 0xe4, 0x73, 0x9f, 0x34, 0xa9, 0xdb, 0x73};
    CHECK_EQ_BUF(hash, expected);
  }
}
