// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../stringutils.h"

#include "minitest.h"
#include <stdbool.h>
#include <stdlib.h>

TEST_MAIN("stringutils - String utility functions") {
  TEST_CASE("string_copy - Copy a string") {
    // Given: A zero-terminated string.
    const char* src = "Hello world!";

    // When: Copying the string to a sufficiently large target.
    char dst[20];
    const bool result = string_copy(src, dst, sizeof(dst));

    // Then: The operation succeeded and the strings are identical.
    CHECK_EQ_BOOL(result, true);
    CHECK_EQ_STR(dst, src);
  }

  TEST_CASE("string_copy - Too short buffer") {
    // Given: A zero-terminated string.
    const char* src = "Hello world!";

    // When: Copying the string to a too small target.
    char dst[4];
    const bool result = string_copy(src, dst, sizeof(dst));

    // Then: The operation did not succeed.
    CHECK_EQ_BOOL(result, false);
  }

  TEST_CASE("numchars_utf8 - Valid Unicode string") {
    // Given: A zero-terminated string with 1-4 byte UTF-8 encoded code points.
    const char src[] = {
        (char)0x41U,  // "A" = U+41    (1/1)
        (char)0xc2U,  // "©" = U+A9    (1/2)
        (char)0xa9U,  //               (2/2)
        (char)0xe1U,  // "Ḉ" = U+1E08  (1/3)
        (char)0xb8U,  //               (2/3)
        (char)0x88U,  //               (3/3)
        (char)0xf0U,  // "🂡" = U+1F0A1 (1/4)
        (char)0x9fU,  //               (2/4)
        (char)0x82U,  //               (3/4)
        (char)0xa1U,  //               (4/4)
        (char)0x00U   // NUL
    };

    // When: Count the number of Unicode characters.
    const size_t result = numchars_utf8(src);

    // Then: The result is correct.
    CHECK_EQ_UINT(result, 4);
  }

  TEST_CASE("numchars_utf8 - Invalid Unicode string") {
    // Given: A zero-terminated invalid UTF-8 string.
    const char invalid[] = {
        (char)0xc0U,  // Invalid 2-byte sequence (1/2)
        (char)0x01U,  //                         (2/2)
        (char)0x00U   // NUL
    };

    // When: Count the number of Unicode characters.
    const size_t result = numchars_utf8(invalid);

    // Then: The result is correct.
    CHECK_EQ_UINT(result, 0);
  }

  TEST_CASE("numchars_utf16 - Valid Unicode string") {
    // Given: A zero-terminated string with code points that would encode to
    // 1-4 bytes in UTF-8.
    const uint16_t src[6] = {
        0x0041U,  // "A" = U+41    (1/1)
        0x00a9U,  // "©" = U+A9    (1/1)
        0x1e08U,  // "Ḉ" = U+1E08  (1/1)
        0xd83cU,  // "🂡" = U+1F0A1 (1/2)
        0xdca1U,  //               (2/2)
        0x0000U   // NUL
    };

    // When: Count the number of Unicode characters.
    const size_t result = numchars_utf16(src);

    // Then: The result is correct.
    CHECK_EQ_UINT(result, 4);
  }

  TEST_CASE("numchars_utf16 - Invalid Unicode string") {
    // Given: A zero-terminated string with invalid UTF-16 encoding.
    const uint16_t invalid[] = {
        0xd800U,  // Invalid surrogate pair sequence (1/2)
        0x0001U,  //                                 (2/2)
        0x0000U   // NUL
    };

    // When: Count the number of Unicode characters.
    const size_t result = numchars_utf16(invalid);

    // Then: The result is correct.
    CHECK_EQ_UINT(result, 0);
  }

  TEST_CASE("utf8_to_utf16 - Valid Unicode string") {
    // Given: A zero-terminated string with 1-4 byte UTF-8 encoded code points.
    const char src[] = {
        (char)0x41U,  // "A" = U+41    (1/1)
        (char)0xc2U,  // "©" = U+A9    (1/2)
        (char)0xa9U,  //               (2/2)
        (char)0xe1U,  // "Ḉ" = U+1E08  (1/3)
        (char)0xb8U,  //               (2/3)
        (char)0x88U,  //               (3/3)
        (char)0xf0U,  // "🂡" = U+1F0A1 (1/4)
        (char)0x9fU,  //               (2/4)
        (char)0x82U,  //               (3/4)
        (char)0xa1U,  //               (4/4)
        (char)0x00U   // NUL
    };

    // When: Converting the string to a sufficiently large target.
    uint16_t dst[6];
    const bool result = utf8_to_utf16(src, dst, sizeof(dst) / 2U);

    // Then: The operation succeeded and the target string has the expected
    // value.
    const uint16_t expected[6] = {
        0x0041U,  // "A" = U+41    (1/1)
        0x00a9U,  // "©" = U+A9    (1/1)
        0x1e08U,  // "Ḉ" = U+1E08  (1/1)
        0xd83cU,  // "🂡" = U+1F0A1 (1/2)
        0xdca1U,  //               (2/2)
        0x0000U   // NUL
    };
    CHECK_EQ_BOOL(result, true);
    CHECK_EQ_BUF(dst, expected);
  }

  TEST_CASE("utf8_to_utf16 - Invalid Unicode string") {
    // Given: Zero-terminated strings with invalid 2-4 byte UTF-8 encodings.
    const char invalid2[] = {
        (char)0xc0U,  // Invalid 2-byte sequence (1/2)
        (char)0x01U,  //                         (2/2)
        (char)0x00U   // NUL
    };
    const char invalid3[] = {
        (char)0xe0U,  // Invalid 3-byte sequence (1/3)
        (char)0x01U,  //                         (2/3)
        (char)0x01U,  //                         (3/3)
        (char)0x00U   // NUL
    };
    const char invalid4[] = {
        (char)0xf0U,  // Invalid 4-byte sequence (1/4)
        (char)0x01U,  //                         (2/4)
        (char)0x01U,  //                         (3/4)
        (char)0x01U,  //                         (4/4)
        (char)0x00U   // NUL
    };
    const char invalidFF[] = {
        (char)0xffU,  // Invalid UTF-8 byte (can never occurr)
        (char)0x00U   // NUL
    };

    // When: Converting the string to a sufficiently large target.
    uint16_t dst[100];
    const bool result2 = utf8_to_utf16(invalid2, dst, sizeof(dst) / 2U);
    const bool result3 = utf8_to_utf16(invalid3, dst, sizeof(dst) / 2U);
    const bool result4 = utf8_to_utf16(invalid4, dst, sizeof(dst) / 2U);
    const bool resultFF = utf8_to_utf16(invalidFF, dst, sizeof(dst) / 2U);

    // Then: The operations did not succeed.
    CHECK_EQ_BOOL(result2, false);
    CHECK_EQ_BOOL(result3, false);
    CHECK_EQ_BOOL(result4, false);
    CHECK_EQ_BOOL(resultFF, false);
  }

  TEST_CASE("utf8_to_utf16 - Unsupported UTF-16 code point") {
    // Given: Zero-terminated strings with invalid code point.
    const char invalid[] = {
        (char)0xedU,  // 3-byte sequence: 0xd800 (1/3)
        (char)0xa0U,  //                         (2/3)
        (char)0x80U,  //                         (3/3)
        (char)0x00U   // NUL
    };

    // When: Converting the string to a sufficiently large target.
    uint16_t dst[100];
    const bool result = utf8_to_utf16(invalid, dst, sizeof(dst) / 2U);

    // Then: The operations did not succeed.
    CHECK_EQ_BOOL(result, false);
  }

  TEST_CASE("utf8_to_utf16 - Too short buffer") {
    // Given: A zero-terminated string.
    const char* src = "Hello world!";

    // When: Converting the string to a too small target.
    uint16_t dst[4];
    const bool result = utf8_to_utf16(src, dst, sizeof(dst) / 2U);

    // Then: The operation did not succeed.
    CHECK_EQ_BOOL(result, false);
  }

  TEST_CASE("utf16_to_utf8 - Valid Unicode string") {
    // Given: A zero-terminated string with code points that will encode to 1-4
    // bytes in UTF-8.
    const uint16_t src[6] = {
        0x0041U,  // "A" = U+41    (1/1)
        0x00a9U,  // "©" = U+A9    (1/1)
        0x1e08U,  // "Ḉ" = U+1E08  (1/1)
        0xd83cU,  // "🂡" = U+1F0A1 (1/2)
        0xdca1U,  //               (2/2)
        0x0000U   // NUL
    };

    // When: Converting the string to a sufficiently large target.
    char dst[11];
    const bool result = utf16_to_utf8(src, dst, sizeof(dst));

    // Then: The operation succeeded and the target string has the expected
    // value.
    const char expected[] = {
        (char)0x41U,  // "A" = U+41    (1/1)
        (char)0xc2U,  // "©" = U+A9    (1/2)
        (char)0xa9U,  //               (2/2)
        (char)0xe1U,  // "Ḉ" = U+1E08  (1/3)
        (char)0xb8U,  //               (2/3)
        (char)0x88U,  //               (3/3)
        (char)0xf0U,  // "🂡" = U+1F0A1 (1/4)
        (char)0x9fU,  //               (2/4)
        (char)0x82U,  //               (3/4)
        (char)0xa1U,  //               (4/4)
        (char)0x00U   // NUL
    };
    CHECK_EQ_BOOL(result, true);
    CHECK_EQ_BUF(dst, expected);
  }

  TEST_CASE("utf16_to_utf8 - Invalid Unicode string") {
    // Given: A zero-terminated string with invalid UTF-16 encoding.
    const uint16_t invalidA[] = {
        0xd800U,  // Invalid surrogate pair sequence (1/2)
        0x0001U,  //                                 (2/2)
        0x0000U   // NUL
    };
    const uint16_t invalidB[] = {
        0xdc00U,  // Invalid surrogate pair sequence (1/2)
        0xd800U,  //                                 (2/2)
        0x0000U   // NUL
    };

    // When: Converting the string to a sufficiently large target.
    char dst[100];
    const bool resultA = utf16_to_utf8(invalidA, dst, sizeof(dst));
    const bool resultB = utf16_to_utf8(invalidB, dst, sizeof(dst));

    // Then: The operation did not succeed.
    CHECK_EQ_BOOL(resultA, false);
    CHECK_EQ_BOOL(resultB, false);
  }

  TEST_CASE("utf16_to_utf8 - Too short buffer") {
    // Given: A zero-terminated string.
    const uint16_t src[] = {0x0041U, 0x0042U, 0x0043U, 0x0000U};

    // When: Converting the string to a too small target.
    char dst[2];
    const bool result = utf16_to_utf8(src, dst, sizeof(dst));

    // Then: The operation did not succeed.
    CHECK_EQ_BOOL(result, false);
  }

  TEST_CASE("utf8_to_utf16_alloc - Valid Unicode string") {
    // Given: A zero-terminated string with 1-4 byte UTF-8 encoded code points.
    const char src[] = {
        (char)0x41U,  // "A" = U+41    (1/1)
        (char)0xc2U,  // "©" = U+A9    (1/2)
        (char)0xa9U,  //               (2/2)
        (char)0xe1U,  // "Ḉ" = U+1E08  (1/3)
        (char)0xb8U,  //               (2/3)
        (char)0x88U,  //               (3/3)
        (char)0xf0U,  // "🂡" = U+1F0A1 (1/4)
        (char)0x9fU,  //               (2/4)
        (char)0x82U,  //               (3/4)
        (char)0xa1U,  //               (4/4)
        (char)0x00U   // NUL
    };

    // When: Converting the string.
    uint16_t* dst = utf8_to_utf16_alloc(src);

    // Then: The operation succeeded and the target string has the expected
    // value.
    const uint16_t expected[6] = {
        0x0041U,  // "A" = U+41    (1/1)
        0x00a9U,  // "©" = U+A9    (1/1)
        0x1e08U,  // "Ḉ" = U+1E08  (1/1)
        0xd83cU,  // "🂡" = U+1F0A1 (1/2)
        0xdca1U,  //               (2/2)
        0x0000U   // NUL
    };
    CHECK_EQ_BOOL(dst != NULL, true);
    if (dst != NULL) {
      CHECK_EQ_BUF(dst, expected);
    }

    // Post: Free allocated memory.
    free(dst);
  }

  TEST_CASE("utf16_to_utf8_alloc - Valid Unicode string") {
    // Given: A zero-terminated string with code points that will encode to 1-4
    // bytes in UTF-8.
    const uint16_t src[6] = {
        0x0041U,  // "A" = U+41    (1/1)
        0x00a9U,  // "©" = U+A9    (1/1)
        0x1e08U,  // "Ḉ" = U+1E08  (1/1)
        0xd83cU,  // "🂡" = U+1F0A1 (1/2)
        0xdca1U,  //               (2/2)
        0x0000U   // NUL
    };

    // When: Converting the string.
    char* dst = utf16_to_utf8_alloc(src);

    // Then: The operation succeeded and the target string has the expected
    // value.
    const char expected[] = {
        (char)0x41U,  // "A" = U+41    (1/1)
        (char)0xc2U,  // "©" = U+A9    (1/2)
        (char)0xa9U,  //               (2/2)
        (char)0xe1U,  // "Ḉ" = U+1E08  (1/3)
        (char)0xb8U,  //               (2/3)
        (char)0x88U,  //               (3/3)
        (char)0xf0U,  // "🂡" = U+1F0A1 (1/4)
        (char)0x9fU,  //               (2/4)
        (char)0x82U,  //               (3/4)
        (char)0xa1U,  //               (4/4)
        (char)0x00U   // NUL
    };
    CHECK_EQ_BOOL(dst != NULL, true);
    if (dst != NULL) {
      CHECK_EQ_BUF(dst, expected);
    }

    // Post: Free allocated memory.
    free(dst);
  }
}
