// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/twofish.h"
#include "minitest.h"

#include <stdint.h>
#include <string.h>

// Some test vectors are from the paper "Twofish: A 128-Bit Block Cipher":
// https://www.schneier.com/wp-content/uploads/2016/02/paper-twofish-paper.pdf

TEST_MAIN("Twofish-CBC-256 cipher") {
  TEST_CASE("ECB Tables Known Answer Test - I=1") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00,
                                   0x00};
    const uint8_t key[32] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0x57,
                                  0xFF,
                                  0x73,
                                  0x9D,
                                  0x4D,
                                  0xC9,
                                  0x2C,
                                  0x1B,
                                  0xD7,
                                  0xFC,
                                  0x01,
                                  0x70,
                                  0x0C,
                                  0xC8,
                                  0x21,
                                  0x6F};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ECB Tables Known Answer Test - I=2") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0x57,
                                   0xFF,
                                   0x73,
                                   0x9D,
                                   0x4D,
                                   0xC9,
                                   0x2C,
                                   0x1B,
                                   0xD7,
                                   0xFC,
                                   0x01,
                                   0x70,
                                   0x0C,
                                   0xC8,
                                   0x21,
                                   0x6F};
    const uint8_t key[32] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0xD4,
                                  0x3B,
                                  0xB7,
                                  0x55,
                                  0x6E,
                                  0xA3,
                                  0x2E,
                                  0x46,
                                  0xF2,
                                  0xA2,
                                  0x82,
                                  0xB7,
                                  0xD4,
                                  0x5B,
                                  0x4E,
                                  0x0D};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ECB Tables Known Answer Test - I=3") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0xD4,
                                   0x3B,
                                   0xB7,
                                   0x55,
                                   0x6E,
                                   0xA3,
                                   0x2E,
                                   0x46,
                                   0xF2,
                                   0xA2,
                                   0x82,
                                   0xB7,
                                   0xD4,
                                   0x5B,
                                   0x4E,
                                   0x0D};
    const uint8_t key[32] = {0x57, 0xFF, 0x73, 0x9D, 0x4D, 0xC9, 0x2C, 0x1B,
                             0xD7, 0xFC, 0x01, 0x70, 0x0C, 0xC8, 0x21, 0x6F,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0x90,
                                  0xAF,
                                  0xE9,
                                  0x1B,
                                  0xB2,
                                  0x88,
                                  0x54,
                                  0x4F,
                                  0x2C,
                                  0x32,
                                  0xDC,
                                  0x23,
                                  0x9B,
                                  0x26,
                                  0x35,
                                  0xE6};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("ECB Tables Known Answer Test - I=4") {
    // Given: A 16-byte plaintext, a key and an all-zero IV (simulate ECB).
    const uint8_t plaintext[16] = {0x90,
                                   0xAF,
                                   0xE9,
                                   0x1B,
                                   0xB2,
                                   0x88,
                                   0x54,
                                   0x4F,
                                   0x2C,
                                   0x32,
                                   0xDC,
                                   0x23,
                                   0x9B,
                                   0x26,
                                   0x35,
                                   0xE6};
    const uint8_t key[32] = {0xD4, 0x3B, 0xB7, 0x55, 0x6E, 0xA3, 0x2E, 0x46,
                             0xF2, 0xA2, 0x82, 0xB7, 0xD4, 0x5B, 0x4E, 0x0D,
                             0x57, 0xFF, 0x73, 0x9D, 0x4D, 0xC9, 0x2C, 0x1B,
                             0xD7, 0xFC, 0x01, 0x70, 0x0C, 0xC8, 0x21, 0x6F};
    const uint8_t iv[16] = {0};

    // When: Encrypt the message.
    uint8_t buffer[16];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[16] = {0x6C,
                                  0xB4,
                                  0x56,
                                  0x1C,
                                  0x40,
                                  0xBF,
                                  0x0A,
                                  0x97,
                                  0x05,
                                  0x93,
                                  0x1C,
                                  0xB6,
                                  0xD4,
                                  0x08,
                                  0xE7,
                                  0xFA};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("Encrypt a message") {
    // Given: A message, and a context initialized with a key and an IV.
    const uint8_t plaintext[32] = "Abcdefgh Hello world! 0187654321";
    const uint8_t key[32] = {0x19, 0x29, 0x39, 0x49, 0x59, 0x69, 0x79, 0x89,
                             0xab, 0xcb, 0xdb, 0xeb, 0xfb, 0xeb, 0xdb, 0xcb,
                             0xab, 0x83, 0x94, 0x76, 0xaf, 0x83, 0x28, 0x74,
                             0xfd, 0x8a, 0x75, 0xc9, 0x00, 0xd0, 0x8c, 0x6d};
    const uint8_t iv[16] = {0x01,
                            0x23,
                            0x45,
                            0x67,
                            0x89,
                            0xab,
                            0xcd,
                            0xef,
                            0x01,
                            0x23,
                            0x45,
                            0x67,
                            0x89,
                            0xab,
                            0xcd,
                            0xef};
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);

    // When: Encrypt the message.
    uint8_t buffer[32];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The ciphertext should be correct.
    const uint8_t expected[32] = {
        0x14, 0x73, 0x3d, 0xa8, 0x1f, 0x5f, 0xef, 0x53, 0xa8, 0x3e, 0xe9,
        0xd4, 0x41, 0x02, 0xc0, 0x22, 0xfa, 0x54, 0x35, 0x3b, 0x19, 0x64,
        0xee, 0x99, 0x94, 0xc4, 0x53, 0x66, 0xed, 0x20, 0x30, 0xae};
    CHECK_EQ_BUF(buffer, expected);
  }

  TEST_CASE("Encrypt+decrypt a message") {
    // Given: A message, a key and an IV.
    const uint8_t plaintext[32] = "Abcdefgh Hello world! 0187654321";
    const uint8_t key[32] = {0x19, 0x29, 0x39, 0x49, 0x59, 0x69, 0x79, 0x89,
                             0xab, 0xcb, 0xdb, 0xeb, 0xfb, 0xeb, 0xdb, 0xcb,
                             0xab, 0x83, 0x94, 0x76, 0xaf, 0x83, 0x28, 0x74,
                             0xfd, 0x8a, 0x75, 0xc9, 0x00, 0xd0, 0x8c, 0x6d};
    const uint8_t iv[16] = {0x01,
                            0x23,
                            0x45,
                            0x67,
                            0x89,
                            0xab,
                            0xcd,
                            0xef,
                            0x01,
                            0x23,
                            0x45,
                            0x67,
                            0x89,
                            0xab,
                            0xcd,
                            0xef};

    // When: Encrypt and decrypt the message.
    uint8_t buffer[32];
    memcpy(buffer, plaintext, sizeof(buffer));
    twofish256_ctx_t ctx;
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_encrypt_buffer(&ctx, buffer, sizeof(buffer));
    twofish_256_init(&ctx, key, iv);
    twofish_256_cbc_decrypt_buffer(&ctx, buffer, sizeof(buffer));

    // Then: The decrypted ciphertext should equal the plaintext.
    CHECK_EQ_BUF(buffer, plaintext);
  }
}
