// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/u128.h"
#include "minitest.h"

#include <stdint.h>

TEST_MAIN("u128 - Unsigned 128-bit integer") {
  TEST_CASE("Constructor") {
    // Given: Two 64-bit words, high and low parts of a 128-bit integer.
    const uint64_t hi = 0x123456789abcdef0;
    const uint64_t lo = 0xaabbccddeeff5555;

    // When: Construct the 128-bit integer.
    u128 value = to_u128(hi, lo);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINT(hi64(value), hi);
    CHECK_EQ_UINT(lo64(value), lo);
  }

  TEST_CASE("Bitwise AND") {
    // Given: Two 128-bit values.
    const u128 a = to_u128(0x123456789abcdef0, 0x9999999999999999);
    const u128 b = to_u128(0x000000000000aaaa, 0xffff000000000001);

    // When: Perform the arithmetic operation.
    u128 value = and128(a, b);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINTX(hi64(value), 0x0000000000008aa0);
    CHECK_EQ_UINTX(lo64(value), 0x9999000000000001);
  }

  TEST_CASE("Bitwise OR") {
    // Given: Two 128-bit values.
    const u128 a = to_u128(0x123456789abcdef0, 0x9999999999999999);
    const u128 b = to_u128(0x000000000000aaaa, 0xffff000000000001);

    // When: Perform the arithmetic operation.
    u128 value = or128(a, b);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINTX(hi64(value), 0x123456789abcfefa);
    CHECK_EQ_UINTX(lo64(value), 0xffff999999999999);
  }

  TEST_CASE("Bitwise XOR") {
    // Given: Two 128-bit values.
    const u128 a = to_u128(0x123456789abcdef0, 0x9999999999999999);
    const u128 b = to_u128(0x000000000000aaaa, 0xffff000000000001);

    // When: Perform the arithmetic operation.
    u128 value = xor128(a, b);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINTX(hi64(value), 0x123456789abc745aU);
    CHECK_EQ_UINTX(lo64(value), 0x6666999999999998U);
  }

  TEST_CASE("Addition") {
    // Given: Two 128-bit values.
    const u128 a = to_u128(0x123456789abcdef0, 0x9999999999999999);
    const u128 b = to_u128(0x000000000000aaaa, 0xffff000000000001);

    // When: Perform the arithmetic operation.
    u128 value = add128(a, b);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINTX(hi64(value), 0x123456789abd899b);
    CHECK_EQ_UINTX(lo64(value), 0x999899999999999a);
  }

  TEST_CASE("Multiplication") {
    // Given: Two 128-bit values.
    const u128 a = to_u128(0x123456789abcdef0, 0x9999999999999999);
    const u128 b = to_u128(0x000000000000aaaa, 0xffff000000000001);

    // When: Perform the arithmetic operation.
    u128 value = mul128(a, b);

    // Then: The high and low parts should be correct.
    CHECK_EQ_UINTX(hi64(value), 0xccdd56789abc788a);
    CHECK_EQ_UINTX(lo64(value), 0x0000999999999999);
  }
}
