// -*- mode: c; tab-width: 2; indent-tabs-mode: nil; -*-
//------------------------------------------------------------------------------
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <https://unlicense.org/>
//------------------------------------------------------------------------------

#include "../crypto/zkdf.h"
#include "minitest.h"

#include <stdint.h>

static const uint8_t TEST_SALT[ZKDF_SALT_BYTES] = {
    1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16,
    17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32};

static const int TEST_STD_PASSES = 1;
static const int TEST_HIGH_PASSES = 20;  // Not too high (keep test times short)

TEST_MAIN("ZKDF key derivation function") {
  TEST_CASE("Empty passphrase - zero salt") {
    // Given: An empty passphrase and an all-zero salt.
    const char passphrase[] = "";
    const uint8_t salt[ZKDF_SALT_BYTES] = {0};

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, salt, TEST_STD_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0xce, 0xcc, 0x6a, 0x9f, 0xf1, 0x83, 0x97, 0x1f, 0x3b, 0x06, 0x85,
        0x51, 0x57, 0x49, 0x01, 0xe9, 0x65, 0xa4, 0x67, 0x3a, 0x0b, 0xae,
        0xdb, 0x6f, 0xaf, 0x7b, 0xc2, 0xfb, 0x3d, 0x9b, 0x1e, 0x95};
    CHECK_EQ_BUF(key, expected);
  }

  TEST_CASE("Empty passphrase - non-zero salt") {
    // Given: An empty passphrase.
    const char passphrase[] = "";

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, TEST_SALT, TEST_STD_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0xf9, 0xc8, 0x42, 0x9a, 0x9d, 0xb8, 0xa2, 0x73, 0xea, 0x6c, 0xae,
        0xdd, 0x28, 0xb5, 0xbc, 0xf5, 0xce, 0x23, 0xfa, 0x76, 0x69, 0x58,
        0x5e, 0x25, 0x85, 0xe5, 0x05, 0x1c, 0x4f, 0xde, 0x57, 0xa2};
    CHECK_EQ_BUF(key, expected);
  }

  TEST_CASE("A short passphrase") {
    // Given: A short passphrase.
    const char passphrase[] = "Hello!";

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, TEST_SALT, TEST_STD_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0x2e, 0xad, 0x75, 0xe3, 0x2b, 0xc1, 0x36, 0x00, 0x1b, 0x72, 0xf7,
        0x29, 0xac, 0xcf, 0xc2, 0x0f, 0x5d, 0x0f, 0x96, 0x5f, 0x79, 0x08,
        0x50, 0x8c, 0x1f, 0x3d, 0x51, 0x10, 0x9d, 0xb7, 0x60, 0x9e};
    CHECK_EQ_BUF(key, expected);
  }

  TEST_CASE("A long passphrase") {
    // Given: A long passphrase.
    const char passphrase[] =
        "Hello world! This is a longer passphrase that should be pretty hard "
        "to guess. Especially if we spice it up with some symbols: "
        "!#¤%&/()=?-_.,+*<>| - And of course some numbers too: 0192837465";

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, TEST_SALT, TEST_STD_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0x41, 0xd2, 0x06, 0xb1, 0xb9, 0x30, 0x89, 0xbd, 0xce, 0xb7, 0x42,
        0x68, 0xe6, 0x9b, 0x28, 0x37, 0x58, 0x8b, 0x2a, 0x6c, 0xb3, 0xa4,
        0x91, 0xe9, 0x07, 0x54, 0x74, 0xbe, 0x54, 0x9d, 0x41, 0x33};
    CHECK_EQ_BUF(key, expected);
  }

  TEST_CASE("A long passphrase, with one character modified") {
    // Given: A long passphrase.
    const char passphrase[] =
        "Iello world! This is a longer passphrase that should be pretty hard "
        "to guess. Especially if we spice it up with some symbols: "
        "!#¤%&/()=?-_.,+*<>| - And of course some numbers too: 0192837465";

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, TEST_SALT, TEST_STD_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0xd6, 0x9f, 0x16, 0x59, 0xe7, 0x5a, 0xea, 0x81, 0xc8, 0x58, 0xe8,
        0x44, 0xe3, 0xcb, 0x3e, 0x4d, 0xc9, 0x2c, 0x6f, 0x38, 0x84, 0xef,
        0x6f, 0x3e, 0xa0, 0x01, 0x28, 0x00, 0xa0, 0x69, 0xa1, 0x5b};
    CHECK_EQ_BUF(key, expected);
  }

  TEST_CASE("Higher computational cost") {
    // Given: A short passphrase.
    const char passphrase[] = "Hello!";

    // When: Generate the key.
    uint8_t key[ZKDF_KEY_BYTES];
    zkdf_make_key(passphrase, TEST_SALT, TEST_HIGH_PASSES, key);

    // Then: The key should have the expected value.
    const uint8_t expected[ZKDF_KEY_BYTES] = {
        0x4f, 0xcf, 0x52, 0x2d, 0x71, 0xbc, 0x60, 0x17, 0x3b, 0xe4, 0xe7,
        0x0d, 0x40, 0x60, 0xe8, 0xed, 0x35, 0x0b, 0x60, 0xf4, 0x7c, 0xc1,
        0x19, 0x17, 0xbf, 0x63, 0x64, 0x0e, 0x65, 0xc3, 0x0c, 0xe4};
    CHECK_EQ_BUF(key, expected);
  }
}
